
/* This unit implements a dynamic data structure for one-dimensional 
   range maximum queries using a red-black tree as a basis. Each subtree 
   stores the maxima of its left and right subtrees. Queries are answered 
   almost like in range trees, except that points are stored in internal 
   nodes, not in leafs.  

   The extra complication is that the maxima need to be maintained
   during insertions, deletions, and rebalancing. This is accomplished
   by some extra updates during these operations; all operations still
   take O(log n) time. */ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>
#include <list>
#include <time.h>
#include <utility>
#include <iostream>

using namespace std;

#define compEQ(a,b) (a.i == b.i && a.j == b.j)
#define compLT(a,b) (a.i < b.i || (a.i==b.i && a.j<b.j))
#define compLEQ(a,b) (a.i < b.i || (a.i==b.i && a.j<=b.j))
#define compGT(a,b) compLEQ(b,a)
#define compGEQ(a,b) compLT(b,a)

#define min2(a,b) ((a)<(b)?(a):(b))
#define min3(a,b,c) ((a)<(b))?((a)<(c)?(a):(c)):((b)<(c)?(b):(c)))
#define max2(a,b) ((a)>(b)?(a):(b))
#define max3(a,b,c) ((a)>(b)?((a)>(c)?(a):(c)):((b)>(c)?(b):(c)))

/* implementation dependent declarations */
typedef enum {
   STATUS_OK,
   STATUS_MEM_EXHAUSTED,
   STATUS_DUPLICATE_KEY,
   STATUS_KEY_NOT_FOUND
} statusEnum;

class keyType {
public:
   int i; // primary key
   int j; // secondary key
};

/* user data stored in tree */
class recType {
public:
   int i; // label of a point
   int v; // value of a point
   int maxl; // maximum value in the left subtree
   int maxr; // maximum value in the right subtree
   int maxli; // label of maximum value in the left subtree
   int maxri; // label of maximum value in the right subtree
};

/* implementation independent declarations */
/* Red-Black tree description */
typedef enum { BLACK, RED } nodeColor;

class nodeType {
public:
    nodeType *left;       /* left child */
    nodeType *right;      /* right child */
    nodeType *parent;     /* parent */
    nodeColor color;            /* node color (BLACK, RED) */
    keyType key;                /* key used for searching */
    recType rec;                /* user data */
};


#define NIL &sentinel           /* all leafs are sentinels */

static nodeType sentinel = { NIL, NIL, 0, BLACK, {0, 0}, {INT_MIN, INT_MIN, INT_MIN,INT_MIN,INT_MIN,INT_MIN}};

class redBlackTree {
private:
   /* last node found, optimizes find/delete operations */
   nodeType *lastFind;
   /* root of Red-Black tree */
   nodeType *root;    
   void updatePathMax(nodeType *x);   
   void updateParentMax(nodeType *x);   
   void rotateLeft(nodeType *x);   
   void rotateRight(nodeType *x);
   void insertFixup(nodeType *x);
   void deleteFixup(nodeType *x);   
   statusEnum find(keyType key, recType *rec);     
   pair<int,int> updateMaximaRec(nodeType *current);
   void makeEmptyRec(nodeType *current);      
public:
   redBlackTree() {root = NIL;};
   ~redBlackTree() {makeEmpty();};
   statusEnum insertNode(keyType key, recType *rec);
   statusEnum deleteNode(keyType key);
   void updateMaximaAll(); // for testing the correctness of dynamic updates
   recType rangeMax(keyType l, keyType r);   
   void makeEmpty();   
};

void redBlackTree::updatePathMax(nodeType *x) {

   /*************************************************   
    * updates subtree maxima in path from x to root *
    *************************************************/ 

    nodeType *y;
    while (x != root) {
      if (x == x->parent->left) {
         y = x->parent->right;
         x->parent->rec.maxl = max3(x->rec.v,x->rec.maxl,x->rec.maxr);
	 if (x->parent->rec.maxl==x->rec.v)
	    x->parent->rec.maxli = x->rec.i;
	 else if (x->parent->rec.maxl==x->rec.maxl)
	    x->parent->rec.maxli = x->rec.maxli;
	 else   
            x->parent->rec.maxli = x->rec.maxri;
	 if (y!=NIL) {
            x->parent->rec.maxr = max3(y->rec.v,y->rec.maxl,y->rec.maxr);
	    if (x->parent->rec.maxr==y->rec.v)
	       x->parent->rec.maxri = y->rec.i;
   	    else if (x->parent->rec.maxr==y->rec.maxl)
   	       x->parent->rec.maxri = y->rec.maxli;
 	    else   
               x->parent->rec.maxri = y->rec.maxri;
	 }
	 else {
  	    x->parent->rec.maxr = INT_MIN;
	    x->parent->rec.maxri = INT_MIN;
	 }
      }
      else {
         y = x->parent->left;
         x->parent->rec.maxr = max3(x->rec.v,x->rec.maxl,x->rec.maxr);
	 if (x->parent->rec.maxr==x->rec.v)
	    x->parent->rec.maxri = x->rec.i;
	 else if (x->parent->rec.maxr==x->rec.maxl)
	    x->parent->rec.maxri = x->rec.maxli;
	 else   
            x->parent->rec.maxri = x->rec.maxri;	 
	 if (y!=NIL) {
            x->parent->rec.maxl = max3(y->rec.v,y->rec.maxl,y->rec.maxr);
	    if (x->parent->rec.maxl==y->rec.v)
	       x->parent->rec.maxli = y->rec.i;
   	    else if (x->parent->rec.maxl==y->rec.maxl)
   	       x->parent->rec.maxli = y->rec.maxli;
 	    else   
               x->parent->rec.maxli = y->rec.maxri;
	 }
	 else {
  	    x->parent->rec.maxl = INT_MIN;
	    x->parent->rec.maxli = INT_MIN;
	 }
      } 
      
      x = x->parent;
    }
}

void redBlackTree::updateParentMax(nodeType *x) {

   /******************************************   
    * updates subtree maxima for parent of x *
    ******************************************/ 

    if (x != root)
    {
      if (x == x->parent->left) {
         x->parent->rec.maxl = max3(x->rec.v,x->rec.maxl,x->rec.maxr);
	 if (x->parent->rec.maxl==x->rec.v)
	    x->parent->rec.maxli = x->rec.i;
	 else if (x->parent->rec.maxl==x->rec.maxl)
	    x->parent->rec.maxli = x->rec.maxli;
	 else   
         x->parent->rec.maxli = x->rec.maxri;
      }
      else {
         x->parent->rec.maxr = max3(x->rec.v,x->rec.maxl,x->rec.maxr);
	 if (x->parent->rec.maxr==x->rec.v)
	    x->parent->rec.maxri = x->rec.i;
	 else if (x->parent->rec.maxr==x->rec.maxl)
	    x->parent->rec.maxri = x->rec.maxli;
	 else   
         x->parent->rec.maxri = x->rec.maxri;	 
      } 
    }
}


void redBlackTree::rotateLeft(nodeType *x) {

   /**************************
    *  rotate node x to left *
    **************************/

    nodeType *y = x->right;

    /* establish x->right link */
    x->right = y->left;
    /* update maximum in the right subtree of x */
    x->rec.maxr = y->rec.maxl;
    x->rec.maxri = y->rec.maxli;
    /* update maximum in the left subtree (to come) of y */
    y->rec.maxl = max3(x->rec.v,x->rec.maxl,x->rec.maxr);
    if (y->rec.maxl == x->rec.v)
       y->rec.maxli = x->rec.i;
    else if (y->rec.maxl == x->rec.maxl) 
       y->rec.maxli = x->rec.maxli;
    else
       y->rec.maxli = x->rec.maxri;
       
    if (y->left != NIL) y->left->parent = x;

    /* establish y->parent link */
    if (y != NIL) y->parent = x->parent;
    if (x->parent) {
        if (x == x->parent->left) {
            x->parent->left = y;
            /* update maximum in the left subtree of x->parent */
            x->parent->rec.maxl = max3(y->rec.v,y->rec.maxl,y->rec.maxr);
	    if (x->parent->rec.maxl==y->rec.v)
	       x->parent->rec.maxli = y->rec.i;
	    else if (x->parent->rec.maxl==y->rec.maxl)
	       x->parent->rec.maxli = y->rec.maxli;
	    else
	       x->parent->rec.maxli = y->rec.maxri;	    
        }
        else {
            x->parent->right = y;
            /* update maximum in the right subtree of x->parent */
            x->parent->rec.maxr = max3(y->rec.v,y->rec.maxl,y->rec.maxr);
	    if (x->parent->rec.maxr==y->rec.v)
	       x->parent->rec.maxri = y->rec.i;
	    else if (x->parent->rec.maxr==y->rec.maxl)
	       x->parent->rec.maxri = y->rec.maxli;
	    else
	       x->parent->rec.maxri = y->rec.maxri;	    	    
        }
    } else {
        root = y;
    }

    /* link x and y */
    y->left = x;
    if (x != NIL) x->parent = y;
    updatePathMax(x);
}

void redBlackTree::rotateRight(nodeType *x) {

   /****************************
    *  rotate node x to right  *
    ****************************/

    nodeType *y = x->left;

    /* establish x->left link */
    x->left = y->right;
    /* update maximum in the left subtree of x */
    x->rec.maxl = y->rec.maxr;
    x->rec.maxli = y->rec.maxri;
    /* update maximum in the right subtree (to come) of y */
    y->rec.maxr = max3(x->rec.v,x->rec.maxl,x->rec.maxr);
    if (y->rec.maxr == x->rec.v)
       y->rec.maxri =x->rec.i;
    if (y->rec.maxr == x->rec.maxl)
       y->rec.maxri =x->rec.maxli;
    else
       y->rec.maxri =x->rec.maxri;

    if (y->right != NIL) y->right->parent = x;

    /* establish y->parent link */
    if (y != NIL) y->parent = x->parent;
    if (x->parent) {
        if (x == x->parent->right) {
            x->parent->right = y;
            /* update maximum in the right subtree of x->parent */
            x->parent->rec.maxr = max3(y->rec.v,y->rec.maxl,y->rec.maxr);
	    if (x->parent->rec.maxr==y->rec.v)
	       x->parent->rec.maxri = y->rec.i;
	    else if (x->parent->rec.maxr==y->rec.maxl)
	       x->parent->rec.maxri = y->rec.maxli;
	    else
	       x->parent->rec.maxri = y->rec.maxri;	    	    
        }
        else {
            x->parent->left = y;
            /* update maximum in the left subtree of x->parent */
            x->parent->rec.maxl = max3(y->rec.v,y->rec.maxl,y->rec.maxr);
	    if (x->parent->rec.maxl==y->rec.v)
	       x->parent->rec.maxli = y->rec.i;
	    else if (x->parent->rec.maxl==y->rec.maxl)
	       x->parent->rec.maxli = y->rec.maxli;
	    else
	       x->parent->rec.maxli = y->rec.maxri;	    	    
        }
    } else {
        root = y;
    }

    /* link x and y */
    y->right = x;
    if (x != NIL) x->parent = y;
    updatePathMax(x);
}

void redBlackTree::insertFixup(nodeType *x) {

   /*************************************
    *  maintain Red-Black tree balance  *
    *  after inserting node x           *
    *************************************/

    /* No changes needed here for maintaining the subtree maxima,
       except in rotation procedures */

    /* check Red-Black properties */
    while (x != root && x->parent->color == RED) {
        /* we have a violation */
        if (x->parent == x->parent->parent->left) {
            nodeType *y = x->parent->parent->right;
            if (y->color == RED) {

                /* uncle is RED */
                x->parent->color = BLACK;
                y->color = BLACK;
                x->parent->parent->color = RED;
                x = x->parent->parent;
            } else {

                /* uncle is BLACK */
                if (x == x->parent->right) {
                    /* make x a left child */
                    x = x->parent;
                    rotateLeft(x);
                }

                /* recolor and rotate */
                x->parent->color = BLACK;
                x->parent->parent->color = RED;
                rotateRight(x->parent->parent);
            }
        } else {

            /* mirror image of above code */
            nodeType *y = x->parent->parent->left;
            if (y->color == RED) {

                /* uncle is RED */
                x->parent->color = BLACK;
                y->color = BLACK;
                x->parent->parent->color = RED;
                x = x->parent->parent;
            } else {

                /* uncle is BLACK */
                if (x == x->parent->left) {
                    x = x->parent;
                    rotateRight(x);
                }
                x->parent->color = BLACK;
                x->parent->parent->color = RED;
                rotateLeft(x->parent->parent);
            }
        }
    }
    root->color = BLACK;
}

statusEnum redBlackTree::insertNode(keyType key, recType *rec) {
    nodeType *current, *parent, *x;

   /***********************************************
    *  allocate node for data and insert in tree  *
    ***********************************************/

    /* find future parent */
    current = root;
    parent = 0;
    while (current != NIL) {
        if (compEQ(key, current->key)) 
            return STATUS_DUPLICATE_KEY;
        parent = current;
        current = compLT(key, current->key) ?
            current->left : current->right;
    }

    /* setup new node */
    if ((x = new nodeType) == 0)
        return STATUS_MEM_EXHAUSTED;
    x->parent = parent;
    x->left = NIL;
    x->right = NIL; 
    x->color = RED;
    x->key = key;
    x->rec = *rec;    

    /* insert node in tree */
    if(parent) {
        if(compLT(key, parent->key)) {
            parent->left = x;
            /* update maximum in the left subtree */
            parent->rec.maxl = x->rec.v; 
	    parent->rec.maxli = x->rec.i; 
        }
        else {
            parent->right = x;
            /* update maximum in the right subtree */
            parent->rec.maxr = x->rec.v;
	    parent->rec.maxri = x->rec.i; 
        }
    } else {
        root = x;
    }
    // to guarantee everything is ok before balancing
    updatePathMax(x);
    
    insertFixup(x);

    /* It is now enough to update all values from x to the root.
       The nodes that got wrong values during rebalancing,
       will be corrected by this step, since they are
       are on the way from x to the root */
    updatePathMax(x);

    lastFind = NULL;

    return STATUS_OK;
}

void redBlackTree::deleteFixup(nodeType *x) {


   /*************************************
    *  maintain Red-Black tree balance  *
    *  after deleting node x            *
    *************************************/
    /* No changes needed here for maintaining the subtree maxima,
       except in rotation procedures */
    while (x != root && x->color == BLACK) {
        if (x == x->parent->left) {
            nodeType *w = x->parent->right;
            if (w->color == RED) {
                w->color = BLACK;
                x->parent->color = RED;
                rotateLeft (x->parent);
                w = x->parent->right;
            }
            if (w->left->color == BLACK && w->right->color == BLACK) {
                w->color = RED;
                x = x->parent;
            } else {
                if (w->right->color == BLACK) {
                    w->left->color = BLACK;
                    w->color = RED;
                    rotateRight (w);
                    w = x->parent->right;
                }
                w->color = x->parent->color;
                x->parent->color = BLACK;
                w->right->color = BLACK;
                rotateLeft (x->parent);
                x = root;
            }
        } else {
            nodeType *w = x->parent->left;
            if (w->color == RED) {
                w->color = BLACK;
                x->parent->color = RED;
                rotateRight (x->parent);
                w = x->parent->left;
            }
            if (w->right->color == BLACK && w->left->color == BLACK) {
                w->color = RED;
                x = x->parent;
            } else {
                if (w->left->color == BLACK) {
                    w->right->color = BLACK;
                    w->color = RED;
                    rotateLeft (w);
                    w = x->parent->left;
                }
                w->color = x->parent->color;
                x->parent->color = BLACK;
                w->left->color = BLACK;
                rotateRight (x->parent);
                x = root;
            }
        }
    }
    x->color = BLACK;
}

statusEnum redBlackTree::deleteNode(keyType key) {
    nodeType *x, *y, *z;

   /*****************************
    *  delete node z from tree  *
    *****************************/

    /* find node in tree */
    if (lastFind && compEQ(lastFind->key, key))
        /* if we just found node, use pointer */
        z = lastFind;
    else {
        z = root;
        while(z != NIL) {
            if(compEQ(key, z->key)) 
                break;
            else
                z = compLT(key, z->key) ? z->left : z->right;
        }
        if (z == NIL) return STATUS_KEY_NOT_FOUND;
    }

    if (z->left == NIL || z->right == NIL) {
        /* y has a NIL node as a child */
        y = z;
    } else {
        /* find tree successor with a NIL node as a child */
        y = z->right;
        while (y->left != NIL) y = y->left;
    }

    /* x is y's only child */
    if (y->left != NIL)
        x = y->left;
    else
        x = y->right;

    /* remove y from the parent chain */
    x->parent = y->parent;
    if (y->parent)
        if (y == y->parent->left) 
            y->parent->left = x;
        else
            y->parent->right = x;
    else
        root = x;

    if (y != z) {
        z->key = y->key;
        z->rec.v = y->rec.v;
	   z->rec.i = y->rec.i;
	//updateParentMax(z);    
    }
    // to guarantee everything is ok before balancing
    updatePathMax(x);

    if (y->color == BLACK)
        deleteFixup(x);

    /* It is now enough to update all values from x to the root.
       The nodes that got wrong values during rebalancing,
       will be corrected by this step, since they are
       are on the way from x to the root */
    //cout << x->rec.v << " " << x->rec.maxl << " " << x->rec.maxr << endl;
    updatePathMax(x);

    delete y;
    lastFind = NULL;

    return STATUS_OK;
}

statusEnum redBlackTree::find(keyType key, recType *rec) {

   /*******************************
    *  find node containing data  *
    *******************************/

    nodeType *current = root;
    while(current != NIL) {
        if(compEQ(key, current->key)) {
            *rec = current->rec;
            lastFind = current;
            return STATUS_OK;
        } else {
            current = compLT (key, current->key) ?
                current->left : current->right;
        }
    }
    return STATUS_KEY_NOT_FOUND;
}


pair<int,int> redBlackTree::updateMaximaRec(nodeType *current) {

    /*******************************
     *  update maxima in subtrees  *
     *******************************/
    pair<int,int> result;
    if (current->left != NIL) {
       result = updateMaximaRec(current->left);
       current->rec.maxl = result.first;
       current->rec.maxli = result.second;
       
    }
    if (current->right != NIL) {
        result = updateMaximaRec(current->right);
	current->rec.maxr = result.first;
	current->rec.maxri = result.second;
    }
    
    /* this would print the tree in the postorder
    if (!current->parent)
       printf("root(%d,%d,%d,%d)",current->key.i,current->rec.v,current->rec.maxl,current->rec.maxr);
    else if (current == current->parent->left)
       printf("lc(%d,%d,%d,%d)",current->key.i,current->rec.v,current->rec.maxl,current->rec.maxr);
    else if (current == current->parent->right)
       printf("rc(%d,%d,%d,%d)",current->key.i,current->rec.v,current->rec.maxl,current->rec.maxr);
    fflush(stdout);
    */
    
    result.first = max3(current->rec.v,current->rec.maxl,current->rec.maxr); 
    if (result.first == current->rec.v)
       result.second = current->rec.i;
    else if (result.first == current->rec.maxl)
       result.second = current->rec.maxli;
    else 
       result.second = current->rec.maxri;      
    return result;
}

void redBlackTree::updateMaximaAll() {
   updateMaximaRec(root);
}

recType redBlackTree::rangeMax(keyType l, keyType r) {

   /*******************************
    *  find maximum from a range  *
    *******************************/

    int max = INT_MIN;
    recType maxRec;
    maxRec.v = max;
    maxRec.i = max;
    nodeType *limit;
    
    /* find the split-node first */
    nodeType *current = root;
    while(current != NIL && (compLEQ(r,current->key) || compGT(l,current->key))) {
        //cout << current->key.i << ":" << max << endl;
        if (compLEQ(l,current->key) && compGEQ(r,current->key)){ 
	    if(current->rec.v > max){
		maxRec = current->rec;
	    	max = max2(max,current->rec.v);
	    }
	}
        if (compLEQ(r, current->key))
            current = current->left;
        else 
            current = current->right;
    }
    if (current == NIL){
       return maxRec; /* no split-node found ==> no points inside the query range */
    }
      
    /* report maximum in path to $l$ */
    limit = current->left;
    max = max2(max,current->rec.v);
    maxRec = current->rec;
    while(limit != NIL) {
       //cout << limit->key.i << ":" << max << "," << maxRec.v << endl; 
          
       if (compLEQ(l, limit->key)) {
           max = max3(max,limit->rec.maxr,limit->rec.v);
	   if ( max == limit->rec.v ) {
	     maxRec = limit->rec;
	   }
	   else if (max==limit->rec.maxr) {
	      maxRec.v = max;
	      maxRec.i = limit->rec.maxri;
	   }
           limit = limit->left;
       }
       else  
          limit = limit->right;
    }
    /* report maximum in path to $r$ */
    limit = current->right;
    while(limit != NIL) {
       //cout << limit->key.i << ":" << max << "," << maxRec.v << endl; 
       if (compGEQ(r, limit->key)) {
           max = max3(max,limit->rec.maxl,limit->rec.v);
	   if ( max == limit->rec.v ) {
	     maxRec = limit->rec;
	   }
	   else if (max==limit->rec.maxl) {
	     maxRec.v = max;
	     maxRec.i = limit->rec.maxli;
	   }	      
           limit = limit->right;
       }
       else  
          limit = limit->left;
    }
    //cout << max << "," << maxRec.v << endl; 

    return maxRec;
}

void redBlackTree::makeEmpty()
{
   if (root!=NIL)
      makeEmptyRec(root);
   root = NIL;
}

void redBlackTree::makeEmptyRec(nodeType *current)
{
   if (current->left != NIL) makeEmptyRec(current->left);
   if (current->right != NIL) makeEmptyRec(current->right);
   delete current;    
}

/*
int main(int argc, char **argv) {
    int i;
    recType rec;
    keyType key;
    statusEnum status;
    keyType left, right;
    int keys[10000]; 
    list<int> values[10000]; 
    time_t t;
    redBlackTree *T = new redBlackTree();

    left.i = atoi(argv[1]);
    right.i = atoi(argv[2]);
    left.j = INT_MIN;
    right.j = INT_MAX;

    srand((unsigned)time(&t));
    int maxval = INT_MIN;
    for(i=0; i<10000; i++) {
       rec.v = rand()%10000; 
       key.i = i;  key.j = i; rec.maxl = INT_MIN; rec.maxr = INT_MIN;
       rec.maxli = i; rec.maxri = i;
       rec.i = key.j;        
       maxval = max2(maxval, rec.v);
       status = T->insertNode(key,&rec);
       keys[i] = key.i;
       values[rec.v].push_back(i);
    }
    printf("Result should be %d\n",maxval);
    printf("Maximum from range [%d,%d] is %d\n",left.i,right.i,T->rangeMax(left,right).v);
    fflush(stdout);
    T->updateMaximaAll();
    printf("Maximum from range [%d,%d] is %d\n",left.i,right.i,T->rangeMax(left,right).v);
    fflush(stdout);    
    for(i=0; i<1000; i++) {
       if (T->rangeMax(left,right).v==INT_MIN) break;
       key.j = T->rangeMax(left,right).i; 
       //key.j = i; 
       //cout <<       T->rangeMax(left,right).v << endl; 
       //cout << key.j << endl;
       fflush(stdout);         
       key.i = keys[key.j]; 
       printf("%d %d\n", T->rangeMax(left,right).v, key.i);
       status = T->deleteNode(key);
       if (status == STATUS_KEY_NOT_FOUND)
          printf("not found\n");
       else	  
          printf("found\n");
    }
    if (T->rangeMax(left,right).v>INT_MIN) 
       printf("Maximum from range [%d,%d] is %d\n",left.i,right.i,T->rangeMax(left,right).v);
    else
       printf("No points in range [%d,%d]\n",left.i,right.i);
    fflush(stdout);
    T->updateMaximaAll();
    if (T->rangeMax(left,right).v>INT_MIN) 
       printf("Maximum from range [%d,%d] is %d\n",left.i,right.i,T->rangeMax(left,right).v);
    else
       printf("No points in range [%d,%d]\n",left.i,right.i);
    delete T;  
    return 0;
}
*/
