// editscript.hpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
//
// Additions / changes by Peter Husemann
//

#ifndef _SWIFT_EDITSCRIPT_
#define _SWIFT_EDITSCRIPT_

#include <deque>
#include <iostream>
#include <iterator>

namespace swift {


class EditOperation
{
    public:
    enum Type {
		Substitution = 0,
		Insertion = 1,
		Deletion = 2,
		Decline = 3,
		Undefined
    };

    EditOperation(Type type, unsigned int length);

    Type type() const;
    unsigned int length() const;
    void setLength(unsigned int length);

    private:
    typedef unsigned int Operation;


// use one integer to store length and type of an operation
// first two bit of an integer are used for the type
// and the rest is used for the length
    enum Shifts {
        TypeShift = sizeof(Operation) * 8 - 2
    };

    Operation encodeOperation(Type type, unsigned int length);

    static const Operation TypeMask;
    static const Operation LengthMask;

    Operation operation_;
};


/*!
    Constructs an edit operation of type \a type and length \a length.
*/
inline
EditOperation::EditOperation(Type type, unsigned int length) :
	operation_(encodeOperation(type, length))
{
}


/*!
    Encodes an operation of type \a type and length \a length.
*/
inline
EditOperation::Operation EditOperation::encodeOperation(
    Type type,
    unsigned int length
)
{
// Operation is a typedef for Integer.
// type is an enum which is casted to integer and shifted 30 (depending on 
// the architecture, here 32 bit) positions to the left. 
// => the type is stored in the first two bits.
// the length is stored in the last 30 bits.
	return (static_cast<Operation>(type) << TypeShift) |
           (length & LengthMask);
}


/*!
    Returns the type of the edit operation.
*/
inline
EditOperation::Type EditOperation::type() const
{
	return static_cast<Type>((operation_ & TypeMask) >> TypeShift);
}


/*!
    Returns the length of the edit operation.
*/
inline
unsigned int EditOperation::length() const
{
	return operation_ & LengthMask;
}


/*!
    Sets the length of the edit operation to \a length.
*/
inline
void EditOperation::setLength(unsigned int length)
{
	operation_ = (operation_ & TypeMask) | (length & LengthMask);
}


class EditScript : public std::deque<EditOperation>
{
    public:
    void append_back(EditOperation const & operation);
    void append_front(EditOperation const & operation);
     friend std::ostream& operator<< (std::ostream&, const EditScript&);
    std::ostream& toString(std::ostream &) const;
};

 inline std::ostream & operator<<(std::ostream & os, const EditScript& editscript)
 {
   editscript.toString(os);
   return os;
 }



/*!
    Appends \a operation to back of the edit script. The operations in the
    edit script are run length packed and the number of operations in the
    edit script will therefore only change if the type of the appended
    operation differs from the last operation.

    Overflow of the run length counter is not checked.
    Overflow at sizeof(int)-2 bit,

    \sa append_front()
*/
inline
void EditScript::append_back(EditOperation const & operation)
{
    if (!empty()) {
        EditOperation & last(back());
        if (last.type() == operation.type()) {
            last.setLength(last.length() + operation.length());
            return;
        }
    }
    push_back(operation);
}


/*!
    Appends \a operation to front of the edit script. The operations in the
    edit script are run length packed and the number of operations in the
    edit script will therefore only change if the type of the appended
    operation differs from the first operation.

    Overflow of the run length counter is not checked.

    \sa append_back
*/
inline
void EditScript::append_front(EditOperation const & operation)
{
    if (!empty()) {
        EditOperation & first(front());
        if (first.type() == operation.type()) {
            first.setLength(first.length() + operation.length());
            return;
        }
    }
    push_front(operation);
}




} // namespace swift

#endif

