// gappedxdropaligner.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

#include "gappedxdropaligner.hpp"
#include <algorithm>
#include <iostream>

namespace swift {


const int GappedXDropAligner::NegativeInfinity = (~(0)-1)/2 - 1;
const unsigned int GappedXDropAligner::ChunkSize = 2097152;


/**
 * Default constructor
 */
GappedXDropAligner::GappedXDropAligner() :
    dpValues_(0),
    dpValuesCapacity_(0),
	stateTable_(0),
	stateTableCapacity_(0)
{
}


/**
 * Destructor
 */
GappedXDropAligner::~GappedXDropAligner()
{
    delete [] dpValues_;
    
	delete [] stateTable_;
	
	if (stateBlocks_.size()) {
		std::deque<StateBlock*>::iterator current(stateBlocks_.begin());
		std::deque<StateBlock*>::iterator end(stateBlocks_.end());
		
		while (current != end)
			delete *current++;
	}
}


/**
 * Returns the score of the gapped X-drop alignment.
 */
void GappedXDropAligner::getAlignment(
	Parameters const & params,
	Result & result
)
{
	
	// Obtain score for left part of alignment if requested and possible.
	int leftScore(0);
	result.queryBegin = params.queryBegin;
	result.subjectBegin = params.subjectBegin;
	
	if ((params.alignDirection == AlignLeftRight ||
		 params.alignDirection == AlignLeft) &&
		(params.queryBegin != 0 && params.subjectBegin != 0)) {

		std::pair<int, int> alignmentLengths(0, 0);
	
		leftScore =
			makeAlignment(
				params.query,
				params.subject,
				params.queryBegin,
				params.subjectBegin,
				params,
				AlignLeft,
				alignmentLengths
            );

        result.queryBegin -= alignmentLengths.first;
        result.subjectBegin -= alignmentLengths.second;
	}

	// Obtain score for middle part of alignment if ungappedLength > 0.
    int middleScore(0);
	unsigned char const * qp(params.query + params.queryBegin);
	unsigned char const * sp(params.subject + params.subjectBegin);
	
	for (int i(0); i < params.ungappedLength; ++i) {
		middleScore += params.matrix[*qp][*sp];
		++qp;
		++sp;
    }

	// Obtain score for right part of alignment if requested and possible.
    int rightScore(0);
	result.queryEnd = params.queryBegin + params.ungappedLength - 1;
	result.subjectEnd = params.subjectBegin + params.ungappedLength - 1;

	if ((params.alignDirection == AlignLeftRight ||
		 params.alignDirection == AlignRight) &&
		(params.queryBegin + params.ungappedLength < params.queryLength &&
		 params.subjectBegin + params.ungappedLength < params.subjectLength)) {
			
		std::pair<int, int> alignmentLengths(0, 0);
			
		rightScore =
			makeAlignment(
				params.query + params.queryBegin + params.ungappedLength - 1,
				params.subject + params.subjectBegin + params.ungappedLength - 1,
				params.queryLength - params.queryBegin - params.ungappedLength,
				params.subjectLength - params.subjectBegin - params.ungappedLength,
				params,
				AlignRight,
				alignmentLengths
            );

        result.queryEnd += alignmentLengths.first;
        result.subjectEnd += alignmentLengths.second;
	}

    result.score = leftScore + middleScore + rightScore;
}


/**
 * Helper method for obtaining the score of the gapped x-drop alignment
 * in the given direction from the starting point.
 */
int GappedXDropAligner::makeAlignment(
	unsigned char const * query,
	unsigned char const * subject,
	int queryLength,
	int subjectLength,
	Parameters const & params,
	AlignmentDirection direction,
	std::pair<int, int> & alignmentLengths
)
{
	// Sanity check
	if (!query || !subject || subjectLength <= 0 || queryLength <= 0) {
		alignmentLengths.first = alignmentLengths.second;
		return 0;
	}

	int bestScore(0);
	int X(params.xDrop);
	int m(params.gapOpen + params.gapExtend);
	if (X < m)
		X = m;
	
	int subjectIncrement(direction == AlignLeft ? -1 : 1);
  
	alignmentLengths.first = alignmentLengths.second = 0;

	//
	// Reserve space for dynamic programming table
	//
    if (dpValuesCapacity_ < subjectLength + 2) {
        if (subjectLength + 2 < 50000) {
            setValueNodeCapacity(subjectLength + 2);
        } else {
            setValueNodeCapacity(50000);
        }
    }
            
        
	dpValues_[0].CC = 0;
	dpValues_[0].FF = -m - params.declinePenalty; 
	dpValues_[0].DD = -m;
	
	int c(-m);
	int i(0);
	for (i = 1; i <= subjectLength; ++i) {
		if (c < -X)
			break;

		dpValues_[i].CC = c;
		dpValues_[i].DD = c - m; 
		dpValues_[i].FF = c - m - params.declinePenalty;
		c -= params.gapExtend;
	}

	//
	// Calculate gapped drop-x alignment
	//
	int j(i);
	for (int j_r(1), tt(0); j_r <= queryLength; ++j_r) {
		
		int* wa(
			direction == AlignLeft ?
			params.matrix[query[queryLength - j_r]] :
			params.matrix[query[j_r]]);
		
		unsigned char const * subjectPtr(
			direction == AlignLeft ?
			&subject[subjectLength - tt] : &subject[tt]);
		
		int e(NegativeInfinity);
		int c(NegativeInfinity);
		int f(NegativeInfinity);
		
		unsigned int p(tt);
		int cb(tt);
	    for (i = tt; i < j; ++i) {
 			ValueNode& dp(dpValues_[p]);
        	int nextC(dp.CC + wa[*(subjectPtr += subjectIncrement)]);
    	    int nextF(dp.FF);
	        int d(dp.DD);

			if (c < f)
				c = f;
			
			if (f > d) {
				d = f;
			} else {
				if (c < d)
					c = d;
			}
			
			if (f > e) {
				e = f;
			} else {
				if (c < e)
					c = e;
			}
			
			if (bestScore - c > X) {
				if (tt == i) {
					++tt;
				} else {
					dp.CC = NegativeInfinity;
				}
			} else {
				cb = i;
				if (c > bestScore) {
					bestScore = c;
					alignmentLengths.first = j_r;
					alignmentLengths.second = i;
				}
				
				if ((c -= m) > (d -= params.gapExtend)) {
					dp.DD = c; 
				} else {
					dp.DD = d;
				}
				
				if (c > (e -= params.gapExtend))
					e = c;

				c += m; 
				if (f < c - params.gapOpen) { 
					dp.FF = c - params.gapOpen - params.declinePenalty; 
				} else {
					dp.FF = f - params.declinePenalty;
				}
				dp.CC = c;
			}
			
			c = nextC;
			f = nextF;
			++p;
		}
		
		if (tt == j)
			break;

		if (cb < j - 1)	{
			j = cb + 1;
		} else {
			while (e >= bestScore - X && j <= subjectLength) {
				dpValues_[j].CC = e;
				dpValues_[j].DD = e - m;
				dpValues_[j].FF = e - params.gapOpen - params.declinePenalty;
				e -= params.gapExtend;
	    		++j;
			}
		}

		if (j <= subjectLength) {
			dpValues_[j].DD =
				dpValues_[j].CC =
				dpValues_[j].FF = NegativeInfinity;
			++j;
		}
	}

	return bestScore;
}


/**
 * Returns the score and the corresponding edit script for the gapped
 * X-drop alignment.
 */
void GappedXDropAligner::getAlignmentWithEditScript(
	Parameters const & params,
	Result & result
)
{
	// Create new edit script or append?
	if (!result.editScript)
		result.editScript = new EditScript();

	// Obtain score for left part of alignment if requested and possible.
	int leftScore(0);
	result.queryBegin = params.queryBegin;
	result.subjectBegin = params.subjectBegin;
	
	if ((params.alignDirection == AlignLeftRight ||
		 params.alignDirection == AlignLeft) &&
		(params.queryBegin != 0 && params.subjectBegin != 0)) {
	
		std::pair<int, int> alignmentLengths(0, 0);
		
		leftScore =
			makeAlignmentWithEditScript(
				params.query,
				params.subject,
				params.queryBegin,
				params.subjectBegin,
				params,
				AlignLeft,
				alignmentLengths,
				*(result.editScript)
            );

        result.queryBegin -= alignmentLengths.first;
        result.subjectBegin -= alignmentLengths.second;
	}

	// Obtain score for middle part of alignment if ungappedLength > 0.
    int middleScore(0);
	unsigned char const * qp(params.query + params.queryBegin);
	unsigned char const * sp(params.subject + params.subjectBegin);
	
	for (int i(0); i < params.ungappedLength; ++i) {
		middleScore += params.matrix[*qp][*sp];
		++qp;
		++sp;
    }

	if (params.ungappedLength > 0)
		result.editScript->push_back(
			EditOperation(EditOperation::Substitution, params.ungappedLength));

	// Obtain score for right part of alignment if requested and possible.
    int rightScore(0);
	result.queryEnd = params.queryBegin + params.ungappedLength - 1;
	result.subjectEnd = params.subjectBegin + params.ungappedLength - 1;

	if ((params.alignDirection == AlignLeftRight ||
		 params.alignDirection == AlignRight) &&
		(params.queryBegin + params.ungappedLength < params.queryLength &&
		 params.subjectBegin + params.ungappedLength < params.subjectLength)) {
	
		std::pair<int, int> alignmentLengths(0, 0);
		
		rightScore =
			makeAlignmentWithEditScript(
				params.query + params.queryBegin + params.ungappedLength - 1,
				params.subject + params.subjectBegin + params.ungappedLength - 1,
				params.queryLength - params.queryBegin - params.ungappedLength,
				params.subjectLength - params.subjectBegin - params.ungappedLength,
				params,
				AlignRight,
				alignmentLengths,
				*(result.editScript)
            );

		result.queryEnd += alignmentLengths.first;
		result.subjectEnd += alignmentLengths.second;
	}

    result.score = leftScore + middleScore + rightScore;
}


/**
 * Helper method for obtaining the score and corresponding edit script
 * of the gapped x-drop alignment in the given direction from the
 * starting point.
 */
int GappedXDropAligner::makeAlignmentWithEditScript(
	unsigned char const * query,
	unsigned char const * subject,
	int queryLength,
	int subjectLength,
	Parameters const & params,
	AlignmentDirection direction,
	std::pair<int, int>& alignmentLengths,
	EditScript& editScript
)
{
	// Sanity check
	if (!query || !subject || subjectLength <= 0 || queryLength <= 0) {
		alignmentLengths.first = alignmentLengths.second;
		return 0;
	}
	
	int bestScore(0);
	int X(params.xDrop);
	int m(params.gapOpen + params.gapExtend);
	if (X < m)
		X = m;
	
	int subjectIncrement(direction == AlignLeft ? -1 : 1);
  
	alignmentLengths.first = alignmentLengths.second = 0;

	//
	// Reserve space for dynamic programming table
	//
    if (dpValuesCapacity_ < subjectLength + 2)
        setValueNodeCapacity(subjectLength + 2);
	
	//
	// Reserve space for traceback states
	//
	if (stateTableCapacity_ < queryLength + 1) {
		delete [] stateTable_;
		stateTable_ = new unsigned char*[queryLength + 1];
		stateTableCapacity_ = queryLength + 1;
	}

	// Clear state blocks
	clearStateBlocks();

	//
	// Initialize DP and traceback state tables
	//
	StateBlock* stateBlock(0);
	
	if (params.gapExtend > 0)
		stateBlock = obtainStateBlock(X/params.gapExtend + 5);
	else
		stateBlock = obtainStateBlock(subjectLength + 3);

	stateTable_[0] = stateBlock->states;

	dpValues_[0].CC = 0;
	dpValues_[0].FF = -m - params.declinePenalty;
	dpValues_[0].DD = -m;
	
	int c(-m);
	int i(0);
	unsigned char* stp(stateTable_[0]);
	for (i = 1; i <= subjectLength; ++i) {
		if (c < -X)
			break;

		dpValues_[i].CC = c;
		dpValues_[i].DD = c - m; 
		dpValues_[i].FF = c - m - params.declinePenalty;
		c -= params.gapExtend;
		stp[i] = 1;
	}
	stateBlock->usedStates = i + 1;
	
	//
	// Calculate gapped drop-x alignment
	//
	int j(i);
	for (int j_r(1), tt(0); j_r <= queryLength; ++j_r) {
		
		if (params.gapExtend > 0)
			stateBlock = obtainStateBlock(j - tt + 5 + X/params.gapExtend);
	    else
			stateBlock = obtainStateBlock(subjectLength - tt + 3);

		stateTable_[j_r] =
			stateBlock->states + stateBlock->usedStates + 1 - tt + 1;
							
		stp = stateTable_[j_r];
		int tt_start(tt); 

		int* wa(
			direction == AlignLeft ?
			params.matrix[query[queryLength - j_r]] :
			params.matrix[query[j_r]]);
		
		unsigned char const * subjectPtr(
			direction == AlignLeft ?
			&subject[subjectLength - tt] : &subject[tt]);
		
		int e(NegativeInfinity);
		int c(NegativeInfinity);
		int f(NegativeInfinity);
		
		unsigned int p(tt);
		int cb(tt);
	    for (i = tt; i < j; ++i) {
 			ValueNode& dp(dpValues_[p]);
        	int nextC(dp.CC + wa[*(subjectPtr += subjectIncrement)]);
    	    int nextF(dp.FF);
	        int d(dp.DD);

			unsigned char std(0);
			unsigned char ste(0);	
			unsigned char st(0);
			
			if (c < f) {
				c = f;
				st = 3;
			}
			
			if (f > d) {
				d = f;
				std = 60;
			} else {
				std = 30;
				if (c < d) {
					c = d;
					st = 2;
				}
			}
			
			if (f > e) {
				e = f;
				ste = 20;
			} else {
				ste = 10;
				if (c < e) {
					c = e;
					st = 1;
				}
			}
			
			if (bestScore - c > X) {
				if (tt == i) {
					++tt;
				} else {
					dp.CC = NegativeInfinity;
				}
			} else {
				cb = i;
				if (c > bestScore) {
					bestScore = c;
					alignmentLengths.first = j_r;
					alignmentLengths.second = i;
				}
				
				if ((c -= m) > (d -= params.gapExtend)) {
					dp.DD = c; 
				} else {
					dp.DD = d;
					st += std;
				}
				
				if (c > (e -= params.gapExtend)) {
					e = c;
				} else {
					st += ste;
				}

				c += m; 
				if (f < c - params.gapOpen) { 
					dp.FF = c - params.gapOpen - params.declinePenalty; 
				} else {
					dp.FF = f - params.declinePenalty;
					st += 5; 
				}
				dp.CC = c;
			}
			
			stp[i] = st;
			c = nextC;
			f = nextF;
			++p;
		}
		
		if (tt == j) {
			break;
		}
		
    	if (cb < j - 1) {
			j = cb + 1;
		} else {
			while (e >= bestScore - X && j <= subjectLength) {
				dpValues_[j].CC = e;
				dpValues_[j].DD = e - m;
				dpValues_[j].FF = e - params.gapOpen - params.declinePenalty;
				e -= params.gapExtend;
				stp[j] = 1;
	    		++j;
			}
		}

		if (j <= subjectLength) {
			dpValues_[j].DD = dpValues_[j].CC = dpValues_[j].FF = NegativeInfinity;
			++j;
		}
		
		stateBlock->usedStates += (std::max(i, j) - tt_start + 1);
	}

	//
	// Derive edit script from traceback
	//
	deriveEditScript(
		stateTable_,
		alignmentLengths.first,
		alignmentLengths.second,
		direction,
		editScript);
	
	return bestScore;
}


/**
 * Helper method for deriving the edit script from the state tables.
 */
void GappedXDropAligner::deriveEditScript(
	unsigned char** state,
	int queryPos,
	int subjectPos,
	AlignmentDirection direction,
	EditScript& editScript) const
{
	unsigned int lastCount(0);
	int lastOperation(0);
	
	int s(0);
	int t(0);
	int k(0);
	
	unsigned int startLength(editScript.size());
	
	while (queryPos > 0 || subjectPos > 0) {
		t = state[queryPos][subjectPos];
		k  = t % 5;
		if (s == 1)
			if ((t/10)%3 == 1)
				k = 1;
			else if ((t/10)%3 == 2)
				k = 3;
			
		if (s == 2)
			if ((t/30) == 1)
				k = 2;
			else if ((t/30) == 2)
				k = 3;
			
		if (s == 3 && ((t/5)%2) == 1)
			k = 3;
		
		if (k == 1)
			--subjectPos;
		else if (k == 2)
			--queryPos;
		else {
			--subjectPos;
			--queryPos;
		}
		s = k;
		
	
		if (lastOperation == k) {
			++lastCount;
		} else {
			if (lastCount > 0) {
				editScript.push_back(
					EditOperation(getEditBlock(lastOperation), lastCount));
			}
			lastCount = 1;
			lastOperation = k;
		}
	}
	
	if (lastCount > 0) {
		editScript.push_back(
			EditOperation(getEditBlock(lastOperation), lastCount));
	}
	
	// Ensure right direction of edit script
	if (direction == AlignRight && editScript.size() > startLength)
		reverse(editScript.begin() + startLength, editScript.end());
}


/**
 * Clear state blocks.
 */
void GappedXDropAligner::clearStateBlocks()
{
	std::deque<StateBlock*>::iterator current(stateBlocks_.begin());
	std::deque<StateBlock*>::iterator end(stateBlocks_.end());

	while (current != end)
		(*current++)->usedStates = 0;
}


/**
 * Obtain state block
 */
GappedXDropAligner::StateBlock*
GappedXDropAligner::obtainStateBlock(unsigned int minimumLength)
{
	unsigned int capacity(std::max(ChunkSize, minimumLength + minimumLength/3));
	
	minimumLength += minimumLength / 3; // Stimulate reuse of ends
	
	std::deque<StateBlock*>::iterator current(stateBlocks_.begin());
	std::deque<StateBlock*>::iterator end(stateBlocks_.end());
	
	while (current != end) {
		if (minimumLength < (*current)->capacity - (*current)->usedStates)
			return *current;
		else if ((*current)->usedStates == 0) {
			// Empty, but too small so resize
			(*current)->setCapacity(capacity);
			return *current;
		}
		++current;
	}

	StateBlock* newBlock = new StateBlock(capacity);
	stateBlocks_.push_back(newBlock);
	
	return newBlock;
}


} // namespace swift

