// gappedxdropaligner.hpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

#ifndef _SWIFT_GAPPEDXDROPALIGNER_
#define _SWIFT_GAPPEDXDROPALIGNER_

#include "assert.hpp"
#include "editscript.hpp"
#include <cstdlib>
#include <iostream>
#include <sstream>
#include <deque>
#include <vector>
#include <new>

namespace swift {

/*!
    Gapped X-drop alignment with affine gap costs.

    Reference:
    Zheng Zhang, Piotr Berman and Webb Miller. Alignments without
    low-scoring regions. Proceedings of the Second Annual International
    Conference on Computational Molecular Biology, 294--301, 1998. ACM Press.

    Conventions:

    AlignLeft:            AlignRight:            AlignLeftRight

       xxxx  |            S                           xxx |
      xxxxxxx|             \                           xxx|
       xxxxxx|              \                         ----S
      -------S               \-------                      \
              \              |xxxxx                         \-----
               \             |xxxx                          |xxxx
                \            | xxx                          | xxx

    where "S" denotes the starting point as passed to the alignment
    generating method. Note that in all cases, the ungapped diagonal piece
    may have a length of zero.
*/
class GappedXDropAligner
{

	/* --- Constants --- */
	private:
	static const int NegativeInfinity;
	static const unsigned int ChunkSize;

	/* --- Types --- */
	public:
	enum AlignmentDirection {
		AlignLeftRight,
		AlignLeft,
		AlignRight
	};

	private:
	typedef struct {
		int CC;
		int DD;
		int FF;
	} ValueNode;

	// State blocks
	typedef struct StateBlockImpl {
		unsigned int capacity;
		unsigned int usedStates;
		unsigned char* states;

		inline StateBlockImpl(unsigned int size);
		inline ~StateBlockImpl();
		inline void setCapacity(unsigned int newCapacity);
	} StateBlock;

	public:
	struct Parameters {
		unsigned char const * query;		/* Query sequence */
		unsigned char const * subject;	/* Subject sequence */
		int queryLength;		/* Length of query sequence */
		int subjectLength;		/* Length of subject sequence */
		int queryBegin;		/* Beginning position in query sequence */
		int subjectBegin;		/* Beginning position in subject sequence */
		AlignmentDirection alignDirection;
								/* Determines in which directions the gapped
								   x-drop alignment is performed, left and
								   right, left only, right only. */
		int ungappedLength;	/* Length of ungapped part of query
								   (from queryBegin) which must be included
								   in the alignment */
		int gapOpen;			/* Gap-opening cost */
		int gapExtend;			/* Gap-extension cost */
		int declinePenalty;	/* Decline to align penalty */
		int xDrop;				/* X dropoff threshold */
		int** matrix;			/* Substitution score matrix */
	};

	struct Result {
		int score;				    /* Score of alignment */
		int queryBegin;		    /* Beginning position of alignment in query */
		int queryEnd;			    /* Ending position of alignment in query */
		int subjectBegin;		    /* Beginning position of alignment in query */
		int subjectEnd;		    /* Ending position of alignment in query */
		EditScript * editScript;	/* Edit script of alignment */
	};

	/* --- Constructors / destructor --- */
	public:
	GappedXDropAligner();
	~GappedXDropAligner();

	/* --- Methods --- */
	public:

	void getAlignment(Parameters const & params, Result & result);
	void getAlignmentWithEditScript(Parameters const & params, Result & result);

	private:
	int makeAlignment(
		unsigned char const * query,
		unsigned char const * subject,
		int queryLength,
		int subjectLength,
		Parameters const & params,
		AlignmentDirection direction,
		std::pair<int, int> & alignmentLengths
    );

	int makeAlignmentWithEditScript(
		unsigned char const * query,
		unsigned char const * subject,
		int queryLength,
		int subjectLength,
		Parameters const & params,
		AlignmentDirection direction,
		std::pair<int, int> & alignmentLengths,
		EditScript& editScript
    );

	inline EditOperation::Type getEditBlock(int tracebackOpcode) const;

	void deriveEditScript(
		unsigned char** state,
		int queryPos,
		int subjectPos,
		AlignmentDirection direction,
		EditScript& editScript
    ) const;

	void clearStateBlocks();
	StateBlock* obtainStateBlock(unsigned int minimumLength);

    inline void setValueNodeCapacity(unsigned int newCapacity);

	/* --- Fields --- */
	private:
    ValueNode * dpValues_;
    unsigned int dpValuesCapacity_;

	unsigned char ** stateTable_;
	unsigned int stateTableCapacity_;
	std::deque<StateBlock *> stateBlocks_;
};



inline EditOperation::Type
GappedXDropAligner::getEditBlock(int tracebackOpcode) const
{
	if (tracebackOpcode == 0) {
		return EditOperation::Substitution;
	} else if (tracebackOpcode == 1) {
		return EditOperation::Insertion;
	} else if (tracebackOpcode == 2) {
		return EditOperation::Deletion;
	} else if (tracebackOpcode == 3) {
		return EditOperation::Decline;
	} else {
    std::cerr << "swift: Unknown opcode." << std::endl;
		exit(-1);
	}
}


/**
 * Constructs a new state block with the given capacity.
 */
inline GappedXDropAligner::StateBlockImpl::StateBlockImpl(
    unsigned int size
) :
	capacity(size),
	usedStates(0),
	states(new unsigned char[size])
{
	ASSERT(size, "Zero sized state block");
}


/**
 * Destructor
 */
inline GappedXDropAligner::StateBlockImpl::~StateBlockImpl()
{
	delete [] states;
}


/**
 * Change the capacity of the state block.
 *
 * The capacity change will force a reallocation of the internal
 * state array.
 */
inline void GappedXDropAligner::StateBlockImpl::setCapacity(
	unsigned int newCapacity)
{
	delete [] states;
	states = new unsigned char[newCapacity];
	capacity = newCapacity;
	usedStates = 0;
}


/**
 * Allocates the requested amount of value nodes.
 */
inline void GappedXDropAligner::setValueNodeCapacity(unsigned int newCapacity)
{
    delete [] dpValues_;

    if (!(dpValues_ = new (std::nothrow) ValueNode[(dpValuesCapacity_ = newCapacity)])) {
        std::cerr
            << "Allocation of " << (newCapacity * sizeof(ValueNode)) << " bytes failed "
            << "(GappedXDropAligner::setValueNodeCapacity)" << std::endl;
        std::cerr.flush();
        exit(-1); // Fail hard and fast
    }
}


} // namespace swift

#endif


