// gram.hpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
// 

#ifndef _SWIFT_GRAM_
#define _SWIFT_GRAM_

#include "masksensitivity.hpp"
#include "encoding.hpp"
#include <boost/cstdint.hpp>

namespace swift {

// --------------------------------------------------------------------------

class Gram
{
    public:
    typedef boost::uint32_t Value;

    Gram(unsigned int size, MaskSensitivity ms);
    ~Gram();

    unsigned int size() const;
    Value value() const;
    bool isValid() const;

    Value flags() const { return flags_; }

    void clear();
    void push_front();
    void push_back();
    void push_front(Value s);
    void push_back(Value s);

    template <typename InputIterator>
    void push_back(InputIterator first, InputIterator last);

    static Value maxValue(unsigned int size);

    private:
    static Value valueMask(unsigned int size);
    static Value flagMask(unsigned int size, MaskSensitivity ms);

    unsigned int size_;
    Value value_;
    Value flags_;
    Value valueMask_;
    Value flagMask_;
    Value frontShift_;
};

// --------------------------------------------------------------------------

/*!
    Returns the size of the gram, i.e. the number of letters.
*/
inline
unsigned int Gram::size() const
{
    return size_;
}


/*!
    Returns the numerical value of the gram.
*/
inline
Gram::Value Gram::value() const
{
    return value_;
}


/*!
    Returns true if the gram is valid. A gram is valid when the constiuent
    letters are taken from the alphabet over which the gram is defined.
*/
inline
bool Gram::isValid() const
{
    return !flags_;
}



inline
void Gram::push_front()
{
    push_front(Encoding::N);
}


inline
void Gram::push_back()
{
    push_back(Encoding::N);
}
    

inline
void Gram::push_front(Value s)
{
    value_ = (value_ >> 2) | ((s << frontShift_) & valueMask_);
    flags_ = (flags_ >> 2) | (((s >> 2) << frontShift_) & flagMask_);
}


inline
void Gram::push_back(Value s)
{
    value_ = ((value_ << 2) | (s & 0x3)) & valueMask_;
    flags_ = ((flags_ << 2) | (s >> 2)) & flagMask_;
}


template <typename InputIterator>
void Gram::push_back(InputIterator first, InputIterator last)
{
    while (first != last) {
        push_back(*first);
        ++first;
    }
}

// --------------------------------------------------------------------------

} // namespace swift

#endif

