// gramchainaligner.hpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
// 

#include "gramchainaligner.hpp"
#include "gramview.hpp"

// DEBUG showChain
#include <deque>
#include <string>
#include <sstream>
#include <climits>

namespace swift {


GramChainAligner::GramChainAligner() :
    gramMap_(GramPositionMap(2048,0.75)),
    queryMaskSensitivity_(MaskInsensitive),
    subjectMaskSensitivity_(MaskInsensitive)
{
}


GramChainAligner::~GramChainAligner()
{
}


void GramChainAligner::alignments(
    Parameters const & params,
    Result & result
)
{
	if (params.subjectLength < static_cast<int>(params.gramSize) ||
        params.queryLength < static_cast<int>(params.gramSize)) {
        return;
    }

	// Ensure that the upper left and lower right corners of dynamic
	// programming matrix are within the band.
	int lowerBound =
		std::min(
			std::max(-params.queryLength, params.lowerDiagonal),
			std::min(params.subjectLength - params.queryLength, 0)
        );

	int upperBound =
		std::max(
			std::min(params.subjectLength, params.upperDiagonal),
			std::max(params.subjectLength - params.queryLength, 0)
        );

	if (upperBound == lowerBound) {
        // FIXME 1090795183
        // Trivial case where there's only one diagonal,
        // so this should be treated specially.
        // Needs only one encoder for each sequence and trivial
        // chaining of the hits.
        std::cout << "Trivial case, one diagonal only" << std::endl;
        exit(1);
	}

    // Find chains
    findChains(
        params.query,
        params.subject,
        params.queryLength,
        params.subjectLength,
        lowerBound,
        upperBound,
        params.maxGap,
        params.minHits,
        params.gramSize
    );
    
    // Rescore chains
    alignmentsFromChains(params, result);

    // Clean up
    chainsFound_.clear();
}


void GramChainAligner::alignWithEditScript(
    Parameters const & params,
    Result & result
)
{
    std::cerr << "GramChainAligner::alignWithEditScript() not implemented"
              << std::endl;
    exit(1);
}


void GramChainAligner::findChains(
    char const *  query,
    char const *  subject,
    int          queryLength,
    int          subjectLength,
    int          lowerDiagonal,
    int          upperDiagonal,
    int          maxGap,
    unsigned int         minChainLength,
    unsigned int gramSize
)
{
    typedef GramView<unsigned char const *> MyGramView;
    typedef MyGramView::const_iterator GramIterator;

    // Clear and reserve space
    gramMap_.clear();
    linkPool_.clear();
    diagonalMap_.clear();
    chainsFound_.clear();
    positionPool_.reserve(upperDiagonal - lowerDiagonal + 2);

    // Define gram views on query and subject sequences
    MyGramView subjectView(
        reinterpret_cast<unsigned char const *>(subject),
        reinterpret_cast<unsigned char const *>(subject + subjectLength),
        gramSize,
        subjectMaskSensitivity_
    );

    MyGramView queryView(
        reinterpret_cast<unsigned char const *>(query),
        reinterpret_cast<unsigned char const *>(query + queryLength),
        gramSize,
        queryMaskSensitivity_
    );

    GramIterator bandBack(subjectView.begin());
    GramIterator bandFront(subjectView.begin());
    GramIterator queryCurrent(queryView.begin());
    GramIterator queryEnd(queryView.end());
    GramIterator subjectEnd(subjectView.end());

    // Initialize gram map with q-grams in first line of the band.
    int fs(0); // Subject position, front
    for ( ; fs < upperDiagonal; ++fs) {
        if ((*bandFront).isValid()) {
            Position * p(positionPool_.pop());
            p->position = fs;
            PositionList pl(p, p);
            std::pair<GramPosition *, bool> ins(
                gramMap_.insert(GramPosition((*bandFront).value(), pl))
            );
            if (!ins.second) {
                ins.first->second.last->next = p;
                ins.first->second.last = p;
                //ins.first->last->next = p;
                //ins.first->last = p;
            }
        }
        ++bandFront;
    }

    // Main loop
    int gy(0); // Query position
    for ( ; queryCurrent != queryEnd; ++queryCurrent, ++gy) {
       
        // Update band back
        if (static_cast<int>(gy) > -lowerDiagonal) {
#if 0 
        /* Fix changes >= to > */
        if (static_cast<int>(gy) >= -lowerDiagonal) {
#endif            
            if ((*bandBack).isValid()) {
                std::pair<GramPosition *, bool> ins(gramMap_.find((*bandBack).value()));
                //std::pair<PositionList *, bool> ins(gramMap_.find((*bandBack).value()));
                Position * t(ins.first->second.first);
                //Position * t(ins.first->first);
                ins.first->second.first = ins.first->second.first->next;
                //ins.first->first = ins.first->first->next;
                positionPool_.push(t);
                if (!ins.first->second.first)
                //if (!ins.first->first)
                    gramMap_.erase((*bandBack).value());
            }
            ++bandBack;
        }

        // Update band front
        if (bandFront != subjectEnd) {
            if ((*bandFront).isValid()) {
                Position * p(positionPool_.pop());
                p->position = fs;
                PositionList pl(p, p);
                std::pair<GramPosition *, bool> ins =
                    gramMap_.insert(GramPosition((*bandFront).value(), pl)
                );
                if (!ins.second) {
                    ins.first->second.last->next = p;
                    ins.first->second.last = p;
                    //ins.first->last->next = p;
                    //ins.first->last = p;
                }
            }
            ++bandFront;
            ++fs;
        }

        // Find q-hits for current q-gram            
        if (!(*queryCurrent).isValid())
            continue;
        //std::pair<PositionList *, bool> matches(
        std::pair<GramPosition *, bool> matches(
            gramMap_.find((*queryCurrent).value())
        );
        if (!matches.second)
            continue;

        // Loop over q-gram hits
        Position * p(matches.first->second.first);
        //Position * p(matches.first->first);
        for ( ; p; p = p->next) {

            // Find a fragment f such that:
            //  (i)  f << g
            //  (ii) max{score(f') - connect(f',g)} is maximized when f' = f.
            //
            // A fragment is a tuple (x,y,l) where x,y is the beginning
            // position and l is the length.
            //
            
            unsigned int gx(p->position);
            int  gd(gx - gy);
            int  fd(0);
            int  lowerd(gd - maxGap);  ///< Diagonal range query lower bound
            int  upperd(gd + maxGap);  ///< Diagonal range query upper bound

            Link * bestChain(0);        ///< Best chain to link to
            int   bestScore(0);        ///< Score of best chain
            int   bestDiagonal(gd);    ///< Diagonal of best chain
            unsigned int  bestHits(0);         ///< Number of q-hits in best (possibly cut) chain.
            
            DiagonalMap::iterator diag(diagonalMap_.lower_bound(lowerd));
            DiagonalMap::iterator dend(diagonalMap_.end());

            while (diag != dend && (fd = (*diag).first) <= upperd) {

                ASSERT(abs(gd - fd) <= maxGap, " Candidate link within range ");

                bool closer(abs(gd - fd) < abs(gd - bestDiagonal));
                LinkList * linkList((*diag).second);

                // Remove farthest-away fragment h from linklist,
                // if no longer linkable.
                {
                    // Invariant: LinkList is never empty.
                    Link * link((*(linkList->begin())).second);
                    ASSERT(gy >= link->y, "gy >= link->y");
                    
                    if (link->y + link->length + maxGap < gy) {
                        if (!link->refCount && link->hits >= minChainLength) {
                            chainsFound_.push_back(link);
                        }

                        linkList->erase(linkList->begin());
                        if (linkList->empty()) {
                            linkListPool_.push_back(linkList);
                            DiagonalMap::iterator t(diag++);
                            diagonalMap_.erase(t);
                            continue; // process remaining diagonals
                        }
                    }
                }

                // Find best chain candidate.
                {
                    if (gd == fd) {
                        // Candidates on same diagonal.
                        // Only the nearest needs to be considered.
                        Link * link((*(--(linkList->end()))).second);
                        
                        int dist(gy - link->y - link->length);
                        int score(link->score - abs(dist));
                        
                        if (dist <= maxGap && score >= bestScore) {
                            bestScore    = score;
                            bestChain    = link;
                            bestDiagonal = fd;
                            bestHits     = link->hits;
                        }
                    } else {
                        // Candidates on other diagonal.
                        int fybmax(gy - gramSize);   // f y-begin max
                        int fyemin(gy - maxGap - 1);   // f y-end min
                        int fyemax(gy - 1);            // f y-end max
                        unsigned int dd(abs(fd - gd));

                        ASSERT(dd <= maxGap, " Candidate link in range ");

                        if (fd < gd) {
                            fyemin += dd;
                        } else {
                            fybmax -= dd;
                            fyemax -= dd;
                        }
                        
                        LinkList::iterator litr(linkList->lower_bound(fybmax));
                        if (litr == linkList->end())
                            --litr; // safe as there's at least one element

                        if (((*litr).second)->y > fybmax) {
                            if (litr == linkList->begin()) {
                                // All candidates have start_y > fybmax -- continue loop
                                ++diag;
                                continue;
                            } else {
                                // At least one has y <= fybmax
                                --litr;
                            }
                        }

                        Link * link((*litr).second);
                        
                        ASSERT(link->y <= fybmax, "link->y <= fybmax");

                        int fye(link->y + link->length - 1);
                        int fxe(link->x + link->length - 1);

                        if (fye >= fyemin) {
                            if (fye > fyemax) { // find cut point
                                fxe -= fyemax - fye;
                                fye = fyemax;
                            }

                            int dist(std::max(gy - fye, static_cast<int>(gx - fxe)) - 1);

                            ASSERT(dist >= 0,      "dist >= 0");
                            ASSERT(dist <= maxGap, "dist <= maxGap");

                            int t(fye - link->y + 1 - link->length);
                            int score(link->score + t - dist);

                            if ((score > bestScore) ||
                                (score == bestScore && closer)) {
                                bestScore    = score;
                                bestChain    = link;
                                bestDiagonal = fd;
                                bestHits     = link->hits + t;
                            }
                        }
                    }
                }
                ++diag;
            }

            /// Insert the current link.
            {
                if (bestChain && bestDiagonal == gd &&
                    bestChain->y + bestChain->length > gy) {
                    
                    ASSERT(bestChain->y + bestChain->length - gramSize + 1 == gy, "overlap point");
                    
                    ++(bestChain->score);
                    ++(bestChain->length);
                    ++(bestChain->hits);
                } else {
                    LinkList * linkList(0);
                    std::pair<DiagonalMap::iterator, bool> ins(
                        diagonalMap_.insert(std::pair<int, LinkList*>(gd, 0)));

                    if (ins.second) {
                        if (!linkListPool_.empty()) {
                            (*ins.first).second = linkListPool_.back();
                            linkListPool_.pop_back();
                        } else {
                            (*ins.first).second = new LinkList(); // FIXME: allocate in pool?
                        }
                    }
                    linkList = (*ins.first).second;

                    Link * link(linkPool_.acquireLink());

                    link->x        = gx;
                    link->y        = gy;
                    link->score    = bestScore + gramSize;
                    link->refCount = 0;
                    link->length   = gramSize;
                    link->next     = bestChain;
                    link->hits     = bestHits + 1;

                    if (bestChain)
                        ++(bestChain->refCount);
                
                    linkList->insert(std::pair<unsigned int, Link *>(gy,link));
                }
            }
        }
    }

    // Collect remaining chains
    DiagonalMap::iterator ditr(diagonalMap_.begin());
    DiagonalMap::iterator dend(diagonalMap_.end());
    for ( ; ditr != dend; ++ditr) {
        LinkList::iterator litr((*ditr).second->begin());
        LinkList::iterator lend((*ditr).second->end());
        for ( ; litr != lend; ++litr) {
            Link * link((*litr).second);
            if (!link->refCount && link->hits >= minChainLength) {
                chainsFound_.push_back(link);
            }
        }    
        (*ditr).second->clear();

        linkListPool_.push_back((*ditr).second);
    }
}


/// Rescore chains into alignments.
void GramChainAligner::alignmentsFromChains(
    Parameters const & params,
    Result & result
)
{
    unsigned int clusterCount(0);
    Link *       link(0);
    Alignment *  a(0);
    EditScript * e(0);
   
    result.clear();
    result.reserve(chainsFound_.size());
    
    ChainVector::iterator chain(chainsFound_.begin());
    ChainVector::iterator chend(chainsFound_.end());

    for ( ; chain != chend; ++chain) {
        if (!(link = *chain))
            continue;

        // Process last chain link
        if (a) {
            e->clear();
        } else {
            a = new Alignment;
            e = new EditScript;
        }

        a->subjectEnd = link->x + link->length - 1;
        a->queryEnd   = link->y + link->length - 1;
        int chainScore(link->length * matchScore_);
        e->push_front(EditOperation(EditOperation::Substitution, link->length));
            
        // Process remaining links
        Link * plink(link);
        link = link->next;
        while (link) {
            int offset((plink->x - plink->y) - (link->x - link->y));

            if (!offset) {
                // Links on same diagonal
                rescoreOnDiagonal(
                    plink,
                    link,
                    params,
                    chainScore,
                    e
                );
            } else {
                // Links on different diagonals
                rescoreOffDiagonal(
                    plink,
                    link,
                    params,
                    chainScore,
                    e
                );
            }
         
            plink = link;
            link  = link->next;
        }
        
        a->queryBegin   = plink->y;
        a->subjectBegin = plink->x;
        a->score        = chainScore;
        a->editScript   = e;

        // As we are processing the chain in reverse order,
        // plink will now point at the first link in the chain.

        if (plink->clusterFlag) {
            plink->clusterFlag = 0;
            plink->clusterNum  = clusterCount;
            ++clusterCount;
            result.push_back(a);
            a = 0;
        } else {
            Alignment * & p(result[plink->clusterNum]);
            if (p->score < chainScore) {
                Alignment * q(p);
                p = a;
                a = q;
                e = a->editScript;
            }
        }
    }

    if (a) {
        delete a;
    }
}


/// Connect and rescore link to plink chain, where
/// link and plink are non-overlapping on same diagonal.
///
/// Entry: score holds the score of the chain
///        beginning in plink.
/// Exit:  score holds the score of the chain
///        beginning in link.
///
/// If editScript is not-null, the edit script is
/// continued.
void GramChainAligner::rescoreOnDiagonal(
    Link const *       plink,
    Link const *       link,
    Parameters const & params,
    int &             score,
    EditScript *       editScript
)
{
    ASSERT(plink->y - link->y - link->length > 0, "gap size > 0");

    char const * q(params.query + link->y + link->length - 1);
    char const * s(params.subject + link->x + link->length - 1);
    
    for (unsigned int gapSize(plink->y - link->y - link->length); gapSize; --gapSize) {
        score += scoreFunction_[*(++q)][*(++s)];
    }
    score += link->length * matchScore_;        

    if (editScript) {
        editScript->append_front(
            EditOperation(EditOperation::Substitution, plink->y - link->y)
        );
    }
}


/// Connect and rescore link to plink chain, where
/// link and plink are on different diagonals and
/// possibly overlapping in either sequence.
///
/// Entry: score holds the score of the chain
///        beginning in plink.
/// Exit:  score holds the score of the chain
///        beginning in link.
///
/// If editScript is not-null, the edit script is
/// continued.
void GramChainAligner::rescoreOffDiagonal(
    Link const *       plink,
    Link const *       link,
    Parameters const & params,
    int &             score,
    EditScript *       editScript
)
{
    static int const NegativeInfinity = INT_MIN;
    static std::vector<int> scoresRL;
    static std::vector<int> scoresLR;
    
    unsigned int linklen(link->length);
    unsigned int lengthRL(0);
    unsigned int lengthLR(0);

    unsigned int pxb(plink->x);
    unsigned int pyb(plink->y);
    unsigned int lxe(link->x + link->length - 1);
    unsigned int lye(link->y + link->length - 1);

    int  offset((pxb - pyb) - (lxe - lye));
    int  extlen(
        std::min(static_cast<int>(pxb - lxe), static_cast<int>(pyb - lye))
    );

    if (extlen <= 1) {
        // Determine cut point
        linklen = offset < 0 ? pxb - link->x : pyb - link->y;
        ASSERT(linklen >= params.gramSize, "linklen >= gramLength_");
    } else {
        unsigned int myextlen(extlen);   // Unsignedness

        if (scoresRL.size() < myextlen + 1) {
            scoresRL.resize(myextlen << 1, 0);
            scoresLR.resize(myextlen << 1, 0);
        }

        char const * q(params.query);
        char const * s(params.subject);

        // Ungapped extension
        scoresLR[0] = scoresRL[myextlen] = 0;
        for (unsigned int i(1); i <= myextlen; ++i) {
            scoresLR[i]            = scoresLR[i - 1] + 
                                     scoreFunction_[q[++lye]][s[++lxe]];
            scoresRL[myextlen - i] = scoresRL[myextlen - i + 1] +
                                     scoreFunction_[q[--pyb]][s[--pxb]];
        }

        // Find best position of gap
        int  extScore(NegativeInfinity);
        for (unsigned int i(0); i < myextlen; ++i) {
            int t(scoresLR[i] + scoresRL[i + 1]);
            if (t > extScore) {
                extScore = t;
                lengthLR = i;
            }
        }
        lengthRL = myextlen - lengthLR - 1;
        score += extScore;
    }
    score += linklen * matchScore_ - 
             scoreFunction_.gapOpenCost() - 
             scoreFunction_.gapExtendCost() * abs(offset);

    if (editScript) {
        editScript->append_front(EditOperation(EditOperation::Substitution, lengthRL));
        
        if (offset < 0)
            editScript->push_front(EditOperation(EditOperation::Deletion, -offset));
        else
            editScript->push_front(EditOperation(EditOperation::Insertion, offset));
            
        editScript->push_front(EditOperation(EditOperation::Substitution, lengthLR + linklen));
    }
}


/*!
    Sets the score function.
*/
void GramChainAligner::setScoreFunction(ScoreFunction const & scoreFunction)
{
    scoreFunction_ = scoreFunction;
}


void GramChainAligner::setMatchScore(int matchScore)
{
    matchScore_ = matchScore;
}


/*!
    Sets the query mask sensitivity to \a ms.
*/
void GramChainAligner::setQueryMaskSensitivity(MaskSensitivity ms)
{
    queryMaskSensitivity_ = ms;
}


/*!
    Returns the query mask sensitivity.
*/
MaskSensitivity GramChainAligner::queryMaskSensitivity() const
{
    return queryMaskSensitivity_;
}


/*!
    Sets the subject mask sensitivity to \a ms.
*/
void GramChainAligner::setSubjectMaskSensitivity(MaskSensitivity ms)
{
    subjectMaskSensitivity_ = ms;
}


/*!
    Returns the subject mask sensitivity.
*/
MaskSensitivity GramChainAligner::subjectMaskSensitivity() const
{
    return subjectMaskSensitivity_;
}


void GramChainAligner::showChain(
    char const * query,
    char const * subject,
    Link *       chain,
    unsigned int gramSize
)
{
    std::cout << "---\n";

    std::deque<std::string> piecesA;
    std::deque<std::string> piecesB;
    
    if (chain) {
        std::cout << "chain hits = " << chain->hits << "\n";
    }

    std::string spacer(gramSize - 2, ' ');
    
    // note: chain is reversed
    while (chain) {
        std::ostringstream ostA("");
        std::ostringstream ostB("");
        std::string q(query + chain->y, chain->length);
        std::string s(subject + chain->x, chain->length);
        std::string qdown(query + chain->y + chain->length, 11);
        std::string sdown(subject + chain->x + chain->length, 11);
        std::string qup;
        std::string sup;
        if (query + chain->y - 11 < query) {
            qup = spacer + std::string("||");
        } else {
            qup = std::string(query + chain->y - 11, 11);
        }
        if (subject + chain->x - 11 < subject) {
            sup = spacer + std::string("||");
        } else {
            sup = std::string(subject + chain->x - 11, 11);
        }

        ostA << qup << " /" << q << "\\ " << qdown << " [" << chain->y << "-" << chain->y + chain->length - 1 << "]";
        ostB << sup << " \\" << s << "/ " << sdown << " [" << chain->x << "-" << chain->x + chain->length - 1 << "]";
        piecesA.push_front(ostA.str());
        piecesB.push_front(ostB.str());
        chain = chain->next;
    }

    std::deque<std::string>::iterator itcA(piecesA.begin());
    std::deque<std::string>::iterator iteA(piecesA.end());
    std::deque<std::string>::iterator itcB(piecesB.begin());
    while (itcA != iteA) {
        std::cout << *itcA << "\n"
             << *itcB << "\n"
             << "\n";
        ++itcA;
        ++itcB;
    }
}


namespace gca_ {

/* ------------------------------------------------------------------------ */
    
PositionPool::PositionPool(void)
    :
    pool_(0),
    freeEntries_(0),
    poolSize_(0),
    unusedEntries_(0)
{
}

/* ------------------------------------------------------------------------ */

PositionPool::~PositionPool(void)
{
    delete [] pool_;
}

/* ------------------------------------------------------------------------ */

void PositionPool::reserve(unsigned int poolSize)
{
    if (poolSize > poolSize_) {
        delete [] pool_;
        pool_ = new Position[poolSize];
    }

    freeEntries_ = 0;
    poolSize_    = unusedEntries_ = poolSize;
}

/* ------------------------------------------------------------------------ */

void PositionPool::clear(void)
{
    freeEntries_   = 0;
    unusedEntries_ = poolSize_;
}

/* ------------------------------------------------------------------------ */

unsigned int const LinkPool::DefaultBlockSize = 4096;

/* ------------------------------------------------------------------------ */

LinkPool::LinkPool(void)
    :
    blockSize_(DefaultBlockSize),
    blockCount_(0),
    usedBlocks_(0),
    unusedBlocks_(0),
    unusedLinks_(0),
    unusedLinksCount_(0),
    recycledLinks_(0)
{
    initialize();
}

/* ------------------------------------------------------------------------ */

LinkPool::LinkPool(unsigned int blockSize)
    :
    blockSize_(blockSize),
    blockCount_(0),
    usedBlocks_(0),
    unusedBlocks_(0),
    unusedLinks_(0),
    unusedLinksCount_(0),
    recycledLinks_(0)
{
    initialize();
}

/* ------------------------------------------------------------------------ */

void LinkPool::initialize(void)
{
    usedBlocks_        = new LinkBlock();
    usedBlocks_->links = new Link[blockSize_];
    usedBlocks_->next  = 0;
    unusedLinks_       = usedBlocks_->links;
    unusedLinksCount_  = blockSize_;
    ++blockCount_;
}

/* ------------------------------------------------------------------------ */

LinkPool::~LinkPool(void)
{
    LinkBlock * b(usedBlocks_);
	LinkBlock * t(0);
	
    unsigned int blocks(0);
    
	while (b) {
		t = b;
		b = b->next;
		delete [] t->links;
		delete t;
        ++blocks;
	}
	
	b = unusedBlocks_;
	while (b) {
		t = b;
		b = b->next;
		delete [] t->links;
		delete t;
        ++blocks;
	}
}

/* ------------------------------------------------------------------------ */

unsigned int LinkPool::getBlockCount() const
{
    return blockCount_;
}

/* ------------------------------------------------------------------------ */

unsigned int LinkPool::getBlockSize() const
{
    return blockSize_;
}

/* ------------------------------------------------------------------------ */

} // namespace gca_



} // namespace swift

