// gramindex.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
// 

#include "gramindex.hpp"
#include "gramview.hpp"
#include <boost/iostreams/device/mapped_file.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/scoped_array.hpp>
#include <cstring>
#include <fstream>
#include <sstream>

namespace swift {

// --------------------------------------------------------------------------

namespace detail {

class MapDeleter
{
    public:
    typedef boost::iostreams::mapped_file_source MappedFile;

    MapDeleter(boost::shared_ptr<MappedFile> mappedFile);
    void operator()(void const * p) const;

    private:
    boost::shared_ptr<MappedFile> mappedFile_;
};


/*!
    Constructs a MappedDeleter functor which on invocation unmaps
    the mapped file \a mappedFile.
*/
MapDeleter::MapDeleter(boost::shared_ptr<MappedFile> mappedFile) :
    mappedFile_(mappedFile)
{
}


/*!
    Unmaps the mapped file and destroys the map. The pointer
    argument \a p is ignored.
*/
void MapDeleter::operator()(void const * p) const
{
}

} // namespace detail

// --------------------------------------------------------------------------

struct GramIndex::Header
{
    typedef boost::uint32_t uint32;

    uint32 headerSize;               ///< Size of header + padding.
    uint32 offsetSize;               ///< Size of Offset type, sizeof(Offset).
    uint32 gramSize;                 ///< Size of grams in index.
    uint32 maskSensitivity;          ///< Mask sensitivity of index.
    GramIndex::size_type gramCount;  ///< Number of grams in index.
    GramIndex::Position maxPosition; ///< Position of last indexed gram.
};

char const * GramIndex::LookupSuffix = ".swl";
char const * GramIndex::PositionsSuffix = ".swp";


/*!
    Constructs a gram index for grams of size \a gramSize, over the
    sequences in the \a database. If \a maskSensitivity is
    \l swift::MaskSensitive, the index will not include grams that
    contain masked letters. Otherwise, all valid grams are indexed.
*/
GramIndex * GramIndex::construct(
    SequenceDatabase const * database,
    unsigned int gramSize,
    MaskSensitivity maskSensitivity
)
{
    typedef GramView<Sequence::const_iterator> MyGramView;
    typedef MyGramView::const_iterator GramIterator;

    if (!database || gramSize == 0)
        return 0;

    // Allocate space for positions lookup table.
    Gram::Value lookupSize(Gram::maxValue(gramSize) + 2); // sentinel  
    Offset * lookup(new Offset[lookupSize]);
    ::memset(lookup, 0, lookupSize * sizeof(Offset));
        
	// First pass: Count gram occurrences
    SequenceDatabase::size_type const sequenceCount(database->count());
    unsigned int invalidGrams(0);

	for (SequenceDatabase::size_type i(0); i < sequenceCount; ++i) {
        boost::scoped_ptr<Sequence> sequence(database->sequence(i));
        if (!sequence.get() || sequence->size() < gramSize)
            continue;

        MyGramView view(
            sequence->begin(),
            sequence->end(),
            gramSize,
            maskSensitivity
        );
        
        for (GramIterator j(view.begin()), k(view.end()); j != k; ++j) {
            if ((*j).isValid()) {
                ++lookup[(*j).value()];
            } else {
                ++invalidGrams;
            }
        }
	}
   
	// Obtain gram count and fill in offsets in positions lookup
    Offset gramCount(0);
    for (Gram::Value i(0); i < lookupSize; ++i) {
        Offset count(lookup[i]);
        lookup[i] = gramCount;
        gramCount += count;
    }
	
    // Allocate positions table
    Position * positions(new Position[gramCount + 1]);

	// Second pass: Fill in gram positions
    Position maxPosition(0);
	for (SequenceDatabase::size_type i(0); i < sequenceCount; ++i) {
        boost::scoped_ptr<Sequence> sequence(database->sequence(i));
        if (!sequence.get() || sequence->size() < gramSize)
            continue;

        MyGramView view(
            sequence->begin(),
            sequence->end(),
            gramSize,
            maskSensitivity
        );
        
        Position position(database->offset(i));
        for (GramIterator j(view.begin()), k(view.end()); j != k; ++j) {
            if ((*j).isValid()) {
                positions[lookup[(*j).value()]++] = position;
                maxPosition = position;
            }
            ++position;
        }
	}
	
	// Adjust offsets back
	::memmove(lookup + 1, lookup, (lookupSize - 1) * sizeof(Offset));
	lookup[0] = 0;
     
    return new GramIndex(lookup, positions, maxPosition,
                         gramSize, maskSensitivity);
}


/*!
    Writes the gram index to a set of files with basename \a name.
    Returns true if the gram index was successfully serialized.
*/
bool GramIndex::serialize(std::string const & name) const
{
    std::ofstream lookupStream(
        lookupFile(name, gramSize()).c_str(),
        std::ios::binary | std::ios::trunc
    );

    std::ofstream positionsStream(
        positionsFile(name, gramSize()).c_str(),
        std::ios::binary | std::ios::trunc
    );

    if (!lookupStream.is_open() || !positionsStream.is_open())
        return false;

    serializeHeader(lookupStream);
    serializeLookup(lookupStream);
    serializePositions(positionsStream);

    lookupStream.close();
    positionsStream.close();

    return true;
}


/*!
    Loads the gram index specified by basename \a name and gram size
    \a gramSize. If no such gram index can be found or an error occurs
    while loading, a null pointer is returned.
*/
GramIndex * GramIndex::unserialize(
    std::string const & name,
    unsigned int gramSize
)
{
    GramIndex * nullIndex(0);

    typedef boost::iostreams::mapped_file_source MappedFile;
    boost::shared_ptr<MappedFile> lookupMap(new MappedFile);
    boost::shared_ptr<MappedFile> positionsMap(new MappedFile);

    try {
        lookupMap->open(lookupFile(name, gramSize));
        positionsMap->open(positionsFile(name, gramSize));
        if (!lookupMap->is_open() || !positionsMap->is_open())
            return nullIndex; 
    } catch (std::ios::failure & e) {
        return nullIndex;
    }
    
    // Simple sanity check
    Header const * header(reinterpret_cast<Header const *>(lookupMap->data()));

    if (header->headerSize > lookupMap->size() ||
        header->offsetSize != sizeof(Offset) ||
        positionsMap->size() != (header->gramCount + 1) * sizeof(Position))
        return nullIndex;

    boost::shared_array<Offset const> lookup(
        reinterpret_cast<Offset const *>(
            lookupMap->data() + header->headerSize
        ),
        detail::MapDeleter(lookupMap)
    );

    boost::shared_array<Position const> positions(
        reinterpret_cast<Position const *>(positionsMap->data()),
        detail::MapDeleter(positionsMap)
    );

    return new GramIndex(
        lookup,
        positions,
        header->maxPosition,
        header->gramSize,
        static_cast<MaskSensitivity>(header->maskSensitivity)
    );
}


/*!
    Serializes the index header information to the stream \a outputStream.
    The serialized header is padded to obtain a size which is a multiple
    of sizeof(Offset).
*/
void GramIndex::serializeHeader(std::ostream & outputStream) const
{
    size_t headerSize(
        ((sizeof(Header) + sizeof(Offset)) / sizeof(Offset)) *
        sizeof(Offset)
    );
    boost::scoped_array<char> headerBuffer(new char[headerSize]);
    ::memset(headerBuffer.get(), 0, headerSize);
    Header & header(*(new (headerBuffer.get()) Header));

    header.headerSize = headerSize;
    header.offsetSize = sizeof(Offset);
    header.gramSize = gramSize_;
    header.maskSensitivity = maskSensitivity_;
    header.gramCount = gramCount();
    header.maxPosition = maxPosition_;

    outputStream.write(headerBuffer.get(), headerSize);
}


/*!
    Serializes the position lookup table to the stream \a outputStream.
*/
void GramIndex::serializeLookup(std::ostream & outputStream) const
{
    outputStream.write(
        reinterpret_cast<char const *>(lookup_.get()),
        sizeof(Offset) * (Gram::maxValue(gramSize()) + 2)
    );
}


/*!
    Serializes the positions table to the stream \a outputStream.
*/
void GramIndex::serializePositions(std::ostream & outputStream) const
{
    outputStream.write(
        reinterpret_cast<char const *>(positions_.get()),
        sizeof(Position) * (gramCount() + 1)
    );
}


/*!
    Returns the name of the lookup file for the index specified by
    \a name using gram size \a gramSize.
*/
std::string GramIndex::lookupFile(
    std::string const & name,
    unsigned int gramSize
)
{
    std::ostringstream fileName;
    fileName << name << '_' << gramSize << ".swl";
    return fileName.str();
}


/*!
    Returns the name of the positions file for the index specified by
    \a name using gram size \a gramSize.
*/
std::string GramIndex::positionsFile(
    std::string const & name,
    unsigned int gramSize
)
{
    std::ostringstream fileName;
    fileName << name << '_' << gramSize << ".swp";
    return fileName.str();
}


/*!
    Constructs a gram index for grams of size \a gramSize. The position
    lists are concatenated in \a positions and the offset for the
    beginning of a list for a given gram can be found in \a lookup.
    The mask sensitivity of the gram index is given by \maskSensitivity.
    
    The constructed gram index takes ownership of the \a lookup and
    \a positions tables.
*/
GramIndex::GramIndex(
    Offset const * lookup,
    Position const * positions,
    Position maxPosition,
    unsigned int gramSize,
    MaskSensitivity maskSensitivity
) :
    gramSize_(gramSize),
    maskSensitivity_(maskSensitivity),
    lookup_(lookup),
    positions_(positions),
    maxPosition_(maxPosition)
{
}


/*!
    Constructs a gram index for grams of size \a gramSize. The position
    lists are concatenated in \a positions and the offset for the
    beginning of a list for a given gram can be found in \a lookup.
    The mask sensitivity of the gram index is given by \maskSensitivity.
*/
GramIndex::GramIndex(
    boost::shared_array<Offset const> lookup,
    boost::shared_array<Position const> positions,
    Position maxPosition,
    unsigned int gramSize,
    MaskSensitivity maskSensitivity
) :
    gramSize_(gramSize),
    maskSensitivity_(maskSensitivity),
    lookup_(lookup),
    positions_(positions),
    maxPosition_(maxPosition)
{
}


// --------------------------------------------------------------------------

} // namespace swift
