// hashset.hpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
// 

#ifndef _SWIFT_HASHSET_
#define _SWIFT_HASHSET_

#include "hashtable.hpp"
#include "identity.hpp"
#include <functional>
#include <utility>

namespace swift
{


template <
    typename Key,
    typename EqualKey = std::equal_to<Key>,
    typename Hasher = Hash<Key>,
    typename IteratorCategory = DefaultHashIteratorTag
>
class HashSet
{
    typedef HashTable<
        Key,
        Key,       // FIXME should be "Key const"
        Identity<Key>,     // FIXME same here
        EqualKey,
        Hasher,
        IteratorCategory
    > HT;

    public:
    typedef typename HT::key_type key_type;
    typedef typename HT::value_type value_type;
    typedef typename HT::const_iterator const_iterator;

    public:
    HashSet(
        size_t initialCapacity = 2048,
        float maxLoadFactor = 0.75,
        EqualKey const & equalKey = EqualKey(),
        Hasher const & hasher = Hasher()
    ) :
        hashTable_(initialCapacity, maxLoadFactor, equalKey, hasher) { }

    const_iterator begin() const
    {
        return hashTable_.begin();
    }

    const_iterator end() const
    {
        return hashTable_.end();
    }

    bool contains(key_type const & key) const
    {
        return hashTable_.contains(key);
    }

    bool find(key_type const & key) const
    {
        return hashTable_.find(key).second;
    }

    bool insert(value_type const & value)
    {
        return hashTable_.insert(value).second;
    }

    bool erase(key_type const & key)
    {
        return hashTable_.erase(key);
    }

    void clear()
    {
        hashTable_.clear();
    }

	size_t capacity() const
    {
        return hashTable_.capacity();
    }

    size_t size() const
    {
        return hashTable_.size();
    }

	float loadFactor() const
    {
        return hashTable_.loadFactor();
    }

	float maxLoadFactor() const
    {
        return hashTable_.maxLoadFactor();
    }

    private:
    HT hashTable_;
}; 


} // namespace swift

#endif

