// hashtable.hpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
// 

#ifndef _SWIFT_HASHTABLE_
#define _SWIFT_HASHTABLE_

#include "assert.hpp"
#include "hash.hpp"
#include <functional>
#include <iterator>
#include <string>
#include <utility>
#include <cstring>

/**
	TODO:
	
	acquireEntry():
		- prevent recursion
		- make inlined
		- improve, or redo as the Queue class
		
	constructor():
		- replace assert with exception??
*/

namespace swift
{


template <
    typename Key,
    typename Value,
    typename ExtractKey,
    typename EqualKey,
    typename Hasher,
    typename IteratorCategory
>
class HashTable;


struct DefaultHashIteratorTag { };
struct FastHashIteratorTag { };

template <
    typename Key,
    typename Value,
    typename ExtractKey,
    typename EqualKey,
    typename Hasher,
    typename IteratorCategory = DefaultHashIteratorTag
>
class HashTableIterationPolicy
{
    protected:
    struct Node
    {
        Node * next;
        Value value;
    };

    public:
    typedef HashTable<
        Key, Value, ExtractKey, EqualKey, Hasher, IteratorCategory
    > hashtable_type;

    HashTableIterationPolicy(hashtable_type const & hashTable) :
        hashTable_(hashTable) { }


    /*!
        Const iterator
    */
    class const_iterator
    {
        public:
        typedef typename hashtable_type::value_type value_type;
        typedef value_type &        reference;
        typedef value_type const &  const_reference;
        typedef value_type *        pointer;
        typedef value_type const *  const_pointer;

        const_iterator(hashtable_type const & hashTable, Node const * node) :
            hashTable_(hashTable),
            node_(node) { }


        /*!
            Returns a const reference to the value at the iterator position.
        */
        const_reference operator*() const
        {
            return node_->value;
        }


        /*!
            Returns a const pointer to the value at the iterator position.
        */
        const_pointer operator->() const
        {
            return &(operator*());
        }


        /*!
            Preincrement the iterator.
        */
        const_iterator & operator++()
        {
            node_ = hashTable_.nextNode(node_);
            return *this;
        }


        /*!
            Postincrements the iterator.
        */
        const_iterator operator++(int)
        {
            const_iterator it(*this);
            ++(*this);
            return it;
        }

        /*!
            Iterator equivalence. Returns \c true if the iterators point
            to the same node.
        */
        bool operator==(const_iterator const & other) const
        {
            return node_ == other.node_;
        }


        /*!
            Iterator inequivalence. Returns \c true if the iterators point
            to different nodes.
        */
        bool operator!=(const_iterator const & other) const
        {
            return !(*this == other);
        }

        private:
        hashtable_type const & hashTable_;
        Node const * node_;
    };

    protected:
    Node * firstNode() const
    {
        for (size_t i(0), j(hashTable_.bucketCapacity_); i < j; ++i)
            if (hashTable_.buckets_[i])
                return hashTable_.buckets_[i];
        return 0;
    }
    void linkNode(Node * node) { }
    void unlinkNode(Node * node) { }
    void unlinkAllNodes() { }

    const_iterator make_const_iterator(Node * node) const
    {
        return const_iterator(hashTable_, node);
    }

    private:
    hashtable_type const & hashTable_;
};


/*!
    Partial specialization of HashTableIterationPolicy. This specialization
    provides fast advancing iterators at the expense of 2 extra pointers per
    hash table element.
*/
template <
    typename Key,
    typename Value,
    typename ExtractKey,
    typename EqualKey,
    typename Hasher
>
class HashTableIterationPolicy<
    Key, Value, ExtractKey, EqualKey, Hasher, FastHashIteratorTag
>
{
    protected:
    struct Node
    {
        Value value;
        Node * next;
        Node * succ;
        Node * pred;
    };


    public:
    typedef HashTable<
        Key, Value, ExtractKey, EqualKey, Hasher, FastHashIteratorTag
    > hashtable_type;

    HashTableIterationPolicy(hashtable_type const & hashTable) :
        hashTable_(hashTable_),
        root_(0) { }


    /*!
        Const iterator
    */
    class const_iterator
    {
        public:
        typedef typename hashtable_type::value_type value_type;
        typedef value_type &        reference;
        typedef value_type const &  const_reference;
        typedef value_type *        pointer;
        typedef value_type const *  const_pointer;

        const_iterator(Node const * node) : node_(node) { }


        /*!
            Returns a const reference to the value at the iterator position.
        */
        const_reference operator*() const
        {
            return node_->value;
        }


        /*!
            Returns a const pointer to the value at the iterator position.
        */
        const_pointer operator->() const
        {
            return &(operator*());
        }


        /*!
            Preincrement the iterator.
        */
        const_iterator & operator++()
        {
            node_ = node_->succ;
            return *this;
        }


        /*!
            Postincrements the iterator.
        */
        const_iterator operator++(int)
        {
            const_iterator it(*this);
            ++(*this);
            return it;
        }

        /*!
            Iterator equivalence. Returns \c true if the iterators point
            to the same node.
        */
        bool operator==(const_iterator const & other) const
        {
            return node_ == other.node_;
        }


        /*!
            Iterator inequivalence. Returns \c true if the iterators point
            to different nodes.
        */
        bool operator!=(const_iterator const & other) const
        {
            return !(*this == other);
        }

        private:
        Node const * node_;
    };


    protected:
    Node * firstNode() const
    {
        return root_;
    }

    void linkNode(Node * node)
    {
        if (root_)
            root_->pred = node;
        node->succ = root_;
        root_ = node;
        node->pred = 0;
    }
    
   
    void unlinkNode(Node * node)
    {
        if (root_ == node) {
            root_ = node->succ;
            if (root_)
                root_->pred = 0;
            return;
        }

        node->pred->succ = node->succ;

        if (node->succ)
            node->succ->pred = node->pred;
    }


    void unlinkAllNodes()
    {
        root_ = 0;
    }


    const_iterator make_const_iterator(Node * node) const
    {
        return const_iterator(node);
    }


    private:
    hashtable_type const & hashTable_;
    Node * root_;
};


/*!
    Hash table.

    Template parameters:

    \li \c Key Key type.
    \li \c Value Value type.
    \li \c ExtractKey Functor to extract the key from a value. Must supply
        the method <tt>Key operator()(Value const & value)</tt>. Must be
        default constructable.
    \li \c EqualKey Functor to compare keys. Must be default constructable.
    \li \c Hasher Functor to generate an integer for a key. Must supply the
        method <tt>size_t operator()(Key const & key)</tt>. Must be default
        constructable.
    \li \c IteratorCategory Determines the iterator category. If set to
        \c HashTableFastIteratorTag, fast iterators are used at the expense
        of increased memory usage. This increases the size of each stored
        node by <tt>2 * sizeof(void *)</tt>. Default is 
        \c HashTableDefaultIteratorTag which minimizes memory usage, but
        at the cost of slower iterators.
*/
template <
    typename Key,
    typename Value,
    typename ExtractKey,
    typename EqualKey = std::equal_to<Key>,
    typename Hasher = Hash<Key>,
    typename IteratorCategory = DefaultHashIteratorTag
>
class HashTable :
    public HashTableIterationPolicy<
        Key, Value, ExtractKey, EqualKey, Hasher, IteratorCategory
    >
{
    public:
    typedef Key      key_type;
    typedef Value    value_type;
    typedef EqualKey equals_type;
    typedef Hasher   hasher_type;
    typedef typename HashTable::Node Node;
    typedef typename HashTable::const_iterator const_iterator;

    public:
    // FIXME 1090807696
    // Constness is important but make pooled allocator difficult.
    // See 1090807721
    //typedef pair<Key /*const*/, Value> value_type;

    private:
    HashTable();

    public:
    /*!
        Constructs a hash table with initial capacity \a initialCapacity and
        maximum load factor of \a maxLoadFactor. The initial capacity must be a
        power of two.
    */
	HashTable(
        size_t initialCapacity = 2048,
        float maxLoadFactor = 0.75,
        EqualKey equalKey = EqualKey(),
        Hasher hasher = Hasher()
    ) :
        HashTableIterationPolicy<
            Key, Value, ExtractKey, EqualKey, Hasher, IteratorCategory
        >(*this),
        bucketCapacity_(0),
        maxLoadFactor_(maxLoadFactor > 0 ? maxLoadFactor : 0.75),
        buckets_(0),
        nodeCapacity_(0),
        usedBlocks_(0),
        freeBlocks_(0),
        unusedEntries_(0),
        unusedEntriesLeft_(0),
        recycledEntries_(0),
        recycledEntriesLeft_(0),
        nodeCount_(0),
        equalKeys_(equalKeys_),
        hashKey_(hasher)
    {
        ASSERT(
            (initialCapacity & (initialCapacity - 1)) == 0 && 
                initialCapacity > 0,
            "Capacity is power of two and larger than zero."
        );
        
        setCapacity(initialCapacity);
    }


    /*!
        Destroys the hash table.
    */
	~HashTable()
    {
        delete [] buckets_;
        
        NodeBlock * b(usedBlocks_);
        NodeBlock * t(0);
        
        while (b) {
            t = b;
            b = b->next;
            delete [] t->entries;
            delete t;
        }
        
        b = freeBlocks_;
        while (b) {
            t = b;
            b = b->next;
            delete [] t->entries;
            delete t;
        }
    }


    /*!
        Returns an iterator to the first node in the hash table, or end()
        if the table is empty.
    */
    const_iterator begin() const
    {
/* to avoid the use of -fpermissive
specify firstNode() and make_const_iterator via baseclass
*/
        Node * node(
HashTableIterationPolicy<
Key, Value, ExtractKey, EqualKey, Hasher, IteratorCategory
    >::firstNode());
        if (node)
            return HashTableIterationPolicy<
Key, Value, ExtractKey, EqualKey, Hasher, IteratorCategory
    >::make_const_iterator(node);
        return end();
    }


    /*!
        Returns an iterator to the end of the hash table.
    */
    const_iterator end() const
    {
        return HashTableIterationPolicy<
Key, Value, ExtractKey, EqualKey, Hasher, IteratorCategory
    >::make_const_iterator(0);
    }


    /*!
        Returns true if the hash table contains an element whose key is \a key.
    */
    bool contains(key_type const & key) const
    {
        Node * p(buckets_[hashKey_(key) & hashMask_]);
        
        while (p) {
            if (equalKeys_(extractKey_(p->value), key))
                return true;
            p = p->next;
        }
        
        return false;
    }


    /*!
        Returns a pair where the second member is true if the hash table
        contains an element with key \a key. The first member contains
        a pointer to the value of the element, or a null pointer if the
        element was not found.
    */
    std::pair<value_type *, bool> find(key_type const & key) const
    {
        Node * p(buckets_[hashKey_(key) & hashMask_]);
        
        while (p) {
            if (equalKeys_(extractKey_(p->value), key))
                return std::pair<value_type *, bool>(&p->value, true);
            p = p->next;
        }
        
        return std::pair<value_type *, bool>(0, false);
    }


    /*!
        Inserts the value \a value in the hash table. If a value with the
        same key already exitst in the hash table, a pair is returned where
        the first member is a pointer to the value and the second member is false.
        If the value was successfully inserted, the first member of the returned
        pair contains a pointer to the inserted value and the second member is
        true.
    */
    std::pair<value_type *, bool> insert(value_type const & value)
    {
        size_t hashedKey(hashKey_(extractKey_(value)));
        
        // FIXME: Is this loop generating a lot of cache misses because nodes
        // in bucket aren't neccessarily consecutively placed in mem?  would it
        // be cheaper with a vector, and set null pointer in when an node has
        // been removed?
        Node * p(buckets_[hashedKey & hashMask_]);

        while (p) {
            if (equalKeys_(extractKey_(p->value), extractKey_(value)))
                return std::pair<value_type *, bool>(&p->value, false);
            p = p->next;
        }
       
        p = acquireNode(); // may trigger capacity change and thus mask change

        p->value = value;
        p->next = buckets_[hashedKey & hashMask_];
        buckets_[hashedKey & hashMask_] = p;

        linkNode(p);

        return std::pair<value_type *, bool>(&p->value, true);
    }	
    

    /*!
        Erases the element whose key is \a key. Returns true if the element
        was erased.
    */
    bool erase(key_type const & key)
    {
        size_t i(hashKey_(key) & hashMask_);
        Node * p(buckets_[i]);

        if (!p)
            return false;
        
        if (equalKeys_(extractKey_(p->value), key)) {
            buckets_[i] = p->next;
            unlinkNode(p);
            recycleNode(p);
            return true;
        }

        Node * q(p);
        p = p->next;	

        while (p) {
            if (equalKeys_(extractKey_(p->value), key)) {
                q->next = p->next;
                unlinkNode(p);
                recycleNode(p);
                return true;
            }

            q = p;
            p = p->next;
        }

        return false;
    }
    

    /*!
        Removes all elements from the hash table.
    */
    void clear()
    {
        memset(buckets_, 0, sizeof(Node *) * bucketCapacity_);
        
        nodeCount_          = 0;
        unusedEntriesLeft_   = 0;
        recycledEntriesLeft_ = 0;

/*
To avoid the use of -fpermissive:

Use BaseClass<T>::unlinkAllNodes();

See:
http://gcc.gnu.org/onlinedocs/gcc/Name-lookup.html
Another, similar example involves calling member functions of a base class:

       template <typename T> struct Base {
           int f();
       };
     
       template <typename T> struct Derived : Base<T> {
           int g() { return f(); };
       };

Again, the call to f() is not dependent on template arguments (there are no arguments that depend on the type T, and it is also not otherwise specified that the call should be in a dependent context). Thus a global declaration of such a function must be available, since the one in the base class is not visible until instantiation time. The compiler will consequently produce the following error message:

       x.cc: In member function `int Derived<T>::g()':
       x.cc:6: error: there are no arguments to `f' that depend on a template
          parameter, so a declaration of `f' must be available
       x.cc:6: error: (if you use `-fpermissive', G++ will accept your code, but
          allowing the use of an undeclared name is deprecated)

To make the code valid either use this->f(), or Base<T>::f(). Using the -fpermissive flag will also let the compiler accept the code, by marking all function calls for which no declaration is visible at the time of definition of the template for later lookup at instantiation time, as if it were a dependent call. We do not recommend using -fpermissive to work around invalid code, and it will also only catch cases where functions in base classes are called, not where variables in base classes are used (as in the example above). 
*/
        HashTableIterationPolicy<
        Key, Value, ExtractKey, EqualKey, Hasher, IteratorCategory
    >::unlinkAllNodes();

        if (!freeBlocks_) {
            freeBlocks_ = usedBlocks_;
        } else {
            NodeBlock * t(freeBlocks_);
            while (t->next)
                t = t->next;
            t->next = usedBlocks_;
        }
        usedBlocks_ = 0;
    }


    /*!
        Returns the bucket capacity of the hash table.
    */
	size_t capacity() const
    {
        return bucketCapacity_;
    }


    /*!
        Returns the number of elements in the hash table.
    */
    size_t size() const
    {
        return nodeCount_;
    }


    /*!
        Returns the current load factor for the hash table.
    */
	float loadFactor() const
    {
        return (float)size()/(float)capacity();
    }


    /*!
        Returns the maximum load factor. When the maximum load factor
        is exceeded, the hash table will reallocate and expand.
    */
	float maxLoadFactor() const
    {
        return maxLoadFactor_;
    }


    private:
    struct NodeBlock
    {
        NodeBlock * next;
        Node * entries;
        size_t size;
    };

    /*!
        \internal
        Acquires an node.
    */
    Node * acquireNode()
    {
        // attempt acquiring a recycled node
        if (recycledEntriesLeft_) {
            Node * t(recycledEntries_);
            recycledEntries_ = t->next;
            --recycledEntriesLeft_;
            
            ++nodeCount_;
            return t;
        }

        // attempt acquiring an unused node
        if (unusedEntriesLeft_) {
            Node * t(unusedEntries_++);
            --unusedEntriesLeft_;

            ++nodeCount_;
            return t;
        }
        
        // try getting a free block
        if (freeBlocks_) {
            NodeBlock * b(freeBlocks_);
            freeBlocks_ = freeBlocks_->next;
            b->next = usedBlocks_;
            usedBlocks_ = b;

            // Optimization of :
            //
            //    m_unusedEntries = b->entries;
            //    m_unusedEntriesLeft = b->size;
            //    return acquireNode();
            //
            // follows.
            ++nodeCount_;
            unusedEntries_ = b->entries + 1;
            unusedEntriesLeft_ = b->size - 1;
            return (b->entries);
        }
        
        // no free blocks, check if maxLoad has been reached
        if (nodeCapacity_ >= (size_t)(bucketCapacity_ * maxLoadFactor_)) {
            setCapacity(bucketCapacity_ << 1);
        }
        
        size_t newBlockSize(
            (size_t)(bucketCapacity_ * maxLoadFactor_) - nodeCapacity_
        );
        
        // TODO : here blockSize limiting may take place
        
        NodeBlock * newBlock(new NodeBlock());
        newBlock->entries = new Node[newBlockSize];
        newBlock->size = newBlockSize;
        nodeCapacity_ += newBlockSize;
        
        // Optimization of :
        //
        //    newBlock->next = m_freeBlocks;
        //    m_freeBlocks = newBlock;
        //    return acquireNode();
        //
        // follows.
        newBlock->next = usedBlocks_;
        usedBlocks_ = newBlock;
        ++nodeCount_;
        unusedEntries_ = newBlock->entries + 1;
        unusedEntriesLeft_ = newBlock->size - 1;
        return (newBlock->entries);
    }


    /*!
        \internal
        Recycles node. The node is put into the pool of recycled entries. It
        is assumed that the node has unlinked prior to the method invocation.
    */
    void recycleNode(Node * node)
    {
        if (node) {
            node->next = recycledEntries_;
            recycledEntries_ = node;
            ++recycledEntriesLeft_;
            
            --nodeCount_;
        }
    }

    /*!
        \internal 
        Increases bucket capacity.  The capacity increase requires rehashing
        all entries into the new buckets, and this step assumes that all
        usedBlocks are fully utilized.
    */
    void setCapacity(size_t newCapacity)
    {
        bucketCapacity_ = newCapacity;
        hashMask_       = bucketCapacity_ - 1;
        
        if (buckets_)
            delete [] buckets_;

        buckets_ = new Node *[bucketCapacity_];
        memset(buckets_, 0, sizeof(Node *) * bucketCapacity_);
        
#if 0
        std::cerr << "Rehashing " << nodeCount_ << " entries." << std::endl;
        std::cerr << "Increased capacity to " << newCapacity << " buckets." << std::endl;
#endif
        
        for (NodeBlock * b(usedBlocks_); b ; b = b->next) {
            Node * e(b->entries);
            for (int i(b->size); i > 0; --i, ++e) {
                size_t k(hashKey_(extractKey_(e->value)) & hashMask_);
                e->next = buckets_[k];
                buckets_[k] = e;
            }
        }
    }


    /*!
        \internal 
        Returns the next node after \a node or a null pointer. Used by
        iterators.
    */
    Node * nextNode(Node const * node) const
    {
        Node * next(node->next);
        if (!next) {
            size_t i(hashKey_(extractKey_(node->value)) & hashMask_);
            while (!next && ++i < bucketCapacity_)
                next = buckets_[i];
        }
        return next;
    }


    size_t bucketCapacity_;
    float maxLoadFactor_;
    Node ** buckets_;

    size_t nodeCapacity_;
    NodeBlock * usedBlocks_;
    NodeBlock * freeBlocks_;

    Node * unusedEntries_;
    size_t unusedEntriesLeft_;

	Node * recycledEntries_;
	size_t recycledEntriesLeft_;

	size_t nodeCount_;
	
	size_t hashMask_;
		
    ExtractKey extractKey_;
    EqualKey equalKeys_;
    Hasher hashKey_;

    friend class HashTableIterationPolicy<
            Key, Value, ExtractKey, EqualKey, Hasher, IteratorCategory
    >;

    friend class HashTable::const_iterator;
};


} // namespace swift

#endif

