// semigappedxdropaligner.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
// 

#include "semigappedxdropaligner.hpp"
#include "assert.hpp"
#include <algorithm>


namespace swift
{


namespace
{
    using std::max;
}


/*!
    Constructs a SemiGappedXDropAligner object. The score function and matrix
    should be configured before the aligner is used.

    \sa setScoreMatrix(), setScoreFunction()
*/
SemiGappedXDropAligner::SemiGappedXDropAligner() :
    rowSizes_(0),
    bestRow_(0),
    insertQrow_(0)
{
    dropOff_        = 38;
    extensionN_     = 10;
    extensionDummy_ = -666;
}


/*!
    Destroys the SemiGappedXDropAligner object.
*/
SemiGappedXDropAligner::~SemiGappedXDropAligner()
{
    delete [] bestRow_;
    delete [] insertQrow_;
}


/*!
    Sets the drop off value to \a dropOff.
*/
void SemiGappedXDropAligner::setDropOff(int dropOff)
{
    dropOff_ = dropOff;
}


/*!
    Sets the score function.
*/
void SemiGappedXDropAligner::setScoreFunction(
    ScoreFunction const & scoreFunction
)
{
    scoreFunction_ = scoreFunction;
}


/*!
    Finds an alignment using the semi gapped X-dropoff algorithm. The
    sequences and the seed position are specified in the parameter object \a p.
    The score and end points of the alignment are returned in the result
    object \a result.
*/
void SemiGappedXDropAligner::findAlignment(
    Parameters const & p,
    Result & result
)
{
    result.score = findAlignmentBeforeSeed(
        p.query,
        p.subject,
        p.queryOffset,
        p.subjectOffset,
        result.queryBegin,
        result.subjectBegin
    );

    result.score += findAlignmentAfterSeed(
        p.query + p.queryOffset,
        p.subject + p.subjectOffset,
        p.querySize - p.queryOffset,
        p.subjectSize - p.subjectOffset,
        result.queryEnd,
        result.subjectEnd
    );

    result.queryEnd   += p.queryOffset;
    result.subjectEnd += p.subjectOffset;

    result.score += 
        scoreFunction_[p.query[p.queryOffset]][p.subject[p.subjectOffset]];
}


void SemiGappedXDropAligner::findAlignmentBeforeSeed(
    Parameters const & p,
    Result & result
)
{
    ASSERT(p.queryOffset < p.querySize, "query offset out of bounds");
    ASSERT(p.subjectOffset < p.subjectSize, "subject offset out of bounds");

    result.score = findAlignmentBeforeSeed(
        p.query,
        p.subject,
        p.queryOffset,
        p.subjectOffset,
        result.queryBegin,
        result.subjectBegin
    );
}


void SemiGappedXDropAligner::findAlignmentAfterSeed(
    Parameters const & p,
    Result & result
)
{
    ASSERT(p.queryOffset < p.querySize, "query offset out of bounds");
    ASSERT(p.subjectOffset < p.subjectSize, "subject offset out of bounds");

    result.score = findAlignmentAfterSeed(
        p.query + p.queryOffset,
        p.subject + p.subjectOffset,
        p.querySize - p.queryOffset,
        p.subjectSize - p.subjectOffset,
        result.queryEnd,
        result.subjectEnd
    );

    result.queryEnd   += p.queryOffset;
    result.subjectEnd += p.subjectOffset;
}


/*!
    Finds an alignment beginning in the seed position (\a queryOffset, \a
    subjectOffset) and ending in (q, s) where q is in the range
    [0, \a queryOffset) and s is in the range [0, \a subjectOffset).
*/
int SemiGappedXDropAligner::findAlignmentBeforeSeed(
    unsigned char const * query,
    unsigned char const * subject,
    unsigned int queryOffset,
    unsigned int subjectOffset,
    unsigned int & queryBegin,
    unsigned int & subjectBegin
)
{
    if (queryOffset == 0 || subjectOffset == 0) {
        queryBegin = queryOffset;
        subjectBegin = subjectOffset;
        return 0;
    }

    unsigned char const * bestSubjectPosition;
    unsigned char const * subjectPosition;
    unsigned char const * startSubjectPosition;

    unsigned char const * bestQueryPosition;
    unsigned char const * queryPosition;

    int bestScore(0);
    int insertS;
    int subjectDistance;
    int oldBest, previousOldBest;
    unsigned char rightOfDropoff;
    int queryCount, subjectCount;

    // Declare processing rows for storing match, insert-subject and
    // insert-query values if current malloced rows aren't big enough
    if (subjectOffset >= rowSizes_) {
        delete [] bestRow_;
        delete [] insertQrow_;
        rowSizes_   = subjectOffset << 1;
        bestRow_    = new int[rowSizes_];
        insertQrow_ = new int[rowSizes_];
    }

    bestSubjectPosition =
        subjectPosition =
        startSubjectPosition = subject + subjectOffset - 1;

    bestQueryPosition = queryPosition = query + queryOffset - 1;

    ASSERT(subject <= subjectPosition, "Invalid subject offset");
    ASSERT(query <= queryPosition, "Invalid query offset");

    // Initialize row pointers
    int rowOffset(subjectPosition - subject);
    int * bestRow(bestRow_ + rowOffset);
    int * insertQrow(insertQrow_ + rowOffset);

    // Set initial row dropoff and column dropoff
    unsigned char const * rowDropoff(subject);
    unsigned char const * columnDropoff(subject + subjectOffset);

    // Using first column of query matrix
    int const * matrixRow(scoreFunction_[*queryPosition]);

    // -----FIRST ROW-----

    // -----FIRST CELL-----
    // Set M value for bottom-right cell
    int match(matrixRow[*subjectPosition]);

    // M must be the best
    *bestRow = match;

    // Only gap opens possible
    *insertQrow = insertS = match - scoreFunction_.gapOpenCost();

    // If this is the best-yet scoring cell
    if (match > bestScore) {
        // Update best start cell data
        bestScore = match;
        bestQueryPosition = queryPosition;
        bestSubjectPosition = subjectPosition;
    }

    subjectDistance = 0;
    --subjectPosition;
    --bestRow;
    --insertQrow;

    // ----- REMAINING CELLS -----
    // For each remaining column in the bottom row, scanning from right-to-left
    while (subjectPosition >= subject) {
        // Set value for M
        match = matrixRow[*subjectPosition] -
                scoreFunction_.gapOpenCost() -
                scoreFunction_.gapExtendCost() * subjectDistance;

        // Determine the best of M and Iy
        if (match > insertS) {
            *bestRow = match;

            // Calculate new Iy
            insertS = max(
                match - scoreFunction_.gapOpenCost(),
                insertS - scoreFunction_.gapExtendCost()
            );
        } else {
            *bestRow = insertS;

            // Since M <= Iy, new Iy must derive from Iy
            insertS -= scoreFunction_.gapExtendCost();
        }

        // Set DUMMY Ix value, which should never be used
        *insertQrow = extensionDummy_;

        // If this is the best-yet scoring cell
        if (match > bestScore) {
            // Update best start cell data
            bestScore = match;
            bestQueryPosition = queryPosition;
            bestSubjectPosition = subjectPosition;
        }

        // If score at current cell is below dropoff
        if (bestScore > *bestRow + dropOff_) {
            // Record dropoff position
            rowDropoff = subjectPosition;
            // And stop processing row
            break;
        }

        --subjectPosition;
        --bestRow;
        --insertQrow;
        ++subjectDistance;
    }

    // Start queryCount at N. Only allow insertS for every Nth row when
    // queryCount reaches 0
    queryCount = extensionN_;

    // -----REMAINING ROWS-----
    while (queryPosition > query && rowDropoff < columnDropoff) {
        --queryPosition;
        --queryCount;
        subjectPosition = columnDropoff - 1;

        // Determine subjectCount for initial subjectPosition. Is used to only
        // allow insertQ when subjectOffset % extensionN_ == 0
        subjectCount = (int)(startSubjectPosition - subjectPosition) % extensionN_;
        if (subjectCount)
            subjectCount = extensionN_ - subjectCount;

        // Reset row pointers to start of rows
        rowOffset  = subjectPosition - subject;
        bestRow    = bestRow_ + rowOffset;
        insertQrow = insertQrow_ + rowOffset;

        // Using next column of query matrix
        matrixRow = scoreFunction_[*queryPosition];

        // ************ All rows we are not allowing insertS
        if (queryCount) {
            // ** No insertQ allowed this column, this cell will only get a
            // DUMMY score
            if (subjectCount) {
                previousOldBest = *bestRow;
                *bestRow = extensionDummy_;

                // Score at this cell is below dropoff
                columnDropoff  = subjectPosition;
                rightOfDropoff = 1;
            } else {
                // ** We are allowing insertQ this column

                // -----FAR RIGHT CELL-----
                // Record some old values
                previousOldBest = *bestRow;

                // Set Ix value
                *bestRow = *insertQrow;
                *insertQrow -= scoreFunction_.gapExtendCost();

                // If score at current cell is below dropoff
                if (bestScore > *bestRow + dropOff_) {
                    // Record dropoff position
                    columnDropoff  = subjectPosition;
                    rightOfDropoff = 1;
                } else {
                    // We are left of the column dropoff for this row
                    rightOfDropoff = 0;
                }

                // Reset subjectCount
                subjectCount = extensionN_;
            }

            --subjectPosition;
            --bestRow;
            --insertQrow;
            --subjectCount;

            // -----CELLS RIGHT OF ROW DROPOFF-----
            while (subjectPosition >= rowDropoff) {
                // ** We are not allowing insertQ this column
                if (subjectCount) {
                    // Calculate new M value, which is also the best
                    oldBest = *bestRow;
                    match = matrixRow[*subjectPosition] + previousOldBest;
                    *bestRow = match;
                    previousOldBest = oldBest;

                    // If this is the best-yet scoring cell
                    if (match > bestScore) {
                        // Update best start cell data
                        bestScore = match;
                        bestQueryPosition = queryPosition;
                        bestSubjectPosition = subjectPosition;
                    }
                } else {
                    // We are allowing insertQ this column

                    // Calculate new M value
                    oldBest = *bestRow;
                    match = matrixRow[*subjectPosition] + previousOldBest;
                    previousOldBest = oldBest;

                    // Determine the best of M and Ix
                    if (match > *insertQrow) {
                        *bestRow = match;

                        // Calculate new Ix
                        *insertQrow = max(
                            match - scoreFunction_.gapOpenCost(),
                            *insertQrow - scoreFunction_.gapExtendCost()
                        );
                    } else {
                        *bestRow = *insertQrow;

                        // Since M <= Ix, new Ix must derive from Ix
                        *insertQrow -= scoreFunction_.gapExtendCost();
                    }

                    // If this is the best-yet scoring cell
                    if (match > bestScore) {
                        // Update best start cell data
                        bestScore = match;
                        bestQueryPosition = queryPosition;
                        bestSubjectPosition = subjectPosition;
                    }

                    // Reset subjectCount
                    subjectCount = extensionN_;
                }

                --subjectPosition;
                --bestRow;
                --insertQrow;
                --subjectCount;
            }

            // -----SINGLE CELL LEFT OF ROW DROPOFF -----
            if (!(bestScore > previousOldBest + dropOff_) &&
                 (subjectPosition >= subject)) {
                // Set value for best
                match = previousOldBest + matrixRow[*subjectPosition];
                *bestRow = match;

                // Set DUMMY values for Ix
                *insertQrow = extensionDummy_;

                if (match + dropOff_ >= bestScore) {
                    // Record dropoff position
                    rowDropoff = subjectPosition;
                }
            }
        } else {
            // ************ Every Nth row we allow insertS

            //
            // -----FAR RIGHT CELL-----

            // ** No insertQ allowed this column, this cell will only get a
            // DUMMY score
            if (subjectCount) {
                previousOldBest = *bestRow;
                *bestRow = extensionDummy_;

                // Score at this cell is below dropoff
                columnDropoff  = subjectPosition;
                rightOfDropoff = 1;
            } else {
                // ** We are allowing insertQ this column
                // Record some old values
                previousOldBest = *bestRow;

                // Set Ix value
                *bestRow = *insertQrow;
                *insertQrow -= scoreFunction_.gapExtendCost();

                // Set DUMMY value for Iy, which should never be used
                insertS = extensionDummy_;

                // If score at current cell is below dropoff
                if (bestScore > *bestRow + dropOff_) {
                    // Record dropoff position
                    columnDropoff = subjectPosition;
                    rightOfDropoff = 1;
                } else {
                    // We are left of the column dropoff for this row
                    rightOfDropoff = 0;
                }

                // Reset subjectCount
                subjectCount = extensionN_;
            }

            --subjectPosition;
            --bestRow;
            --insertQrow;
            --subjectCount;

            // -----CELLS RIGHT OF ROW DROPOFF-----
            while (subjectPosition >= rowDropoff) {
                // ** We are not allowing insertQ this column
                if (subjectCount) {
                    // Remember old M value (for cell below this one)
                    oldBest = *bestRow;
                    match = matrixRow[*subjectPosition] + previousOldBest;
                    previousOldBest = oldBest;

                    // Determine the best of M and Iy
                    if (match > insertS) {
                        *bestRow = match;

                        // Calculate new Iy
                        insertS = max(
                            match - scoreFunction_.gapOpenCost(),
                            insertS - scoreFunction_.gapExtendCost()
                        );
                    } else {
                        *bestRow = insertS;

                        // Since M <= Iy, new Iy must derive from Iy
                        insertS -= scoreFunction_.gapExtendCost();
                    }

                    // If this is the best-yet scoring cell
                    if (match > bestScore) {
                        // Update best start cell data
                        bestScore = match;
                        bestQueryPosition = queryPosition;
                        bestSubjectPosition = subjectPosition;
                    }

                    // If score at current cell (and cells to its right) are
                    // below dropoff
                    if (rightOfDropoff) {
                        if (bestScore > *bestRow + dropOff_) {
                            // Record dropoff position
                            columnDropoff = subjectPosition;
                        } else {
                            // We are left of the column dropoff for this row
                            rightOfDropoff = 0;
                        }
                    }
                } else {
                    // ** We are allowing insertQ this column
                    // Remember old M value (for cell below this one)
                    oldBest = *bestRow;
                    match = matrixRow[*subjectPosition] + previousOldBest;
                    previousOldBest = oldBest;

                    // Determine the best of M, Ix and Iy
                    if (match > insertS) {
                        if (match > *insertQrow) {
                            // Match is largest
                            *bestRow = match;

                            // Calculate new Ix
                            *insertQrow = max(
                                match - scoreFunction_.gapOpenCost(),
                                *insertQrow - scoreFunction_.gapExtendCost()
                            );

                            // Calculate new Iy
                            insertS = max(
                                match - scoreFunction_.gapOpenCost(),
                                insertS - scoreFunction_.gapExtendCost()
                            );

                            // If this is the best-yet scoring cell
                            if (match > bestScore) {
                                // Update best start cell data
                                bestScore = match;
                                bestQueryPosition = queryPosition;
                                bestSubjectPosition = subjectPosition;
                            }
                        } else {
                            // insertQ is largest
                            *bestRow = *insertQrow;

                            // Calculate new Ix
                            *insertQrow -= scoreFunction_.gapExtendCost();

                            // Calculate new Iy
                            insertS = max(
                                match - scoreFunction_.gapOpenCost(),
                                insertS - scoreFunction_.gapExtendCost()
                            );
                        }
                    } else {
                        if (insertS > *insertQrow) {
                            // insertS is largest
                            *bestRow = insertS;

                            // Calculate new Ix
                            *insertQrow = max(
                                match - scoreFunction_.gapOpenCost(),
                                *insertQrow - scoreFunction_.gapExtendCost()
                            );

                            // Calculate new Iy
                            insertS -= scoreFunction_.gapExtendCost();
                        } else {
                            // insertQ is largest
                            *bestRow = *insertQrow;

                            // Calculate new Ix
                            *insertQrow -= scoreFunction_.gapExtendCost();

                            // Calculate new Iy
                            insertS = max(
                                match - scoreFunction_.gapOpenCost(),
                                insertS - scoreFunction_.gapExtendCost()
                            );
                        }
                    }

                    // If score at current cell (and cells to its right) are
                    // below dropoff
                    if (rightOfDropoff) {
                        if (bestScore > *bestRow + dropOff_) {
                            // Record dropoff position
                            columnDropoff = subjectPosition;
                        } else {
                            // We are left of the column dropoff for this row
                            rightOfDropoff = 0;
                        }
                    }

                    // Reset subjectCount
                    subjectCount = extensionN_;
                }

                --subjectPosition;
                --bestRow;
                --insertQrow;
                --subjectCount;
            }

            // -----SINGLE CELL LEFT OF ROW DROPOFF -----
            if (!(bestScore > previousOldBest + dropOff_) &&
                (subjectPosition >= subject)) {
                // Calculate match value
                match = previousOldBest + matrixRow[*subjectPosition];

                // Set value for best
                *bestRow = max(match, insertS);

                // Calculate new Iy
                insertS = max(
                    match - scoreFunction_.gapOpenCost(),
                    insertS - scoreFunction_.gapExtendCost()
                );
 
                // Set DUMMY values for Ix
                *insertQrow = extensionDummy_;

                --subjectPosition;
                --bestRow;
                --insertQrow;
            }

            // -----CELLS LEFT OF ROW DROPOFF -----
            if (!(bestScore > *(bestRow + 1) + dropOff_)) {
                while (subjectPosition >= subject) {
                    // Set value for Iy and best
                    *bestRow = insertS;
                    insertS = insertS - scoreFunction_.gapExtendCost();

                    // Set DUMMY values for Ix
                    *insertQrow = extensionDummy_;

                    // If score at current cell is below dropoff
                    if (bestScore > *bestRow + dropOff_) {
                        // Stop processing row
                        --subjectPosition;
                        break;
                    }

                    --subjectPosition;
                    --bestRow;
                    --insertQrow;
                }
            }

            // Record dropoff position
            rowDropoff = subjectPosition + 1;

            // Clear insertS for next row
            insertS = extensionDummy_;

            // Reset queryCount
            queryCount = extensionN_;
        }
    }

    queryBegin   = bestQueryPosition - query;
    subjectBegin = bestSubjectPosition - subject;
    return bestScore;
}
 

/*!
    Finds an alignment beginning in the seed position (\a queryOffset, \a
    subjectOffset) and ending in (q, s) where q is in the range
    (\a queryOffset, querySize) and s is in the range
    (\a subjectOffset, subjectSize).
*/
int SemiGappedXDropAligner::findAlignmentAfterSeed(
    unsigned char const * query,
    unsigned char const * subject,
    unsigned int querySize,
    unsigned int subjectSize,
    unsigned int & queryEndOut,
    unsigned int & subjectEndOut
)
{
    if (querySize <= 1 || subjectSize <= 1) {
        queryEndOut = 0;
        subjectEndOut = 0;
        return 0;
    }

    int bestScore = 0;
    int insertS, rowOffset;
    int subjectDistance;
    int oldBest, previousOldBest;
    unsigned char leftOfDropoff;
    int queryCount, subjectCount;

    unsigned char const * startSubjectPosition(subject + 1);
    unsigned char const * subjectPosition(startSubjectPosition);
    unsigned char const * bestSubjectPosition(startSubjectPosition);
    unsigned char const * subjectEnd(subject + subjectSize);

    unsigned char const * queryPosition(query + 1);
    unsigned char const * bestQueryPosition(queryPosition);
    unsigned char const * queryEnd(query + querySize);

    // Set initial row dropoff and column dropoff
    unsigned char const * rowDropoff(subject + subjectSize - 1);
    unsigned char const * columnDropoff(subject);

    // Declare processing rows for storing match, insert-subject and
    // insert-query values if current malloced rows aren't big enough
    if (subjectSize >= rowSizes_) {
        delete [] bestRow_;
        delete [] insertQrow_;
        rowSizes_ = subjectSize << 1;
        bestRow_    = new int[rowSizes_];
        insertQrow_ = new int[rowSizes_];
    }

    // Initialize rows
    int * bestRow(bestRow_ + 1);
    int * insertQrow(insertQrow_ + 1);


    // -----FIRST ROW-----

    // Using first column of query matrix
    ASSERT(*queryPosition <= 0xc, "Invalid symbol in query");
    int const * matrixRow(scoreFunction_[*queryPosition]);

    // -----FIRST CELL-----
    // Set M value for top-left cell
    ASSERT(*subjectPosition <= 0xc, "Invalid symbol in subject");
    int match(matrixRow[*subjectPosition]);

    // M must be the best
    *bestRow = match;

    // Only gap opens possible
    *insertQrow = insertS = match - scoreFunction_.gapOpenCost();

    // If this is the best-yet scoring cell
    if (match > bestScore) {
        // Update best start cell data
        bestScore = match;
        bestQueryPosition = queryPosition;
        bestSubjectPosition = subjectPosition;
    }

    subjectDistance = 0;
    ++subjectPosition;
    ++bestRow;
    ++insertQrow;

    // ----- REMAINING CELLS -----
    // For each remaining columns in the top row, scanning from left-to-right
    while (subjectPosition < subjectEnd) {
        // Set value for M
        match = matrixRow[*subjectPosition] -
                scoreFunction_.gapOpenCost() - 
                scoreFunction_.gapExtendCost() * subjectDistance;

        // Determine the best of M and Iy
        if (match > insertS) {
            *bestRow = match;

            // Calculate new Iy
            insertS = max(
                match - scoreFunction_.gapOpenCost(),
                insertS - scoreFunction_.gapExtendCost()
            );
        } else {
            *bestRow = insertS;

            // Since M <= Iy, new Iy must derive from Iy
            insertS -= scoreFunction_.gapExtendCost();
        }

        // Set DUMMY Ix value, which should never be used
        *insertQrow = extensionDummy_;

        // If this is the best-yet scoring cell
        if (match > bestScore) {
            // Update best start cell data
            bestScore = match;
            bestQueryPosition = queryPosition;
            bestSubjectPosition = subjectPosition;
        }

        // If score at current cell is below dropoff
        if (bestScore > *bestRow + dropOff_) {
            // Record dropoff position
            rowDropoff = subjectPosition;
            // And stop processing row
            break;
        }

        ++subjectPosition;
        ++bestRow;
        ++insertQrow;
        ++subjectDistance;
    }

    // Start queryCount at N. Only allow insertS for every Nth row when
    // queryCount reaches 0
    queryCount = extensionN_;

    ++queryPosition;
    --queryCount;

    // -----REMAINING ROWS-----
    while (queryPosition < queryEnd && rowDropoff > columnDropoff) {
        subjectPosition = columnDropoff + 1;

        // Determine subjectCount for initial subjectPosition. Is used to only
        // allow insertQ when subjectOffset % extensionN_ == 0
        subjectCount = (int)(subjectPosition - startSubjectPosition) % extensionN_;
        if (subjectCount) {
            subjectCount = extensionN_ - subjectCount;
        }

        // Reset rows
        rowOffset  = subjectPosition - subject;
        bestRow    = bestRow_ + rowOffset;
        insertQrow = insertQrow_ + rowOffset;

        // Using next column of query matrix
        matrixRow = scoreFunction_[*queryPosition];

        // ************ All rows we are not allowing insertS
        if (queryCount) {
            // ** No insertQ allowed this column, this cell will only get a
            // DUMMY score
            if (subjectCount) {
                previousOldBest = *bestRow;
                *bestRow = extensionDummy_;

                // Score at this cell is below dropoff
                columnDropoff = subjectPosition;
                leftOfDropoff = 1;
            } else {
                // ** We are allowing insertQ this column
                // -----FAR LEFT CELL-----
                // Record some old values
                previousOldBest = *bestRow;

                // Set Ix value
                *bestRow = *insertQrow;
                *insertQrow -= scoreFunction_.gapExtendCost();

                // If score at current cell is below dropoff
                if (bestScore > *bestRow + dropOff_) {
                    // Record dropoff position
                    columnDropoff = subjectPosition;
                    leftOfDropoff = 1;
                } else {
                    // We are right of the column dropoff for this row
                    leftOfDropoff = 0;
                }

                // Reset subjectCount
                subjectCount = extensionN_;
            }

            ++subjectPosition;
            ++bestRow; 
            ++insertQrow;
            --subjectCount;

            // -----CELLS LEFT OF ROW DROPOFF-----
            while (subjectPosition <= rowDropoff) {
                // ** We are not allowing insertQ this column
                if (subjectCount) {
                    // Calculate new M value, which is also the best
                    oldBest = *bestRow;
                    match = matrixRow[*subjectPosition] + previousOldBest;
                    *bestRow = match;
                    previousOldBest = oldBest;

                    // If this is the best-yet scoring cell
                    if (match > bestScore) {
                        // Update best start cell data
                        bestScore = match;
                        bestQueryPosition = queryPosition;
                        bestSubjectPosition = subjectPosition;
                    }
                } else {
                    // We are allowing insertQ this column
                    // Calculate new M value
                    oldBest = *bestRow;
                    match = matrixRow[*subjectPosition] + previousOldBest;
                    previousOldBest = oldBest;

                    // Determine the best of M and Ix
                    if (match > *insertQrow) {
                        *bestRow = match;

                        // Calculate new Ix
                        *insertQrow = max(
                            match - scoreFunction_.gapOpenCost(),
                            *insertQrow - scoreFunction_.gapExtendCost()
                        );
                    } else {
                        *bestRow = *insertQrow;

                        // Since M <= Ix, new Ix must derive from Ix
                        *insertQrow -= scoreFunction_.gapExtendCost();
                    }

                    // If this is the best-yet scoring cell
                    if (match > bestScore) {
                        // Update best start cell data
                        bestScore = match;
                        bestQueryPosition = queryPosition;
                        bestSubjectPosition = subjectPosition;
                    }

                    // Reset subjectCount
                    subjectCount = extensionN_;
                }

                ++subjectPosition;
                ++bestRow; 
                ++insertQrow; 
                --subjectCount;
            }

            // -----SINGLE CELL RIGHT OF ROW DROPOFF -----
            if (!(bestScore > previousOldBest + dropOff_) && 
                (subjectPosition < subjectEnd)) {
                // Set value for best
                match = previousOldBest + matrixRow[*subjectPosition];
                *bestRow = match;

                // Set DUMMY values for Ix
                *insertQrow = extensionDummy_;

                if (match + dropOff_ >= bestScore) {
                    // Record dropoff position
                    rowDropoff = subjectPosition;
                }
            }
        } else {
            // ************ Every Nth row we allow insertS
            // -----FAR LEFT CELL-----

            // ** No insertQ allowed this column, this cell will only get a
            // DUMMY score
            if (subjectCount) {
                previousOldBest = *bestRow;
                *bestRow = extensionDummy_;

                // Score at this cell is below dropoff
                columnDropoff = subjectPosition;
                leftOfDropoff = 1;
            } else {
                // ** We are allowing insertQ this column
                // Record some old values
                previousOldBest = *bestRow;

                 // Set Ix value
                *bestRow = *insertQrow;
                *insertQrow -= scoreFunction_.gapExtendCost();

                // Set DUMMY value for Iy, which should never be used
                insertS = extensionDummy_;

                // If score at current cell is below dropoff
                if (bestScore > *bestRow + dropOff_) {
                    // Record dropoff position
                    columnDropoff = subjectPosition;
                    leftOfDropoff = 1;
                } else {
                    // We are right of the column dropoff for this row
                    leftOfDropoff = 0;
                }

                // Reset subjectCount
                subjectCount = extensionN_;
            }

            ++subjectPosition;
            ++bestRow;
            ++insertQrow;
            --subjectCount;

            // -----CELLS LEFT OF ROW DROPOFF-----
            while (subjectPosition <= rowDropoff) {
                // ** We are not allowing insertQ this column
                if (subjectCount) {
                    // Remember old M value (for cell below this one)
                    oldBest = *bestRow;
                    match = matrixRow[*subjectPosition] + previousOldBest;
                    previousOldBest = oldBest;

                    // Determine the best of M and Iy
                    if (match > insertS) {
                        *bestRow = match;

                        // Calculate new Iy
                        insertS = max(
                            match - scoreFunction_.gapOpenCost(),
                            insertS - scoreFunction_.gapExtendCost()
                        );
                    } else {
                        *bestRow = insertS;

                        // Since M <= Iy, new Iy must derive from Iy
                        insertS -= scoreFunction_.gapExtendCost();
                    }

                    // If this is the best-yet scoring cell
                    if (match > bestScore) {
                        // Update best start cell data
                        bestScore = match;
                        bestQueryPosition = queryPosition;
                        bestSubjectPosition = subjectPosition;
                    }

                    // If score at current cell (and cells to its left) are
                    // below dropoff
                    if (leftOfDropoff) {
                        if (bestScore > *bestRow + dropOff_) {
                            // Record dropoff position
                            columnDropoff = subjectPosition;
                        } else {
                            // We are right of the column dropoff for this row
                            leftOfDropoff = 0;
                        }
                    }
                } else {
                    // ** We are allowing insertQ this column
                    // Remember old M value (for cell below this one)
                    oldBest = *bestRow;
                    match = matrixRow[*subjectPosition] + previousOldBest;
                    previousOldBest = oldBest;

                    // Determine the best of M, Ix and Iy
                    if (match > insertS) {
                        if (match > *insertQrow) {
                            // Match is largest
                            *bestRow = match;

                            // Calculate new Ix
                            *insertQrow = max(
                                match - scoreFunction_.gapOpenCost(),
                                *insertQrow - scoreFunction_.gapExtendCost()
                            );

                            // Calculate new Iy
                            insertS = max(
                                match - scoreFunction_.gapOpenCost(),
                                insertS - scoreFunction_.gapExtendCost()
                            );

                            // If this is the best-yet scoring cell
                            if (match > bestScore) {
                                // Update best start cell data
                                bestScore = match;
                                bestQueryPosition = queryPosition;
                                bestSubjectPosition = subjectPosition;
                            }
                        } else {
                            // insertQ is largest
                            *bestRow = *insertQrow;

                            // Calculate new Ix
                            *insertQrow -= scoreFunction_.gapExtendCost();

                            // Calculate new Iy
                            insertS = max(
                                match - scoreFunction_.gapOpenCost(),
                                insertS - scoreFunction_.gapExtendCost()
                            );
                        }
                    } else {
                        if (insertS > *insertQrow) {
                            // insertS is largest
                            *bestRow = insertS;

                            // Calculate new Ix
                            *insertQrow = max(
                                match - scoreFunction_.gapOpenCost(),
                                *insertQrow - scoreFunction_.gapExtendCost()
                            );

                            // Calculate new Iy
                            insertS -= scoreFunction_.gapExtendCost();
                        } else {
                            // insertQ is largest
                            *bestRow = *insertQrow;

                            // Calculate new Ix
                            *insertQrow -= scoreFunction_.gapExtendCost();

                            // Calculate new Iy
                            insertS = max(
                                match - scoreFunction_.gapOpenCost(),
                                insertS - scoreFunction_.gapExtendCost()
                            );
                        }
                    }

                    // If score at current cell (and cells to its left) are
                    // below dropoff
                    if (leftOfDropoff) {
                        if (bestScore > *bestRow + dropOff_) {
                            // Record dropoff position
                            columnDropoff = subjectPosition;
                        } else {
                            // We are right of the column dropoff for this row
                            leftOfDropoff = 0;
                        }
                    }

                    // Reset subjectCount
                    subjectCount = extensionN_;
                }

                ++subjectPosition;
                ++bestRow;
                ++insertQrow; 
                --subjectCount;
            }

            // -----SINGLE CELL RIGHT OF ROW DROPOFF -----
            if (!(bestScore > previousOldBest + dropOff_) && 
                (subjectPosition < subjectEnd)) {
                // Calculate match value
                match = matrixRow[*subjectPosition] + previousOldBest;

                // Set value for best
                *bestRow = max(match, insertS);

                // Calculate new Iy
                insertS = max(
                    match - scoreFunction_.gapOpenCost(),
                    insertS - scoreFunction_.gapExtendCost()
                );

                // Set DUMMY values for Ix
                *insertQrow = extensionDummy_;

                ++subjectPosition;
                ++bestRow;
                ++insertQrow;
            }

            // -----CELLS RIGHT OF ROW DROPOFF -----
            if (!(bestScore > *(bestRow - 1) + dropOff_)) {
                while (subjectPosition < subjectEnd) {
                    // Set value for Iy and best
                    *bestRow = insertS;
                    insertS = insertS - scoreFunction_.gapExtendCost();

                    // Set DUMMY values for Ix
                    *insertQrow = extensionDummy_;

                    // If score at current cell is below dropoff
                    if (bestScore > *bestRow + dropOff_) {
                        // Stop processing row
                        ++subjectPosition;
                        break;
                    }

                    ++subjectPosition;
                    ++bestRow;
                    ++insertQrow;
                }
            }

            // Record dropoff position
            rowDropoff = subjectPosition - 1;

            // Clear insertS for next row
            insertS = extensionDummy_;

            // Reset queryCount
            queryCount = extensionN_;
        }

        ++queryPosition;
        --queryCount;
    }

    queryEndOut   = bestQueryPosition - query;
    subjectEndOut = bestSubjectPosition - subject;
    return bestScore;
}


} // namespace swift

