// sequence.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
// 

#include "sequence.hpp"
#include "encoding.hpp"

namespace swift {


size_t const Sequence::npos = size_t(-1);


/*!
    Construct an empty Sequence object.
*/
Sequence::Sequence() :
    buffer_(0),
    begin_(0),
    end_(0),
    shareable_(true)
{
}


/*!
    Construct a Sequence object of length \a size, beginning at position
    \a offset in \a buffer.
*/
Sequence::Sequence(
    Buffer const & buffer,
    size_type size,
    size_type offset
) :
    buffer_(buffer),
    begin_(buffer_.get() + offset),
    end_(begin_ + size),
    shareable_(true)
{
}


/*!
    Construct a Sequence object from another sequence \a other, using
    the range [\a pos, \a pos + \a n).
*/
Sequence::Sequence(
    Sequence const & other,
    size_type pos,
    size_type n
) :
    buffer_(0),
    begin_(0),
    end_(0),
    shareable_(true)
{
    if (pos > other.size())
        throw std::out_of_range("position out of range");

    n = std::min(n, other.size() - pos);

    if (other.shareable_) {
        buffer_ = other.buffer_;
        begin_  = other.begin_ + pos;
        end_    = begin_ + n;
    } else {
        cloneBuffer(other.begin_ + pos, other.begin_ + pos + n);
    }
}


/*!
    Constructs a Sequence object from a the string \a string of
    size \a size.
*/
Sequence::Sequence(char const * string, size_type size) :
    buffer_(new value_type[size]),
    begin_(buffer_.get()),
    end_(begin_ + size),
    shareable_(true)
{
    for (size_t i(0); i < size; ++i){
        begin_[i] = Encoding::ToValue[string[i]];
}
}


/*!
    Reverse-complements the sequence.
TODO:probable error in case of (un)even length of sequence (in the middle)
0---(=)---7
*/
void Sequence::reverseComplement()
{
    accessCow();

    int halfLength(size() >> 1);
    pointer first(begin_);
    pointer last(end_);

    while (--halfLength >= 0) {
        char c(Encoding::Complement[*first]);
        *first++ = Encoding::Complement[*--last];
        *last    = c;
    }
}


// --------------------------------------------------------------------------

} // namespace swift



