// sort_permute_iterator.hpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//  

#ifndef _SWIFT_SORT_PERMUTE_ITERATOR_
#define _SWIFT_SORT_PERMUTE_ITERATOR_

#include <boost/iterator/iterator_facade.hpp>
#include <boost/tuple/tuple.hpp>
#include <iterator>


namespace swift
{


template <
    class SortIterator,
    class PermuteIterator
>
struct sort_permute_iterator_dereferencer
{
    typedef boost::tuple<
        typename std::iterator_traits<SortIterator>::value_type,
        typename std::iterator_traits<PermuteIterator>::value_type
    > value_type;

    typedef boost::tuple<
        typename std::iterator_traits<SortIterator>::value_type &,
        typename std::iterator_traits<PermuteIterator>::value_type &
    > ref_type;
};


/*!
    sort_permute_iterator is a partial solution to the problem of
    sorting two sequences simultaneously. E.g. given two arrays of 
    equal length, we wish to sort the first and keep the second synced.
    In other words, we wish to sort a sequence of pairs by their
    first member.

    boost::zip_iterator allows processings iterator ranges in parallel,
    but it is unfortunately not compatible with std::sort().

    sort_permute_iterator abuses the iterator notation somewhat, and is
    therefore not conforming. It works with std::sort(), however.

    The template parameters are:

    \li \c SortIterator Iterator type for the sorted sequence.
    \li \c PermuteIterator Iterator type for the permuted sequence.
*/
template <
    class SortIterator,
    class PermuteIterator
>
class sort_permute_iterator :
    public boost::iterator_facade<
        sort_permute_iterator<SortIterator, PermuteIterator>,
        typename sort_permute_iterator_dereferencer<
            SortIterator, PermuteIterator
        >::value_type,
        std::random_access_iterator_tag,
        typename sort_permute_iterator_dereferencer<
            SortIterator, PermuteIterator
        >::ref_type,
        typename std::iterator_traits<SortIterator>::difference_type
    >
{
    public:
    typedef typename sort_permute_iterator<
        SortIterator,
        PermuteIterator
    >::difference_type difference_type;

    sort_permute_iterator() {}

    sort_permute_iterator(SortIterator si, PermuteIterator pi) :
        si_(si), pi_(pi)
    {
    }


    private:
    typedef typename sort_permute_iterator_dereferencer<
        SortIterator,
        PermuteIterator
    >::ref_type deref_type;

    void increment()
    {
        ++si_;
        ++pi_;
    }

    void decrement()
    {
        --si_;
        --pi_;
    }

    bool equal(sort_permute_iterator const & other) const
    {
        return si_ == other.si_;
    }

    deref_type dereference() const
    {
        return deref_type(*si_, *pi_);
    }

    void advance(difference_type n)
    {
        si_ += n;
        pi_ += n;
    }

    difference_type distance_to(sort_permute_iterator const & other) const
    {
        return other.si_ - si_;
    }

    SortIterator si_;
    PermuteIterator pi_;
        
    friend class boost::iterator_core_access;
};


template <
    class SortIterator,
    class PermuteIterator
>
struct sort_permute_iterator_compare :
    public std::binary_function<
        typename sort_permute_iterator_dereferencer<
            SortIterator,
            PermuteIterator
        >::value_type,
        typename sort_permute_iterator_dereferencer<
            SortIterator,
            PermuteIterator
        >::value_type,
        bool
    >
{
    typedef typename sort_permute_iterator_dereferencer<
        SortIterator,
        PermuteIterator
    >::value_type T;

    bool operator()(T const & t1, T const & t2) const
    {
        return boost::get<0>(t1) < boost::get<0>(t2);
    }
};


/*!
    Returns a sort_permute_iterator for the given domain iterators.
    This is a convenience method that is equivalent to
    <tt>sort_permute_iterator<SortIterator, PermuteIterator>(si, pi)</tt>
    where \c si and \c pi are the domain iterators of type \c SortIterator
    and \c PermuteIterator , respectively.
*/
template <
    class SortIterator,
    class PermuteIterator
>
sort_permute_iterator<SortIterator, PermuteIterator>
make_sort_permute_iterator(SortIterator si, PermuteIterator pi)
{
    return sort_permute_iterator<SortIterator, PermuteIterator>(si, pi);
}


} // namespace swift

#endif

