// swiftdatabase.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

#include "swiftdatabase.hpp"
#include "swiftdatabasebase.hpp"
#include "idmap.hpp"
#include <boost/shared_ptr.hpp>
#include <boost/scoped_array.hpp>
#include <boost/iostreams/device/array.hpp>
#include <boost/iostreams/device/mapped_file.hpp>
#include <boost/iostreams/stream_buffer.hpp>
#include <algorithm>
#include <cstdio>

namespace swift {
namespace detail {

// SwiftDatabaseImpl class declaration --------------------------------------

struct SwiftDatabaseImpl : private SwiftDatabaseBase
{
    typedef SwiftDatabase::size_type size_type;

    SwiftDatabaseImpl();

    size_type size() const;
    size_type count() const;

    std::string const & name() const;

    std::string text(size_type n) const;

    Sequence * sequence(
        size_type n,
        Sequence::size_type begin,
        Sequence::size_type end
    ) const;

    size_type offset(size_type n) const;
    Sequence::size_type size(size_type n) const;

    size_type find(size_type position, size_type index) const;
    size_type find(std::string const & key) const;

    static boost::shared_ptr<SwiftDatabaseImpl> open(
        std::string const & name
    );

    char * loadData(FILE * file, long offset, size_t size) const;
    char * loadData(FILE * file, long offset, size_t size, char * buffer) const;

    FILE * textFile_;
    FILE * sequenceFile_;

    OffsetVector textOffset_;
    OffsetVector sequenceOffset_;

    boost::shared_ptr<IdMap> sequenceIdMap_; ///< Sequence id map

    std::string name_;

    mutable boost::scoped_array<char> textBuffer_;
    unsigned int textBufferSize_;
};

// SwiftDatabaseImpl class definition ---------------------------------------

/*!
    Constructs a SwiftDatabaseImpl object.
*/
SwiftDatabaseImpl::SwiftDatabaseImpl() :
    textBuffer_(0),
    textBufferSize_(0)
{
}


/*!
    \internal

    Returns the size of the database. This is equivalent to the sum of the
    sizes of all sequences in the database.
*/
SwiftDatabaseImpl::size_type SwiftDatabaseImpl::size() const
{
    return sequenceOffset_.back();
}


/*!
    \internal

    Returns the number of sequences in the database.
*/
SwiftDatabaseImpl::size_type SwiftDatabaseImpl::count() const
{
    return sequenceOffset_.size() - 1;
}


/*!
    \internal

    Returns the index of the sequence that contains position \a position,
    or \l count() if no such sequence exists. The argument \a index
    is used as hint to where the position may be found. It must be
    in the closed range [0, count()].
*/
SwiftDatabaseImpl::size_type
SwiftDatabaseImpl::find(size_type position, size_type index) const
{
    OffsetVector::const_iterator h(sequenceOffset_.begin() + index);
    OffsetVector::const_iterator i(sequenceOffset_.begin());
    OffsetVector::const_iterator j(sequenceOffset_.end());

    if (*h > position) {
        j = h;
    } else {
        i = h;
    }

    OffsetVector::const_iterator p(std::upper_bound(i, j, position));
    --p;

    return (p - sequenceOffset_.begin());
}


/*!
    \internal

    Returns the index of the sequence with key \a key, or the value of
    \l count() if no such sequence exists.
*/
SwiftDatabaseImpl::size_type
SwiftDatabaseImpl::find(std::string const & key) const
{
    std::pair<bool, uint32> result(sequenceIdMap_->find(key));
    return (result.first ? result.second : count());
}


/*!
    \internal

    Opens the database with name \a name. A SwiftDatabaseImpl object
    is returned if the opening was successful, otherwise a null pointer
    is returned.
*/
boost::shared_ptr<SwiftDatabaseImpl>
SwiftDatabaseImpl::open(std::string const & name)
{
    typedef boost::iostreams::array_source ArraySource;
    typedef boost::iostreams::stream_buffer<ArraySource> ArrayStreamBuffer;
    typedef boost::iostreams::mapped_file_source MappedFile;

    boost::shared_ptr<SwiftDatabaseImpl> nullPimpl;
    std::string baseName(name.substr(0, name.rfind(IndexSuffix)));

    // Map header
    MappedFile indexMap;
    try {
        indexMap.open(name + IndexSuffix);
    } catch (std::ios::failure & e) {
        return nullPimpl;
    }
    if (!indexMap.is_open() || indexMap.size() < sizeof(IndexHeader))
        return nullPimpl;

    IndexHeader const * header(
        reinterpret_cast<IndexHeader const *>(indexMap.data())
    );
    uint32 offsetCount(header->sequenceCount + 1);
    uint32 headerSize(sizeof(IndexHeader) + 2 * sizeof(uint32) * offsetCount);

    boost::shared_ptr<SwiftDatabaseImpl> pimpl(new SwiftDatabaseImpl);
    //todo differenciate linux (/) windows (\) mac?
    std::string::size_type p(baseName.rfind('/'));
    pimpl->name_ = baseName.substr(p == std::string::npos ? 0 : p + 1);

    // Load offsets
    pimpl->textOffset_.reserve(offsetCount);
    pimpl->sequenceOffset_.reserve(offsetCount);

    uint32 const * i(
        reinterpret_cast<uint32 const *>(indexMap.data() +
        sizeof(IndexHeader))
    );
    uint32 const * j(i + offsetCount);
    uint32 const * k(j + offsetCount);

    pimpl->textOffset_.insert(pimpl->textOffset_.begin(), i, j);
    pimpl->sequenceOffset_.insert(pimpl->sequenceOffset_.begin(), j, k);

    // Load id map
    ArraySource as(indexMap.data() + headerSize,
                   indexMap.size() - headerSize);
    ArrayStreamBuffer asb(as);
    std::istream is(&asb);
    IdMapFactory::Source idMap;
    pimpl->sequenceIdMap_.reset(idMap.read(is));

    // Open text and sequence files
    pimpl->textFile_ = fopen((baseName + TextSuffix).c_str(), "rb");
    pimpl->sequenceFile_ = fopen((baseName + SequenceSuffix).c_str(), "rb");
    if (!pimpl->textFile_ || !pimpl->sequenceFile_)
        return nullPimpl;

    return pimpl;
}


/*!
    \internal

    Loads \a size bytes from position \a offset in file \a file. If the data
    was successfully loaded, a pointer to the allocated memory is returned.
    If an error occurs, a null pointer is returned.
*/
char * SwiftDatabaseImpl::loadData(
    FILE * file,
    long offset,
    size_t size
) const
{
    // FIXME - must get offset types "synchronized"

    long currentOffset(ftell(file));

    if (currentOffset < 0)
        return 0;

    if (fseek(file, offset - currentOffset, SEEK_CUR) < 0)
        return 0;

    char * data(new char[size]);
    if (fread(data, sizeof(char), size, file) < size) {
        delete [] data;
        return 0;
    }

    return data;
}


/*!
    \internal

    Loads \a size bytes from position \a offset in file \a file to the buffer
    \a buffer. It assumed that the buffer is large enough to hold \a size
    bytes. If an error occurs, a null pointer is returned.
*/
char * SwiftDatabaseImpl::loadData(
    FILE * file,
    long offset,
    size_t size,
    char * buffer
) const
{
    // FIXME - must get offset types "synchronized"

    long currentOffset(ftell(file));

    if (currentOffset < 0)
        return 0;

    if (fseek(file, offset - currentOffset, SEEK_CUR) < 0)
        return 0;

    if (fread(buffer, sizeof(char), size, file) < size)
        return 0;

    return buffer;
}


/*!
    \internal

    Returns the text associated with the \a n th sequence. It is
    assumed that \a n is a valid sequence number.
*/
std::string SwiftDatabaseImpl::text(size_type n) const
{
    size_t textSize(textOffset_[n + 1] - textOffset_[n]);

    if (textBufferSize_ < textSize) {
        textBuffer_.reset(new char[textSize << 1]);
    }

    return std::string(
        loadData(textFile_, textOffset_[n], textSize, textBuffer_.get())
    );
}


/*!
    \internal

    Returns the substring [\a begin, \a end) of the \a n th sequence.
    It is assumed that \a n is a valid sequence number and that the
    substring range is valid.
*/
Sequence * SwiftDatabaseImpl::sequence(
    size_type n,
    Sequence::size_type begin,
    Sequence::size_type end
) const
{
    long offset(sequenceOffset_[n] + begin);
    size_t sz(std::min(end, size(n)) - begin);
    Sequence::Buffer buffer(
        (unsigned char *)loadData(sequenceFile_, offset, sz)
    );
    return new Sequence(buffer, sz);
}


/*!
    \internal

    Returns the offset for the \a n th sequence. It is assumed that
    \a n is a valid sequence number.
*/
SwiftDatabaseImpl::size_type
SwiftDatabaseImpl::offset(size_type n) const
{
    return sequenceOffset_[n];
}


/*!
    \internal

    Returns the size of the \a n th sequence. It is assumed that
    \a n is a valid sequence number.
*/
Sequence::size_type
SwiftDatabaseImpl::size(size_type n) const
{
    return sequenceOffset_[n + 1] - sequenceOffset_[n];
}


/*!
    \internal

    Returns the basename for the database currently opened.
*/
std::string const & SwiftDatabaseImpl::name() const
{
    return name_;
}

} // namespace detail

// SwiftDatabase class definition -------------------------------------------


/*!
    Constructs a SwiftDatabase object.
*/
SwiftDatabase::SwiftDatabase()
{
}


/*!
    Constructs a SwiftDatabase object for the database with name \a name.
*/
SwiftDatabase::SwiftDatabase(std::string const & name)
{
    open(name);
}


/*!
    Opens the database with name \a name. A previously opened database will
    be closed.

    \sa close()
*/
void SwiftDatabase::open(std::string const & name)
{
    pimpl_ = ImplType::open(name);
}


/*!
    Closes the database.

    \sa open()
*/
void SwiftDatabase::close()
{
    if (isOpen()) {
        pimpl_.reset();
    }
}


/*!
    Returns true if the database is open for access.

    \sa open()
*/
bool SwiftDatabase::isOpen() const
{
    return (!!pimpl_);
}


/*!
    Returns the text associated with the \a n th sequence.
*/
std::string SwiftDatabase::text(size_type n) const
{
    return pimpl_->text(n);
}


Sequence * SwiftDatabase::sequence(
    size_type n,
    Sequence::size_type begin,
    Sequence::size_type end
) const
{
    return pimpl_->sequence(n, begin, end);
}


Sequence::size_type SwiftDatabase::offset(size_type n) const
{
    return pimpl_->offset(n);
}


Sequence::size_type SwiftDatabase::size(size_type n) const
{
    return pimpl_->size(n);
}


SwiftDatabase::size_type SwiftDatabase::size() const
{
    return pimpl_->size();
}


SwiftDatabase::size_type SwiftDatabase::count() const
{
    return pimpl_->count();
}


/*!
    Returns the name of the database.

    \sa open()
*/
std::string const & SwiftDatabase::name() const
{
    return pimpl_->name();
}


/*!
    Returns the index for the sequence with key \a key, or \l count()
    if no such sequence exists.
*/
SwiftDatabase::size_type SwiftDatabase::find(std::string const & key) const
{
    return pimpl_->find(key);
}


SwiftDatabase::size_type SwiftDatabase::find(
    size_type position,
    size_type index
) const
{
    return pimpl_->find(position, index);
}

// --------------------------------------------------------------------------


} // namespace swift

