// timer.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
// 

#include "timer.hpp"
#include <sstream>
#include <cstdlib>

namespace swift {


/**
 * Constructor.
 *
 * @param useUSecPrecision      If true, usec precision is used.
 */
Timer::Timer(bool useUSecPrecision /* = true */) :
    isRunning_(false),
    useUSecPrecision_(useUSecPrecision),
    userTime_(0),
    systemTime_(0),
    realTime_(0)
{
}


/**
 * Destructor
 */
Timer::~Timer()
{
}


/**
 * Starts the timer
 */
void Timer::start()
{
    if (isRunning_)
        return;

    isRunning_ = true;
    timeBegin_ = time(0);

    if (getrusage(RUSAGE_SELF, &usageBegin_) == -1) {
        std::cerr << "getrusage() failed." << std::endl;
        exit(-1);
    }
}


/**
 * Stops the timer
 */
void Timer::stop()
{
    if (!isRunning_)
        return;

    isRunning_ = false;
    timeEnd_ = time(0);

    if (getrusage(RUSAGE_SELF, &usageEnd_) == -1) {
        std::cerr << "getrusage() failed." << std::endl;
        exit(-1);
    }

    long utime(usageEnd_.ru_utime.tv_sec - usageBegin_.ru_utime.tv_sec);
    long stime(usageEnd_.ru_stime.tv_sec - usageBegin_.ru_stime.tv_sec);
    long rtime(timeEnd_ - timeBegin_);
    
    if (useUSecPrecision_) {
        utime = utime * 1000000 +
                usageEnd_.ru_utime.tv_usec - usageBegin_.ru_utime.tv_usec;
        stime = stime * 1000000 +
                usageEnd_.ru_stime.tv_usec - usageBegin_.ru_stime.tv_usec;
        rtime = rtime * 1000000;
    }
    
    userTime_ += utime;
    systemTime_ += stime;
    realTime_ += rtime;
}


/**
 * Resets the timer.
 */
void Timer::reset()
{
    userTime_ = 0;
    systemTime_ = 0;
    realTime_ = 0;

    if (isRunning_) {
        timeBegin_ = time(0);
        if (getrusage(RUSAGE_SELF, &usageBegin_) == -1) {
            std::cerr << "getrusage() failed." << std::endl;
            exit(-1);
        }
    }
}


/**
 * Returns the time elapsed since the timer was started.
 */
TimeElapsed Timer::getTimeElapsed()
{
    if (!isRunning_)
        return TimeElapsed(userTime_, systemTime_, realTime_);

    timeEnd_ = time(0);
    if (getrusage(RUSAGE_SELF, &usageEnd_) == -1) {
        std::cerr << "getrusage() failed." << std::endl;
        exit(-1);
    }

    long utime(usageEnd_.ru_utime.tv_sec - usageBegin_.ru_utime.tv_sec);
    long stime(usageEnd_.ru_stime.tv_sec - usageBegin_.ru_stime.tv_sec);
    long rtime(timeEnd_ - timeBegin_);

    if (useUSecPrecision_) {
        utime = utime * 1000000 +
                usageEnd_.ru_utime.tv_usec - usageBegin_.ru_utime.tv_usec;
        stime = stime * 1000000 +
                usageEnd_.ru_stime.tv_usec - usageBegin_.ru_stime.tv_usec;
        rtime = rtime * 1000000;
    }

    return TimeElapsed(
        userTime_ + utime,
        systemTime_ + stime,
        realTime_ + rtime);
}


/**
 * Returns true if the timer has been started.
 */
bool Timer::isRunning()
{
    return isRunning_;
}


/**
 * Constructor
 */
TimeElapsed::TimeElapsed() :
    userTime_(0),
    systemTime_(0),
    realTime_(0)
{
}


/**
 * Constructor
 */
TimeElapsed::TimeElapsed(
    long userTime,
    long systemTime,
    long realTime) :
    userTime_(userTime),
    systemTime_(systemTime),
    realTime_(realTime)
{
}


/**
 * Copy constructor
 */
TimeElapsed::TimeElapsed(const TimeElapsed& timeElapsed)
{
    userTime_ = timeElapsed.userTime_;
    systemTime_ = timeElapsed.systemTime_;
    realTime_ = timeElapsed.realTime_;    
}


/**
 * Destructor
 */
TimeElapsed::~TimeElapsed()
{
}


/**
 * Assignment operator
 */
TimeElapsed& TimeElapsed::operator=(const TimeElapsed& source)
{
    userTime_ = source.userTime_;
    systemTime_ = source.systemTime_;
    realTime_ = source.realTime_;    

    return *this;
}


/**
 * Adding assignment operator
 */
const TimeElapsed& TimeElapsed::operator+=(const TimeElapsed& rhs)
{
    userTime_ += rhs.userTime_;
    systemTime_ += rhs.systemTime_;
    realTime_ += rhs.realTime_;

    return *this;
}


/**
 * Subtracting assignment operator
 */
const TimeElapsed& TimeElapsed::operator-=(const TimeElapsed& rhs)
{
    userTime_ -= rhs.userTime_;
    systemTime_ -= rhs.systemTime_;
    realTime_ -= rhs.realTime_;

    return *this;
}


/**
 * Addition operator
 */
TimeElapsed operator+(const TimeElapsed& lhs, const TimeElapsed& rhs)
{
    TimeElapsed t(lhs);
    return t += rhs;
}


/**
 * Subtraction operator
 */
TimeElapsed operator-(const TimeElapsed& lhs, const TimeElapsed& rhs)
{
    TimeElapsed t(lhs);
    return t -= rhs;
}


/**
 * Operator for displaying elapsed time
 */
std::ostream& operator<<(std::ostream& outStream, TimeElapsed& timeElapsed)
{
    outStream
        << "Real time         : " << timeToString(timeElapsed.realTime_) 
        << "\nCPU-time (user)   : " << timeToString(timeElapsed.userTime_)
        << "\nCPU-time (kernel) : " << timeToString(timeElapsed.systemTime_)
        << std::endl;

    return outStream;
}


std::string timeToString(long time)
{
    static std::ostringstream buf;
    static unsigned int minutes;
    static unsigned int seconds;
    static unsigned int miliseconds;
    
    buf.str("");
    minutes = time / 1000000 / 60;
    seconds = time / 1000000 - minutes * 60;
    miliseconds = (time - seconds * 1000000) / 1000;

    buf << minutes << 'm' << seconds << '.' << miliseconds << 's';

    return buf.str();
}


} // namespace swift

