// createcommand.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
//
//
// Additions / changes by Peter Husemann
//

#include "createcommand.hpp"
#include "fastastreamiterator.hpp"
#include "sequence.hpp"
#include "swiftdatabasebuilder.hpp"
#include <fstream>

namespace swift
{

/*!
    Constructs a CreateCommand object.
*/
CreateCommand::CreateCommand() :
    visibleOptions_("Valid options")
{
    namespace po = boost::program_options;

    visibleOptions_.add_options()
        ("help,h", "produces this help message.")
        ("verbose,v", "print extra information.");

    hiddenOptions_.add_options()
        ("database", po::value<std::string>(&databasePrefix_),
        "Database prefix")
        ("input", po::value<NamesVector>(&inputFiles_),
        "Input files");
}


void CreateCommand::execute(int argc, char ** argv)
{
  //SwiftCommand::SwiftExecutable = argv[0];
  if (processArguments(argc, argv))
        createDatabase();
}


bool CreateCommand::processArguments(int argc, char ** argv)
{
    namespace po = boost::program_options;

    inputFiles_.clear();

    try {
        po::positional_options_description positionals;
        positionals.add("database", 1);
        positionals.add("input", -1);

        po::options_description all;
        all.add(visibleOptions_).add(hiddenOptions_);

        po::variables_map vm;
        po::store(po::command_line_parser(argc, argv).
                  options(all).positional(positionals).run(), vm);
        po::notify(vm);

        if (vm.count("help") || vm.count("input") < 1) {
            displayUsage();
            return false;
        }

    } catch (po::unknown_option & e) {
      std::cerr << SwiftCommand::SwiftExecutable <<": " << e.what() << "\n"
            << "Type '"<< SwiftCommand::SwiftExecutable <<" help' for usage information."
                  << std::endl;
        return false;
    } catch (std::exception & e) {
        std::cerr << e.what() << std::endl;
        return false;
    }

    return true;
}


void CreateCommand::displayUsage(std::ostream & outStream) const
{
    outStream <<
        "Usage:\n"
        "  "<< SwiftCommand::SwiftExecutable <<" create [options] DATABASE [FASTA...]\n"
        "\n"
        "  Preprocesses multi-FASTA files into a SWIFT database.  To query\n"
        "  the database an index must be built with the 'index' command.\n"
        "\n"
        "  The constructed database consists of three files with name\n"
        "  prefix DATABASE and the following suffices:\n"
        "\n"
        "    .swi   -   database meta-information\n"
        "    .sws   -   sequence data\n"
        "    .swt   -   sequence descriptions\n"
        "\n"
        "  Examples:\n"
        "    "<< SwiftCommand::SwiftExecutable <<" create mydatabase myfasta1.fa myfasta2.fa\n"
        "\n"
        "  After Creating the tatabease it is likely that you want to\n"
        "  create an index. See:\n"
        "    index\n"
        "\n"
        << visibleOptions_ << std::endl;
}


void CreateCommand::createDatabase()
{
    SwiftDatabaseBuilder builder;

    builder.setName(databasePrefix_);

    NamesVector::const_iterator i(inputFiles_.begin());
    NamesVector::const_iterator j(inputFiles_.end());

    for ( ; i != j; ++i) {
        std::ifstream fastaStream((*i).c_str());
        if (!fastaStream.is_open()) {
            std::cout << "-- Cannot open file '" << *i << "'. "
                         "Skipping." << std::endl;
            continue;
        }

        for (FastaStreamIterator ii(fastaStream), jj; ii != jj; ++ii) {
            FastaSequence const & fasta(*ii);

            // FIXME  1148220183
            // THIS STINKS; it will reallocate for each sequence
            Sequence sequence(fasta.sequence.c_str(), fasta.sequence.size());
            //Sequence sequence;

            builder.addSequence(
                sequence, // FIXME
                fasta.defline,
                extractId(fasta.defline)
            );
        }

        fastaStream.close();
    }
}


/*!
    \internal

    Extracts a sequence id from a Fasta \a defline. The id is assumed to
    be the substring [0, p), where p denotes the position of the first
    white-space. If no such position exists, the returned id is empty.
*/
std::string CreateCommand::extractId(std::string const & defline) const
{
    std::string::size_type p(defline.find_first_of(' '));
    if (p != std::string::npos)
        return defline.substr(0, p);

    return std::string();
}


} // namespace swift
