// epsilonfilter.hpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the 
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

#ifndef _SWIFT_EPSILONFILTER_
#define _SWIFT_EPSILONFILTER_

#include "gramindex.hpp"
#include "hashset.hpp"
#include <boost/utility.hpp>
#include <vector>
#include <iosfwd>


namespace swift {


class EpsilonFilter : public boost::noncopyable
{
    public:
    struct Match
    {
        Match(int ss, unsigned int qs, unsigned int qe);

        unsigned int subjectEnd(unsigned int matchWidth) const;

        int          subjectStart;
        unsigned int queryStart;
        unsigned int queryEnd;
    };

    typedef std::vector<Match> Matches;

    struct MatchProperties
    {
        /*!
            Returns the minimum match size in query sequence.
        */
// todo/fixme
// cases can occur, where a match is smaller than the requested size
        unsigned int minMatchSize() const
        {
            return minHits + gramSize - 1;
        }

        /*!
            Match width.
        */
        unsigned int width;

        /*!
            Maximum gap between q-grams in epsilon match in zone.
        */
        unsigned int maxGap;

        /*!
            Minimum number of q-grams in epsilon match in zone.
        */
        unsigned int minHits;

        /*!
            Size of q-grams in match.
        */
        unsigned int gramSize;
    };

    private:
    /*!
        Bin structure.
    */
    struct Bin
    {
        /*!
            Position of first hit in bin.
        */
        int first;

        /*!
            Position of last hit in bin.
        */
        int last;

        /*!
            Number of hits in a bin (counting once per position).
        */
        unsigned int count;
    };

    typedef HashSet<
        unsigned int,
        std::equal_to<unsigned int>,
        Hash<unsigned int>,
        FastHashIteratorTag
    > TaggedBins;

    public:
    EpsilonFilter(GramIndex const & gramIndex);
    ~EpsilonFilter();

    template <class GramIterator>
    void operator()(
        GramIterator first,
        GramIterator last,
        Matches & matches
    ) throw();

    /// Configure filter.
    bool configure(
        unsigned int minLength,
        double epsilon,
        unsigned int minBinShift = 0
    ) throw();

    MatchProperties const & matchProperties() const;

    private:
    unsigned int   subjectSize_;
    unsigned int   gramSize_;
    Bin *          bins_;
    unsigned int   binCount_;
    TaggedBins taggedBins_;

    unsigned int binMask_;
    unsigned int binShift_;
    unsigned int binSize_;
    unsigned int minMatchLength_;
    unsigned int zoneWidth_;
    unsigned int threshold_;
    unsigned int trapMin_;
    unsigned int binMax_;
    int ic_;
    int id_;

    MatchProperties matchProperties_;
    GramIndex const & gramIndex_;
};


inline
EpsilonFilter::Match::Match(int ss, unsigned int qs, unsigned int qe) :
    subjectStart(ss),
    queryStart(qs),
    queryEnd(qe)
{
}


/*!
    Returns \c subjectEnd from the range <tt>[subjectStart, subjectEnd)</tt>,
    given the width of the match parallelogram.

    \internal
    Note that due to the stair-case we need to subtract one from the otherwise
    obvious calculation of subjectEnd. 

                             subjectStart    .--- subjectEnd
                                  v          v
                                  012345678901
                    queryStart> 0 OOOOO++++++
                                1 +OOOOO     
                                2 + OOOOO    
                                3 +  OOOOO   
                                4 +   OOOOO  
                                5 +    OOOOO 
                                6 +     OOOOO
                    queryEnd>   7 

*/
inline
unsigned int EpsilonFilter::Match::subjectEnd(unsigned int matchWidth) const
{
    return subjectStart + (queryEnd - queryStart) + matchWidth - 1;
}


inline
EpsilonFilter::MatchProperties const & EpsilonFilter::matchProperties() const
{
    return matchProperties_;
}


/*!
    Finds epsilon zones using the grams in [\a first, \a last) which is
    defined by iterator of type \a GramIterator. Matches are returned in
    \a matches.
*/
template <typename GramIterator>
void EpsilonFilter::operator()(
    GramIterator first,
    GramIterator last,
    Matches & matches
) throw()
{

// hint GramIterator:
//typedef GramView<unsigned char const *> MyGramView;
//typedef MyGramView::const_iterator GramIterator;

    typedef GramIndex::Positions MyPositions;

    unsigned int const binMask = binMask_;
    unsigned int const binShift = binShift_;
    unsigned int const zoneWidth = zoneWidth_;
    unsigned int const threshold = threshold_;
    unsigned int const trapMin = trapMin_;
    unsigned int const binMax = binMax_;

    int ia = subjectSize_;
    int ib = -zoneWidth_;
    int ic = ic_;
    int id = id_;
    
    unsigned int const offset = ic; // Offset from zero for query sequence

    // Loop: hits
    for (; first != last; ++ic, ++id, ++ia, ++first) {
        if ((*first).isValid()) {
            MyPositions pl = gramIndex_.positions((*first).value());
            MyPositions::iterator pi = pl.begin();
            MyPositions::iterator pj = pl.end();

            for (; pi != pj; ++pi) {
                unsigned int e = ia - *pi;
                unsigned int k = e >> binShift;
                Bin * bp(bins_ + (k % binMax));
                
                if (bp->last < id) {
                    if (bp->count >= threshold) {
                        matches.push_back(
                            Match(k << binShift, bp->first, bp->last)
                        );
                        taggedBins_.erase(k);
                    }
                    bp->count = 0;
                }
                if (bp->count == 0)
                    bp->first = ic;

                if (bp->last < ic) {
                    bp->last = ic;
                    if (++bp->count == threshold) {
                        taggedBins_.insert(k);
                    }
                }

                // Check for double bin hit
                if ((e & binMask) < zoneWidth) {
                    if (bp == bins_)
                        bp = bins_ + (binMax - 1);
                    else
                        --bp;

                    if (bp->last < id) {
                        if (bp->count >= threshold) {
                            matches.push_back(
                                Match((k-1) << binShift, bp->first, bp->last)
                            );
                            taggedBins_.erase(k - 1);
                        }
                        bp->count = 0;
                    }
                    if (bp->count == 0)
                        bp->first = ic;
                    
                    if (bp->last < ic) {
                        bp->last = ic;
                        if (++bp->count == threshold) {
                            taggedBins_.insert(k - 1);
                        }
                    }
                }
            }
        }

        if (static_cast<unsigned int>(ib++) == binMask) {
            ib = 0;

            unsigned int k((ic - zoneWidth) >> binShift);
            Bin * bp(bins_ + (k % binMax));
            
            if (bp->count >= threshold) {
                matches.push_back(Match(k << binShift, bp->first, bp->last));
                taggedBins_.erase(k);
            }
            bp->count = 0;
            bp->last = 0;
        }
    }
    
    // Collect remaining regions from tagged bins
    typedef TaggedBins::const_iterator BinIter;
    for (BinIter i(taggedBins_.begin()), j(taggedBins_.end()); i != j; ++i) {
        unsigned int k(*i);
        Bin * bp(bins_ + (k % binMax));

        matches.push_back(Match(k << binShift, bp->first, bp->last));

        bp->count = 0;
        bp->last  = 0;
    }
    taggedBins_.clear();

    // Adjust coordinates
    {
        int const d(subjectSize_ - binSize_ + 1);
        int const s(offset);
        int const e(gramSize_ - offset);
        Matches::iterator i(matches.begin());
        Matches::iterator j(matches.end());
        for ( ; i != j; ++i) {
            Match & m(*i);
            m.queryStart  -= s;
            m.queryEnd    += e;
            m.subjectStart = d - m.subjectStart + m.queryStart;
        }
    }

    ic_ = ic + trapMin;
    id_ = id + trapMin;
}


std::ostream & operator<<(
    std::ostream & out,
    EpsilonFilter::MatchProperties const & mp
);


} // end namespace swift

#endif

