// helpcommand.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
//
//
// Additions / changes by Peter Husemann
//

#include "createcommand.hpp"
#include "helpcommand.hpp"
#include "indexcommand.hpp"
#include "infocommand.hpp"
#include "querycommand.hpp"
#include "versioncommand.hpp"
#include <boost/scoped_ptr.hpp>


namespace swift
{


/*!
    Functor to display help information given a command name.
*/
struct DisplayCommand
{
    void operator()(std::string const & commandName);
};


/*!
    Displays the help for the command with name \a commandName.
*/
void DisplayCommand::operator()(std::string const & commandName)
{

    boost::scoped_ptr<SwiftCommand> command;

    if (commandName == "query") {
        command.reset(new QueryCommand);
    } else if (commandName == "create") {
        command.reset(new CreateCommand);
    } else if (commandName == "index") {
        command.reset(new IndexCommand);
    } else if (commandName == "info") {
        command.reset(new InfoCommand);
    } else if (commandName == "help") {
        command.reset(new HelpCommand);
    } else if (commandName == "version") {
        command.reset(new VersionCommand);
    } else {
        std::cout << "Unknown command '" << commandName << "'\n"
            "Type '"<< SwiftCommand::SwiftExecutable <<" help' for usage information." << std::endl;
        return;
    }

    command->displayUsage();
}


/*!
    Constructs a HelpCommand object.
*/
HelpCommand::HelpCommand() :
    visibleOptions_("Valid options")
{
    namespace po = boost::program_options;

    visibleOptions_.add_options()
        ("help,h", "produces this help message.");

    hiddenOptions_.add_options()
        ("commands", po::value<CommandsVector>(&commands_),
        "Commands");
}


void HelpCommand::execute(int argc, char ** argv)
{
  //SwiftCommand::SwiftExecutable= argv[0];

    if (!processArguments(argc, argv))
        return;

    if (commands_.empty()) {
        displayHelp();
    } else {
        displayCommandsHelp();
    }
}


/*!
    Displays usage help for the 'help' command.
*/
void HelpCommand::displayUsage(std::ostream & outStream) const
{
    outStream <<
        "Usage:\n  "
        << SwiftCommand::SwiftExecutable <<" help SUBCOMMAND\n"
        "\n"
        "  Describes the usage of this program or its subcommands.\n"
        "\n"
        "  Examples:\n"
        "    "<< SwiftCommand::SwiftExecutable <<" help index\n"
        "\n"
        << visibleOptions_ << std::endl;
}


/*!
    Displays the usage information for the specified commands.
*/
void HelpCommand::displayCommandsHelp() const
{
    std::for_each(commands_.begin(), commands_.end(), DisplayCommand());
}


/*!
    Displays general help for the swift program.
*/
void HelpCommand::displayHelp() const
{
    std::cout <<
        "Usage:\n"
        "  "<< SwiftCommand::SwiftExecutable <<" <subcommand> [options] [args]\n"
        "\n"
        "  Type '"<< SwiftCommand::SwiftExecutable <<" help <subcommand>' for help on a specific\n"
        "  subcommand.\n"
        "\n"
        "Available subcommands:\n"
        "  create   - database creation\n"
        "  help     - usage help\n"
        "  index    - index construction\n"
        "  info     - database information\n"
        "  query    - database query\n"
        "  version  - Swift version and license information\n"
        "\n"
        "Swift is a tool for DNA sequence alignment.\n"
        "  For additional information, see http://swift.sourceforge.org/\n\n"
        "If you publish results where swift was involved,\n"
        "  please cite the following reference:\n"
        "    K. Rasmussen, J. Stoye, and E. W. Myers.\n"
        "    Efficient q-Gram Filters for Finding All epsilon-Matches over a Given Length\n"
        "    J. Comp. Biol., 13(2):296-308, 2006.\n"
        << std::endl;
}


/*!
    Processes command line arguments. Returns \c true if the arguments
    were successfully processed. Returns \c false if an error occurred, or
    if the arguments contains a 'shortcut' option such as '--help'.
*/
bool HelpCommand::processArguments(int argc, char ** argv)
{
    namespace po = boost::program_options;

    try {
        po::positional_options_description positionals;
        positionals.add("commands", -1);

        po::options_description all;
        all.add(visibleOptions_).add(hiddenOptions_);

        po::variables_map vm;
        po::store(po::command_line_parser(argc, argv).
                  options(all).positional(positionals).run(), vm);
        po::notify(vm);

        if (vm.count("help")) {
            displayUsage();
            return false;
        }
    } catch (po::unknown_option & e) {
      std::cerr <<  SwiftCommand::SwiftExecutable <<": " << e.what() << "\n"
          << "Type '"<< SwiftCommand::SwiftExecutable <<" help' for usage information."
                  << std::endl;
        return false;
    } catch (std::exception & e) {
        std::cerr << e.what() << std::endl;
        return false;
    }

    return true;
}


} // namespace swift
