// infocommand.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
//
//
// Additions / changes by Peter Husemann
//

#include "infocommand.hpp"
#include "swiftdatabase.hpp"
#include "thousands.hpp"
#include <boost/format.hpp>
#include <iomanip>
#include <iostream>

namespace swift
{

/*!
    Constructs a InfoCommand object.
*/
InfoCommand::InfoCommand() :
    visibleOptions_("Valid options")
{
    namespace po = boost::program_options;

    visibleOptions_.add_options()
        ("help,h", "produces this help message.")
        ("verbose,v", "print extra information.");

    hiddenOptions_.add_options()
        ("database", po::value<std::string>(&databasePrefix_),
        "Database prefix");
}


void InfoCommand::execute(int argc, char ** argv)
{
  //SwiftCommand::SwiftExecutable = argv[0];

  if (processArguments(argc, argv))
        showInfo();
}


bool InfoCommand::processArguments(int argc, char ** argv)
{
    namespace po = boost::program_options;

    try {
        po::positional_options_description positionals;
        positionals.add("database", 1);

        po::options_description all;
        all.add(visibleOptions_).add(hiddenOptions_);

        po::variables_map vm;
        po::store(po::command_line_parser(argc, argv).
                  options(all).positional(positionals).run(), vm);
        po::notify(vm);

        if (vm.count("help") || vm.count("database") < 1) {
            displayUsage();
            return false;
        }

    } catch (po::unknown_option & e) {
      std::cerr <<  SwiftCommand::SwiftExecutable <<": " << e.what() << "\n"
            << "Type '"<< SwiftCommand::SwiftExecutable <<" help' for usage information."
                  << std::endl;
        return false;
    } catch (std::exception & e) {
        std::cerr << e.what() << std::endl;
        return false;
    }

    return true;
}


void InfoCommand::showInfo()
{
    SwiftDatabase database(databasePrefix_);

    if (!database.isOpen()) {
        std::cout << "Cannot open database '" << databasePrefix_
            << "'." << std::endl;
        return;
    }

    SwiftDatabase::size_type minLength(database.size());
    SwiftDatabase::size_type maxLength(0);
    for (SwiftDatabase::size_type i(0), j(database.count()); i < j; ++i) {
        SwiftDatabase::size_type length(database.size(i));
        if (length < minLength)
            minLength = length;
        if (length > maxLength)
            maxLength = length;
    }

    SwiftDatabase::size_type averageLength(database.size() / database.count());

    boost::format fmt("%1%");
    fmt.modify_item(1, std::setw(Thousands::size(database.size()) + 3));

    std::cout
        << "Database '" << databasePrefix_ << "'"
        << "\n   sequences              : "
        << fmt % Thousands::string(database.count());
    std::cout
        << "\n   sequence length, min   : "
        << fmt % Thousands::string(minLength);
    std::cout
        << "\n                    max   : "
        << fmt % Thousands::string(maxLength);
    std::cout
        << "\n                    avg   : "
        << fmt % Thousands::string(averageLength);
    std::cout
        << "\n                    total : "
        << fmt % Thousands::string(database.size())
        << std::endl;
}


void InfoCommand::displayUsage(std::ostream & outStream) const
{

}


} // namespace swift
