// querycommand.cpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
//
// Additions / changes by Peter Husemann
//

#include "querycommand.hpp"
#include "blasttablereporter.hpp"
#include "epsilonaligner.hpp"
#include "epsilonfilter.hpp"
#include "gramindex.hpp"
#include "gramview.hpp"
#include "karlinaltschul.hpp"
#include "sequence.hpp"
#include "sort_permute_iterator.hpp"
#include "swiftdatabase.hpp"
#include "timer.hpp"
#include <boost/function_output_iterator.hpp>
#include <boost/scoped_ptr.hpp>
#include <algorithm>
#include <iostream>
#include "version.hpp"


namespace swift
{


class ReporterAppender
{
    public:
    ReporterAppender(Reporter & reporter);

    void operator()(Alignment * alignment) const;
    void setQueryId(unsigned int queryId);
    void setSubjectId(unsigned int subjectId);
    void setForwardQueryPositions();
    void setBackwardQueryPositions(unsigned int querySize);

    private:
    Reporter & reporter_;
    unsigned int queryId_;
    unsigned int subjectId_;
    unsigned int querySize_;
};


inline
ReporterAppender::ReporterAppender(Reporter & reporter) :
    reporter_(reporter),
    queryId_(0),
    subjectId_(0),
    querySize_(0)
{
}


inline
void ReporterAppender::setForwardQueryPositions()
{
    querySize_ = 0;
}


inline
void ReporterAppender::setBackwardQueryPositions(unsigned int querySize)
{
    querySize_ = querySize;
}


inline
void ReporterAppender::operator()(Alignment * alignment) const
{
    if (querySize_ > 0) {
        alignment->queryBegin = querySize_ - alignment->queryBegin;
        alignment->queryEnd   = querySize_ - alignment->queryEnd;
    }
    reporter_.append(alignment, queryId_, subjectId_);
}


inline
void ReporterAppender::setQueryId(unsigned int queryId)
{
    queryId_ = queryId;
}


inline
void ReporterAppender::setSubjectId(unsigned int subjectId)
{
    subjectId_ = subjectId;
}


namespace {

struct OrderBySubjectStart
{
    inline
    bool operator()(
        EpsilonFilter::Match const & a,
        EpsilonFilter::Match const & b
    )
    {
        return a.subjectStart < b.subjectStart;
    }
};

}



unsigned int const QueryCommand::DefaultMinBin = 4;
int const QueryCommand::DefaultMatchScore = 1;
int const QueryCommand::DefaultMismatchScore = -3;
int const QueryCommand::DefaultGapOpenCost = 5;
int const QueryCommand::DefaultGapExtendCost = 2;
double const QueryCommand::DefaultXDrop = 77.0;
unsigned int const QueryCommand::DefaultStrand = 3;
unsigned int const QueryCommand::DefaultMasking = QueryCommand::MaskingSeeding;



/*!
    Constructs a QueryCommand object.
*/
QueryCommand::QueryCommand() :
    visibleOptions_("Valid options")
{
    namespace po = boost::program_options;

    visibleOptions_.add_options()
        ("help,h", ": produces this help message")
        ("verbose,v", ": print extra information")
        ("gram",
         po::value<unsigned int>(&gramSize_)->default_value(11),
         ": q-hit length")
        ("maxerror",
         po::value<double>(&maxMatchError_),
         ": maximum error rate of matches")
        ("minlen",
          po::value<unsigned int>(&minMatchLength_),
         ": minimum length of matches (->length of the query side ?!)")
        ("strand",
         po::value<unsigned int>(&strand_)->
            default_value(QueryCommand::DefaultStrand),
         ": query strand to consider. ARG can be one of:\n"
         "    1 = plus strand\n"
         "    2 = minus strand\n"
         "    3 = both strands")
        ("mask",
         po::value<unsigned int>(&maskingArg_)->
            default_value(QueryCommand::DefaultMasking),
         ": behaviour on masked query sequence. ARG can\n"
         "  be one of:\n"
         "    1 = do not allow epsilon matches in or into\n"
         "        masked sequence.\n"
         "    2 = do not allow seeding of epsilon matches\n"
         "        in masked sequence, but allow extension\n"
         "        into masked sequence.\n"
         "    3 = ignore masking.")
/*
TODO: selfcompare seems to be not implemented
        ("self-compare",
         po::bool_switch(&selfCompare_),
         ": compare subject with itself")*/
        ("match",
         po::value<int>(&matchScore_)->
            default_value(QueryCommand::DefaultMatchScore),
         ": nucleotide match score")
        ("mismatch",
         po::value<int>(&mismatchScore_)->
            default_value(QueryCommand::DefaultMismatchScore),
         ": nucleotide mismatch score")
        ("gapopen",
         po::value<int>(&gapOpenCost_)->
            default_value(QueryCommand::DefaultGapOpenCost),
         ": gap open cost")
        ("gapextend",
         po::value<int>(&gapExtendCost_)->
            default_value(QueryCommand::DefaultGapExtendCost),
         ": gap extend cost")
        ("xdrop",
         po::value<double>(&xDrop_)->
            default_value(QueryCommand::DefaultXDrop),
         ": gapped X-dropoff value in bits")
        ("minbin",
         po::value<unsigned int>(&minBin_)->
            default_value(QueryCommand::DefaultMinBin),
         ": minimum size of bin shift")
        ("output",
         po::value<std::string>(&outputName_),
         ": output the query results to file ARG instead of\n"
         "  to the console")
        ("compassemb",
         po::bool_switch(),
        ": change output to a format optimized\n"
            "  for the comperative assembly project");


    hiddenOptions_.add_options()
        ("subject", po::value<std::string>(&subjectsPrefix_),
         "subject name prefix")
        ("query", po::value<std::string>(&queriesPrefix_),
         "query name prefix");
}


void QueryCommand::execute(int argc, char ** argv)
{
  //SwiftCommand::SwiftExecutable = argv[0];
  if (processArguments(argc, argv))
        querySubject();
}


bool QueryCommand::processArguments(int argc, char ** argv)
{
    namespace po = boost::program_options;

    outputName_.clear();

    try {
        po::positional_options_description positionals;
        positionals.add("subject", 1);
        positionals.add("query", 1);

        po::options_description all;
        all.add(visibleOptions_).add(hiddenOptions_);

        po::variables_map vm;
        po::store(po::command_line_parser(argc, argv).
                  options(all).positional(positionals).run(), vm);
        po::notify(vm);

        if (vm.count("help") || vm.count("query") < 1 ||
            !vm.count("maxerror") || !vm.count("minlen")) {
            displayUsage();
            return false;
        }

        if (maskingArg_ < MaskingCount) {
            masking_ = static_cast<Masking>(maskingArg_);
        } else {
          std::cerr << SwiftCommand::SwiftExecutable <<": Invalid mask argument.\n" << std::endl;
            return false;
        }

        compAssembOutput_ = vm["compassemb"].as<bool>();

    } catch (po::unknown_option & e) {
      std::cerr << SwiftCommand::SwiftExecutable <<": " << e.what() << "\n"
            << "Type '"<< SwiftCommand::SwiftExecutable <<" help' for usage information."
                  << std::endl;
        return false;
    } catch (std::exception & e) {
        std::cerr << e.what() << std::endl;
        return false;
    }

    return true;
}


void QueryCommand::querySubject()
{
    if (!openOutput()) // sets the outputfile if given on the command line
        return;

    openQueries();
    filter();

    openSubjects();
    align();

    closeOutput();
}


void QueryCommand::filter()
{
    using std::cout;
    using std::endl;

    const unsigned int PLUS_STRAND = 0x1;
    const unsigned int MINUS_STRAND = 0x2;
    const unsigned int BOTH_STRANDS = 0x3;

    Timer filterTimer;
    filterTimer.start();

    MaskSensitivity maskSensitivity(
        masking_ == MaskingAll || masking_ == MaskingSeeding ?
        MaskSensitive : MaskInsensitive
    );

    boost::scoped_ptr<GramIndex> gramIndex(
        GramIndex::unserialize(subjectsPrefix_, gramSize_)
    );

    if (!gramIndex) {
        cout << "Cannot open gram index. Aborting" << endl;
        exit(-1);
    }

    EpsilonFilter filter(*gramIndex);
    filter.configure(minMatchLength_, maxMatchError_, minBin_);

    matchZones_.clear();
    matchZones_.reserve(queries_->count());
    unsigned int plusZones = 0;
    unsigned int minusZones= 0;

    EpsilonFilter::Matches * matches = 0;

    for (unsigned int i = 0, j = queries_->count(); i < j; ++i) {
        MatchZones zones;
        boost::scoped_ptr<Sequence> query(queries_->sequence(i));

        //if (strand_ & 0x1) { // FIXME - strand
	if (strand_ == PLUS_STRAND || strand_ == BOTH_STRANDS) { // FIXME - strand
            if (!matches)
                matches = new EpsilonFilter::Matches;

            //GramView<Sequence::const_iterator> gramView(
            GramView<Sequence::value_type const *> gramView(
                query->begin(),
                query->end(),
                gramSize_,
                maskSensitivity
            );

            filter(gramView.begin(), gramView.end(), *matches);

            if (!matches->empty()) {
                zones.plusStrand = matches;
                matches          = 0;
            }
        } else {
            zones.plusStrand = 0;
        }

        //if (strand_ & 0x2) { // FIXME - strand
	if (strand_ == MINUS_STRAND || strand_ == BOTH_STRANDS) { // FIXME - strand
            query->reverseComplement();

            if (!matches)
                matches = new EpsilonFilter::Matches;

            //GramView<Sequence::const_iterator> gramView(
            GramView<Sequence::value_type const *> gramView(
                query->begin(),
                query->end(),
                gramSize_,
                maskSensitivity
            );

            filter(gramView.begin(), gramView.end(), *matches);

            if (!matches->empty()) {
                zones.minusStrand = matches;
                matches           = 0;
            }
        } else {
            zones.minusStrand = 0;
        }

        if (zones.plusStrand || zones.minusStrand) {
            zones.queryId = i;
            matchZones_.push_back(zones);

            if (zones.plusStrand) {
                plusZones += zones.plusStrand->size();
            }
            if (zones.minusStrand) {
                minusZones += zones.minusStrand->size();
            }
        }
    }

    if (matches) {
        delete matches;
    }

    matchProperties_ = filter.matchProperties();

    cout << "\nPlus zones  : " << plusZones
         << "\nMinus zones : " << minusZones
         << endl;

    filterTimer.stop();
    TimeElapsed filterTime(filterTimer.getTimeElapsed());
    cout << "\nFiltration time\n" << filterTime << endl;
}



void QueryCommand::align()
{
    typedef std::vector<MatchZones>::iterator ZIterator;

    Timer verifyTimer;
    verifyTimer.start();

    MaskSensitivity seedingMaskSensitivity(
        masking_ == MaskingAll || masking_ == MaskingSeeding ?
        MaskSensitive : MaskInsensitive
    );

    MaskSensitivity extensionMaskSensitivity(
        masking_ == MaskingAll ? MaskSensitive : MaskInsensitive
    );


    boost::shared_ptr<KarlinAltschul const> ka(
        new KarlinAltschul(matchScore_, mismatchScore_)
    );

    EpsilonAligner aligner;

    aligner.setSeedingScoreFunction(
        ScoreFunction(
            ScoreMatrix(matchScore_, mismatchScore_, seedingMaskSensitivity),
            gapOpenCost_,
            gapExtendCost_
        ),
        matchScore_
    );

    aligner.setSeedingMaskSensitivity(seedingMaskSensitivity);

    aligner.setExtensionScoreFunction(
        ScoreFunction(
            ScoreMatrix(matchScore_, mismatchScore_, extensionMaskSensitivity),
            gapOpenCost_,
            gapExtendCost_
        ),
        KarlinAltschul(matchScore_, mismatchScore_).rawScore(xDrop_)
    );

    aligner.setGramParameters(
        gramSize_,
        matchProperties_.maxGap,
        matchProperties_.minHits
    );

    BlastTableReporter reporter(
        queries_,
        subjects_,
        ka,
        *outputStream_
    );

    ReporterAppender reportAppender(reporter);

    reporter.clear();
    reporter.message(Version::publication());

    for (ZIterator i(matchZones_.begin()), j(matchZones_.end()); i != j; ++i) {
        MatchZones const & mz(*i);
        boost::scoped_ptr<Sequence> query(queries_->sequence(mz.queryId));

        reportAppender.setQueryId(mz.queryId);

        if (mz.plusStrand) {
            reportAppender.setForwardQueryPositions();
            verifyQuery(
                *query,
                *mz.plusStrand,
                matchProperties_,
                aligner,
                reportAppender
            );

            delete mz.plusStrand;
        }

        if (mz.minusStrand) {
            reportAppender.setBackwardQueryPositions(query->size());
            query->reverseComplement();
            verifyQuery(
                *query,
                *mz.minusStrand,
                matchProperties_,
                aligner,
                reportAppender
            );

            delete mz.minusStrand;
        }

        reporter.report(compAssembOutput_);
        reporter.clear();
    }

    verifyTimer.stop();
    TimeElapsed verifyTime(verifyTimer.getTimeElapsed());
    std::cout << "\nVerification time\n" << verifyTime << std::endl;
}


void QueryCommand::verifyQuery(
    Sequence const & query,
    EpsilonFilter::Matches & matches,
    EpsilonFilter::MatchProperties const & matchProperties,
    EpsilonAligner & aligner,
    ReporterAppender & appender
)
{
    typedef EpsilonFilter::Matches::const_iterator MIterator;

    splitMatches(matches, matchProperties, subjectIds_);
    sortMatches(matches, subjectIds_);

    ASSERT(matches.size() == subjectIds_.size(), "Inconsistent vector lengths");

    unsigned int previousSubjectId(subjects_->count());
    boost::scoped_ptr<Sequence const> subject(0);

    unsigned int ii(0);
    for (MIterator i(matches.begin()), j(matches.end()); i != j; ++i, ++ii) {

        unsigned int subjectId(subjectIds_[ii]);
        if (subjectId != previousSubjectId) {
            subject.reset(subjects_->sequence(subjectId));
            appender.setSubjectId(subjectId);
            previousSubjectId = subjectId;
        }

        EpsilonFilter::Match const & m((*i));
        EpsilonRegion region(
            m.queryStart,
            m.queryEnd,
            m.subjectStart - m.queryStart,
            m.subjectEnd(matchProperties.width) - m.queryEnd
        );

        aligner.alignments(
            query,
            *subject,
            region,
            boost::make_function_output_iterator(appender)
        );
    }
}


/*!
    Splits the \a matches such that each match corresponds to one subject
    sequence. The id of the corresponding subject sequence to
    <tt>matches[i]</tt> is found in <tt>ids[i]</tt>. Split matches that
    are shorter than \a minMatchSize are discarded. Hence, on return both
    \a matches and \a ids may be empty.
*/
void QueryCommand::splitMatches(
    EpsilonFilter::Matches & matches,
    EpsilonFilter::MatchProperties const & matchProperties,
    IdVector & ids
) const
{
    ids.resize(matches.size());

    for (size_t i(0), j(matches.size()); i < j; ) {
        EpsilonFilter::Match & m(matches[i]);

        unsigned int si(subjects_->find(std::max(0, m.subjectStart)));
        unsigned int sj(
            subjects_->find(
                std::min(
                    m.subjectEnd(matchProperties.width), (unsigned int)subjects_->size()
                ) - 1,
                si
            )
        );

        if (si == sj) {
            ASSERT(
                m.queryEnd - m.queryStart >= matchProperties.minMatchSize(),
                "Match too short"
            );
            ids[i] = si;
            m.subjectStart -= subjects_->offset(si);
            ++i;
            continue;
        }

        bool splitsFound(false);
        for ( ; si <= sj; ++si) {
            EpsilonFilter::Match n(m);

            int x(subjects_->offset(si + 1) - n.subjectStart + n.queryStart);
            if (x < n.queryEnd)
                n.queryEnd = x;

            x = subjects_->offset(si) - n.subjectStart -
                matchProperties.width + 1;
            if (x < 0)
                x = 0;
            n.queryStart   += x;
            n.subjectStart += x - subjects_->offset(si);

            if (n.queryEnd - n.queryStart < matchProperties.minMatchSize())
                continue;

            matches.push_back(n);
            ids.push_back(si);
            splitsFound = true;
        }

        matches[i] = matches.back();
        ids[i] = ids.back();

        if (splitsFound)
            ++i;

        if (j == matches.size())
            --j;

        matches.pop_back();
        ids.pop_back();
    }
}


/*!
    Sorts the \a matches and the \a ids in increasing order of the ids.
*/
void QueryCommand::sortMatches(
    EpsilonFilter::Matches & matches,
    IdVector & ids
) const
{
    std::sort(
        make_sort_permute_iterator(ids.begin(), matches.begin()),
        make_sort_permute_iterator(ids.end(), matches.end()),
        sort_permute_iterator_compare<
            IdVector::const_iterator,
            EpsilonFilter::Matches::const_iterator
        >()
    );
}


void QueryCommand::openSubjects()
{
    using std::cout;
    using std::endl;
    using std::exception;

    cout << "Opening subjects database: " << subjectsPrefix_ << "\n";

    SwiftDatabase * db(0);
    subjects_.reset(db = new SwiftDatabase(subjectsPrefix_));
    if (!db->isOpen()) {
        cout << "Cannot open subjects database. Aborting." << endl;
        exit(-1);
    }
}


void QueryCommand::openQueries()
{
    using std::cout;
    using std::endl;
    using std::exception;

    cout << "Opening queries database: " << queriesPrefix_ << "\n";

    SwiftDatabase * db(0);
    queries_.reset(db = new SwiftDatabase(queriesPrefix_));
    if (!db->isOpen()) {
        cout << "Cannot open queries database. Aborting." << endl;
        exit(-1);
    }
}


bool QueryCommand::openOutput()
{
    if (outputName_.empty()) {
        outputStream_ = &std::cout;
        outputFile_   = 0;
    } else {
        outputStream_ =
        outputFile_   = new std::ofstream(outputName_.c_str(),
                                          std::ios::out | std::ios::trunc);
        if (!outputFile_->is_open()) {
            std::cerr << "Cannot open `" << outputName_ << "' for output.\n";
            return false;
        }
        outputFile_->sync_with_stdio(false);
    }
    return true;
}


void QueryCommand::closeOutput()
{
    if (outputFile_) {
        outputFile_->flush();
        outputFile_->close();
        delete outputFile_;
        outputFile_ = 0;
    }
}


void QueryCommand::displayUsage(std::ostream & outStream) const
{
    outStream <<
        "Usage:\n"
        "  "<< SwiftCommand::SwiftExecutable <<" query [options] DATABASE [QUERY...]\n"
        "\n"
        "  Queries the DATABASE ... \n"
        "\n"
        "  The options --maxerror, --minlen and --gram are mandatory.\n"
        "\n"
        "Examples:\n"
        "  "<< SwiftCommand::SwiftExecutable <<" query --maxerror 0.01 --minlen 20  --gram 11 sequences query \n"
        "\n"
        "Make sure to create the database and the index of query and database beforehand\n"
        "\n"
        << visibleOptions_ << std::endl;
}


} // namespace swift

