// querycommand.hpp
//
// Copyright (c) 2004-2006, Kim Roland Rasmussen <krr@elreki.net>
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
//
// Additions / changes by Peter Husemann
//

#ifndef _SWIFT_QUERYCOMMAND_
#define _SWIFT_QUERYCOMMAND_

#include "epsilonfilter.hpp"
#include "masksensitivity.hpp"
#include "swiftcommand.hpp"
#include <boost/program_options.hpp>
#include <boost/shared_ptr.hpp>
#include <fstream>
#include <iostream>
#include <string>
#include <vector>


namespace swift
{

class GramIndex;
class SequenceDatabase;

class EpsilonAligner;
class ReporterAppender;

class QueryCommand : public SwiftCommand
{
    public:
    QueryCommand();

    virtual void execute(int argc, char ** argv);
    virtual void displayUsage(std::ostream & outStream = std::cout) const;

    private:
    typedef std::vector<unsigned int> IdVector;

    enum Masking
    {
        MaskingAll = 1,
        MaskingSeeding = 2,
        MaskingIgnore = 3,
        MaskingCount
    };

    bool processArguments(int argc, char ** argv);
    void querySubject();
    void filter();
    void align();


    void verifyQuery(
        Sequence const & query,
        EpsilonFilter::Matches & matches,
        EpsilonFilter::MatchProperties const & matchProperties,
        EpsilonAligner & aligner,
        ReporterAppender & appender
    );

    void splitMatches(
        EpsilonFilter::Matches & matches,
        EpsilonFilter::MatchProperties const & matchProperties,
        IdVector & ids
    ) const;

    void sortMatches(
        EpsilonFilter::Matches & matches,
        IdVector & ids
    ) const;

    void openSubjects();
    void openQueries();

    bool openOutput();
    void closeOutput();

    struct MatchZones
    {
        MatchZones() : queryId(0), plusStrand(0), minusStrand(0) {}

        unsigned int queryId;
        EpsilonFilter::Matches * plusStrand;
        EpsilonFilter::Matches * minusStrand;
    };


    boost::program_options::options_description visibleOptions_;
    boost::program_options::options_description hiddenOptions_;

    std::string subjectsPrefix_;
    std::string queriesPrefix_;
    unsigned int gramSize_;
    double maxMatchError_;
    unsigned int minMatchLength_;
    unsigned int strand_;
    unsigned int maskingArg_;
    Masking masking_;
    bool selfCompare_;
    int matchScore_;
    int mismatchScore_;
    int gapOpenCost_;
    int gapExtendCost_;
    double xDrop_;
    unsigned int minBin_;
    MaskSensitivity maskSensitivity_;

    std::string outputName_;
    std::ostream * outputStream_;
    std::ofstream * outputFile_;
    bool compAssembOutput_;

    std::vector<MatchZones> matchZones_;
    EpsilonFilter::MatchProperties matchProperties_;

    IdVector subjectIds_;

    boost::shared_ptr<SequenceDatabase> queries_;
    boost::shared_ptr<SequenceDatabase> subjects_;


    static unsigned int const DefaultMinBin;
    static int const DefaultMatchScore;
    static int const DefaultMismatchScore;
    static int const DefaultGapOpenCost;
    static int const DefaultGapExtendCost;
    static double const DefaultXDrop;
    static unsigned int const DefaultStrand;
    static unsigned int const DefaultMasking;
};


} // namespace swift

#endif

