#!/bin/bash

WORK_DIR="`pwd`/validation_work"
REF_PATH="`pwd`/reference.fasta"
SCAF_PATH="`pwd`/scaffolds.fasta"
HOSTFILE="`pwd`/hosts"
LOCALWORK="/node/local/path/"
GRAM=11
MAXGAP=0
MAXERROR=0.05
MINLEN=30
NUMJOBS=10

usage() {
	echo
	echo "Usage:"
	echo "Parameter		(Default value)"
	echo
	echo "--------------------------------------------"
	echo "-h | --help			Display this text"
	echo "--maxerror VALUE	(=$MAXERROR)	Allow error rate of VALUE when doing alignments with swift"
	echo "--minlen VALUE		(=$MINLEN)	Only report alignments with length >= VALUE"
	echo "--maxgap VALUE		(=$MAXGAP)	Restrict gaps to at max VALUE when doing col. chaining (0 = unlimited)"
	echo "--gram VALUE		(=$GRAM)	Use VALUE-gram for swift"
	echo "-r | --reference PATH	(=$REF_PATH)	Path to reference sequence"
	echo "-s | --scaffolds PATH	(=$SCAF_PATH)	Path to scaffolds"
	echo "-w | --work PATH	(=$WORK_DIR)	Absolute path to final and intermediary results directory (must be visible to all nodes)"
	echo "-o | --hosts PATH	(=$HOSTFILE)	Path to hosts file"
	echo "-l | --localwork PATH	(=$LOCALWORK)	Absolute path to local work directory (will be same on all nodes)"
	echo "-n | --numjobs NUMJOBS	(=$NUMJOBS)	Number of distributed jobs to be created"
	echo
}

SCRIPT_DIR="$(cd -P "$(dirname "${BASH_SOURCE[0]}")" && pwd)"
if [ $# -lt 2 ]; then
	usage
	exit 1
fi

while [ $# -gt 0 ]; do
	case $1 in 
		-h|--help) usage; exit 1;;
		--maxerror) MAXERROR=$2; shift 2;;
		--minlen) MINLEN=$2; shift 2;;
		--maxgap) MAXGAP=$2; shift 2;;
		--gram) GRAM=$2; shift 2;;
		-r|--reference) REF_PATH=$2; shift 2;;
		-s|--scaffolds) SCAF_PATH=$2; shift 2;;
		-w|--work) WORK_DIR=$2; shift 2;;
		-o|--hosts) HOSTFILE=$2; shift 2;;
		-l|--localwork) LOCALWORK=$2; shift 2;;
		-n|--numjobs) NUMJOBS=$2; shift 2;;
		*) shift 1;;
	esac
done
SCAFID=`basename $SCAF_PATH`
SCAFID=${SCAFID%.*}
REFID="`basename $REF_PATH`"
REFID=${REFID%.*}

REFMD5=`openssl dgst -md5 $REF_PATH | cut -f 2 -d ' '`
REFMD5=${REFMD5:0:6}
SCAFMD5=`openssl dgst -md5 $SCAF_PATH | cut -f 2 -d ' '`
SCAFMD5=${SCAFMD5:0:6}

LOCALWORK="${LOCALWORK}/${REFID}_$REFMD5/${SCAFID}_$SCAFMD5/"

IDENTIFIER="${REFID}_$SCAFID"

echo "Using settings"
echo "--------------"
echo "Maxerror	$MAXERROR"
echo "Minlength	$MINLEN"
echo "Maxgap		$MAXGAP"
echo "Gram		$GRAM"
echo
echo "Work dir	$WORK_DIR"
echo "Reference	$REF_PATH"
echo "Scaffolds	$SCAF_PATH"
echo "Scaffolds identifier	$SCAFID"
echo "Reference identifier	$REFID"
echo "Local working directory	$LOCALWORK"

if [ ! -f $SCAF_PATH ]; then
	echo "Could not find file $SCAF_PATH"
	exit 1
fi

if [ ! -f $REF_PATH ]; then
	echo "Could not find file $REF_PATH"
	exit 1
fi
echo "--------------"

## Create working directory if doesn't exist
mkdir -p $WORK_DIR

## Concatenate reference sequence
echo "$SCRIPT_DIR/bin/concatenate_fasta $REF_PATH $WORK_DIR/concat_reference_${REFID}.fasta $WORK_DIR/boundaries_$REFID"
$SCRIPT_DIR/bin/concatenate_fasta $REF_PATH $WORK_DIR/concat_reference_${REFID}.fasta $WORK_DIR/boundaries_$REFID

## Build swift index
echo "$SCRIPT_DIR/swift/build/swift/swift create $WORK_DIR/index_$REFID $WORK_DIR/concat_reference_${REFID}.fasta"
$SCRIPT_DIR/swift/build/swift/swift create $WORK_DIR/index_$REFID $WORK_DIR/concat_reference_${REFID}.fasta
echo "$SCRIPT_DIR/swift/build/swift/swift index $WORK_DIR/index_$REFID --gram $GRAM"
$SCRIPT_DIR/swift/build/swift/swift index $WORK_DIR/index_$REFID --gram $GRAM

## Generate distributable jobs
echo "$SCRIPT_DIR/src/generate_jobs.py -i $WORK_DIR/index_$REFID -E $SCAF_PATH -r $WORK_DIR -n $NUMJOBS -g $GRAM -e $MAXERROR -l $MINLEN -w $LOCALWORK -s $SCRIPT_DIR/swift/build/swift"
$SCRIPT_DIR/src/generate_jobs.py -i $WORK_DIR/index_$REFID -E $SCAF_PATH -r $WORK_DIR -n $NUMJOBS -g $GRAM -e $MAXERROR -l $MINLEN -w $LOCALWORK -s $SCRIPT_DIR/swift/build/swift

## Divide EST sequences into chunks of somewhat even swift computation time

echo "$SCRIPT_DIR/bin/split_seqs $SCAF_PATH $WORK_DIR/querySeq_$SCAFID $NUMJOBS"
$SCRIPT_DIR/bin/split_seqs $SCAF_PATH $WORK_DIR/querySeq_$SCAFID $NUMJOBS

## Distribute the computations

chmod 700 $WORK_DIR/pre_jobs_* $WORK_DIR/post_jobs_*

echo "python $SCRIPT_DIR/src/distribute.py -m 7000 -o $HOSTFILE -j $WORK_DIR/jobs_index_$IDENTIFIER -p $WORK_DIR/pre_jobs_index_$IDENTIFIER -s $WORK_DIR/post_jobs_index_$IDENTIFIER -w $WORK_DIR/pre_task_jobs_index_$IDENTIFIER"
python $SCRIPT_DIR/src/distribute.py -m 7000 -o $HOSTFILE -j $WORK_DIR/jobs_index_$IDENTIFIER -p $WORK_DIR/pre_jobs_index_$IDENTIFIER -s $WORK_DIR/post_jobs_index_$IDENTIFIER -w $WORK_DIR/pre_task_jobs_index_$IDENTIFIER

## Concatenate results

rm -f $WORK_DIR/${IDENTIFIER}.results
rm -f $WORK_DIR/reordered_scafs.fasta

i=0
while [ ! $i -eq $NUMJOBS ]
do
	cat $WORK_DIR/index_${IDENTIFIER}_${i}.results >> $WORK_DIR/${IDENTIFIER}.results
	cat $WORK_DIR/querySeq_${SCAFID}_${i}.fasta >> $WORK_DIR/reordered_scafs.fasta
	i=$(($i+1))
done

## Get scaffold lengths
echo "$SCRIPT_DIR/bin/calc_estl $WORK_DIR/reordered_scafs.fasta > $WORK_DIR/estl_$SCAFID"
$SCRIPT_DIR/bin/calc_estl $WORK_DIR/reordered_scafs.fasta > $WORK_DIR/estl_$SCAFID

## Extract tuples from swift results
echo "$SCRIPT_DIR/bin/extract_tuples $WORK_DIR/$IDENTIFIER.results $WORK_DIR/estl_$SCAFID $WORK_DIR/tuples_$IDENTIFIER"
$SCRIPT_DIR/bin/extract_tuples $WORK_DIR/$IDENTIFIER.results $WORK_DIR/estl_$SCAFID $WORK_DIR/tuples_$IDENTIFIER

## Perform colinear chaining
echo "$SCRIPT_DIR/bin/col_chain $WORK_DIR/tuples_$IDENTIFIER $WORK_DIR/boundaries_$REFID $WORK_DIR/correct_alignments_$IDENTIFIER $MAXGAP"
$SCRIPT_DIR/bin/col_chain $WORK_DIR/tuples_$IDENTIFIER $WORK_DIR/boundaries_$REFID $WORK_DIR/correct_alignments_$IDENTIFIER $MAXGAP

GENOME_LENGTH=`grep '^[^>]' $WORK_DIR/concat_reference_${REFID}.fasta | tr -d '\n' | wc -m`
NUMSCAFFOLDS=`grep '^>' $SCAF_PATH | wc -l`

## Calculate final values
echo "$SCRIPT_DIR/bin/calc_coverages $WORK_DIR/correct_alignments_$IDENTIFIER $WORK_DIR/reordered_scafs.fasta $GENOME_LENGTH > $WORK_DIR/validation_results.txt"
$SCRIPT_DIR/bin/calc_coverages $WORK_DIR/correct_alignments_$IDENTIFIER $WORK_DIR/reordered_scafs.fasta $GENOME_LENGTH > $WORK_DIR/validation_results.txt
echo "$SCRIPT_DIR/bin/calc_norm_n50 $WORK_DIR/correct_alignments_$IDENTIFIER $WORK_DIR/reordered_scafs.fasta >> $WORK_DIR/validation_results.txt"
$SCRIPT_DIR/bin/calc_norm_n50 $WORK_DIR/correct_alignments_$IDENTIFIER $WORK_DIR/reordered_scafs.fasta $GENOME_LENGTH >> $WORK_DIR/validation_results.txt
echo "$SCRIPT_DIR/src/png_script.sh $WORK_DIR/correct_alignments_$IDENTIFIER $WORK_DIR/${IDENTIFIER} $NUMSCAFFOLDS"
$SCRIPT_DIR/src/png_script.sh $WORK_DIR/correct_alignments_$IDENTIFIER $WORK_DIR/${IDENTIFIER} $NUMSCAFFOLDS

cat $WORK_DIR/validation_results.txt
