#!/bin/bash

# This script does the same as validate.sh but with a precomputed swift results file

WORK_DIR="`pwd`/validation_work"
REF_PATH="`pwd`/reference.fasta"
SCAF_PATH="`pwd`/scaffolds.fasta"
SWIFT_PATH="`pwd`/swift.results"

MAXGAP=0

usage() {
	echo validate_noswift.sh: A script for running validation based on a precomputed swift alignments file
	echo
	echo "Usage:"
	echo "Parameter		(Default value)"
	echo
	echo "--------------------------------------------"
	echo "-h | --help			Display this text"
	echo "--maxgap VALUE		(=$MAXGAP)	Restrict gaps to at max VALUE when doing colinear chaining (0 = unlimited)"
	echo "-t | --swift	PATH	(=$SWIFT_PATH)	Path to swift results file"
	echo "-r | --reference PATH	(=$REF_PATH)	Path to reference sequence"
	echo "-s | --scaffolds PATH	(=$SCAF_PATH)	Path to scaffolds"
	echo "-w | --work PATH	(=$WORK_DIR)	Where to store intermediary and final results"
	echo
}

SCRIPT_DIR="`dirname $0`"

if [ $# -lt 2 ]; then
	usage
	exit 1
fi

while [ $# -gt 0 ]; do
	case $1 in 
		-h|--help) usage; exit 1;;
		--maxgap) MAXGAP=$2; shift 2;;
		-t|--swift) SWIFT_PATH=$2; shift 2;;
		-r|--reference) REF_PATH=$2; shift 2;;
		-s|--scaffolds) SCAF_PATH=$2; shift 2;;
		-w|--work) WORK_DIR=$2; shift 2;;
		*) shift 1;;
	esac
done
IDENTIFIER="`basename $WORK_DIR`"

echo "Using settings"
echo "--------------"
echo "Maxgap		$MAXGAP"
echo
echo "Work dir	$WORK_DIR"
echo "Swift results	$SWIFT_PATH"
echo "Reference	$REF_PATH"
echo "Scaffolds	$SCAF_PATH"
echo "Identifier	$IDENTIFIER"

if [ ! -f $SCAF_PATH ]; then
	echo "Could not find file $SCAF_PATH"
	exit 1
fi

if [ ! -f $REF_PATH ]; then
	echo "Could not find file $REF_PATH"
	exit 1
fi

if [ ! -f $SWIFT_PATH ]; then
	echo "Could not find file $SWIFT_PATH"
	exit 1
fi
echo "--------------"

## Create working directory if doesn't exist
mkdir -p $WORK_DIR

## Concatenate first!
echo "$SCRIPT_DIR/bin/concatenate_fasta $REF_PATH $WORK_DIR/concat_reference_${IDENTIFIER}.fasta $WORK_DIR/boundaries_$IDENTIFIER"
$SCRIPT_DIR/bin/concatenate_fasta $REF_PATH $WORK_DIR/concat_reference_${IDENTIFIER}.fasta $WORK_DIR/boundaries_$IDENTIFIER

## Get scaffold lengths
echo "$SCRIPT_DIR/bin/calc_estl $SCAF_PATH > $WORK_DIR/scl_$IDENTIFIER"
$SCRIPT_DIR/bin/calc_estl $SCAF_PATH > $WORK_DIR/scl_$IDENTIFIER

## Extract tuples from swift results
echo "$SCRIPT_DIR/bin/extract_tuples $SWIFT_PATH $WORK_DIR/scl_$IDENTIFIER $WORK_DIR/tuples_$IDENTIFIER"
$SCRIPT_DIR/bin/extract_tuples $SWIFT_PATH $WORK_DIR/scl_$IDENTIFIER $WORK_DIR/tuples_$IDENTIFIER

## Perform colinear chaining
echo "$SCRIPT_DIR/bin/col_chain $WORK_DIR/tuples_$IDENTIFIER $WORK_DIR/boundaries_$IDENTIFIER $WORK_DIR/correct_alignments_$IDENTIFIER $MAXGAP"
$SCRIPT_DIR/bin/col_chain $WORK_DIR/tuples_$IDENTIFIER $WORK_DIR/boundaries_$IDENTIFIER $WORK_DIR/correct_alignments_$IDENTIFIER $MAXGAP

GENOME_LENGTH=`grep '^[^>]' $WORK_DIR/concat_reference_${IDENTIFIER}.fasta | tr -d '\n' | wc -m`
NUMSCAFFOLDS=`grep '^>' $SCAF_PATH | wc -l`

## Calculate final values
echo "$SCRIPT_DIR/bin/calc_coverages $WORK_DIR/correct_alignments_$IDENTIFIER $SCAF_PATH $GENOME_LENGTH > $WORK_DIR/validation_results.txt"
$SCRIPT_DIR/bin/calc_coverages $WORK_DIR/correct_alignments_$IDENTIFIER $SCAF_PATH $GENOME_LENGTH > $WORK_DIR/validation_results.txt
echo "$SCRIPT_DIR/bin/calc_norm_n50 $WORK_DIR/correct_alignments_$IDENTIFIER $SCAF_PATH >> $WORK_DIR/validation_results.txt"
$SCRIPT_DIR/bin/calc_norm_n50 $WORK_DIR/correct_alignments_$IDENTIFIER $SCAF_PATH $GENOME_LENGTH >> $WORK_DIR/validation_results.txt
echo "$SCRIPT_DIR/src/png_script.sh $WORK_DIR/correct_alignments_$IDENTIFIER $WORK_DIR/${IDENTIFIER} $NUMSCAFFOLDS"
$SCRIPT_DIR/src/png_script.sh $WORK_DIR/correct_alignments_$IDENTIFIER $WORK_DIR/${IDENTIFIER} $NUMSCAFFOLDS

cat $WORK_DIR/validation_results.txt
